/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "Canon_BJC_8500_Instance.hpp"
#include "defines.hpp"

#include <iostream>
#include <strstream>

const static bool fTestNoCompression = true;

typedef enum {
   CANON_MEDIA_ID_PLAIN     =  0,
   CANON_MEDIA_ID_COATED    =  1,
   CANON_MEDIA_ID_TRANS     =  2,
   CANON_MEDIA_ID_BACKPRINT =  3,
   CANON_MEDIA_ID_CLOTH     =  4,
   CANON_MEDIA_ID_GLOSSY    =  5,
   CANON_MEDIA_ID_GLOSSFILM =  6,
   CANON_MEDIA_ID_HIGHRES   =  7,
   CANON_MEDIA_ID_ENVELOPE  =  8,
   CANON_MEDIA_ID_POSTCARD  =  9,
   CANON_MEDIA_ID_THICK     =  9,
   CANON_MEDIA_ID_FULLXXXXX = 10,
   CANON_MEDIA_ID_BANNER    = 11,
   CANON_MEDIA_ID_OTHER     = 15
} CANON_MEDIA_ID;

Canon_BJC_8500_Instance::
Canon_BJC_8500_Instance (Device *pDevice)
   : DeviceInstance (pDevice)
{
   if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::Canon_BJC_8500_Instance ()" << endl;

   ptlPrintHead_d.x    = 0;
   ptlPrintHead_d.y    = 0;
   fHaveInitialized_d  = false;
   fHaveSetupPrinter_d = false;
}

Canon_BJC_8500_Instance::
~Canon_BJC_8500_Instance ()
{
   if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::~Canon_BJC_8500_Instance ()" << endl;
}

void Canon_BJC_8500_Instance::
initializeInstance ()
{
   if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::initializeInstance ()" << endl;

   if (fHaveInitialized_d)
      return;

   fHaveInitialized_d = true;
}

void Canon_BJC_8500_Instance::
setupPrinter ()
{
   if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::setupPrinter ()" << endl;

   if (fHaveSetupPrinter_d)
      return;

   fHaveSetupPrinter_d = true;

   setPrintColor ();

   DevicePrintMode *pDPM    = getCurrentPrintMode ();

   DeviceCommand *pCommands = getCommands ();
   DeviceData    *pData     = getDeviceData ();
   BinaryData    *pCmd      = 0;

   /* The default is:
   **    <deviceData name="cmdSetPageMode" type="binary">_ESC_ "(a" HEX2S(00,01) HEX(01)</deviceData>
   */
   if (pData)
   {
      // search for: <deviceData name="cmdSetPageMode" type="binary">_ESC_ "(a" HEX2S(00,01) HEX(01)</deviceData>
      if (pData->getBinaryData ("cmdSetPageMode", &pCmd))
      {
         pDevice_d->sendBinaryDataToDevice (pCmd);
      }
   }
   else
   {
      if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::setupPrinter () Error: There is no device data for this device!" << endl;
   }

   pCmd = pCommands->getCommandData ("cmdSetPageID");
   if (pCmd)
   {
      if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::setupPrinter () cmdPageID = " << *pCmd << endl;

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }
   else
   {
      if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::setupPrinter () Error: There is no cmdPageID defined for this device!" << endl;
   }

   HardCopyCap *pHCC = getCurrentForm ()->getHardCopyCap ();
   int          cx   = pHCC->getCx ();
   int          cy   = pHCC->getCy ();

   /* The default is:
   **    <deviceData name="cmdSetPageMargins" type="binary">_ESC_ "(g" HEX2S(00,03) "%c" HEX(01) "%c"</deviceData>
   **    <deviceData name="minPageLength" type="integer">220</deviceData>
   **    <deviceData name="minRightMargin" type="integer">80</deviceData>
   */
   if (pData)
   {
      // search for: <deviceData name="cmdSetPageMargins" type="binary">_ESC_ "(g" HEX2S(00,03) "%c" HEX(01) "%c"</deviceData>
      if (pData->getBinaryData ("cmdSetPageMargins", &pCmd))
      {
         // convert from hundreds of a millimeter to tens of an inch
         int iPageLength     = cy / 254;
         int iRightMargin    = cx / 254;
         int iMinPageLength  = 0;
         int iMinRightMargin = 0;

         if (  pData->getIntData ("minPageLength", &iMinPageLength)
            && pData->getIntData ("minRightMargin", &iMinRightMargin)
            )
         {
            pDevice_d->sendPrintfToDevice (pCmd,
                                           min (iMinPageLength, iPageLength),
                                           min (iMinRightMargin, iRightMargin));
         }
         else
         {
            if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::setupPrinter: Error: could not find minPageLength or minPageLength" << endl;
         }
      }
      // search for: <deviceData name="cmdSetPageMargins2" type="binary">_ESC_ "(p" HEX2S(00,08) "%d%d"</deviceData>
      else if (pData->getBinaryData ("cmdSetPageMargins2", &pCmd))
      {
         // convert from hundreds of a millimeter to sixtys of an inch
         int iPageLength     = (6 * cy) / 254;
         int iRightMargin    = (6 * cx) / 254;
         int iMaxRightMargin = 0;

         // search for: <deviceData name="maxRightMargin" type="integer">480</deviceData>
         if (pData->getIntData ("maxRightMargin", &iMaxRightMargin))
         {
            pDevice_d->sendPrintfToDevice (pCmd,
                                           min (1380, iPageLength),
                                           min (iMaxRightMargin, iRightMargin));
         }
         else
         {
            if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::setupPrinter: Error: could not find maxRightMargin" << endl;
         }
      }
      else
      {
         if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::setupPrinter: Error: could not find cmdSetPageMargins or cmdSetPageMargins2" << endl;
      }
   }

   DeviceResolution *pDR = getCurrentResolution ();

   // Select the resolution
   pDevice_d->sendBinaryDataToDevice (pDR);

   /* The default is:
   **    <deviceData name="cmdSetImage" type="binary">_ESC_ "(t" HEX2S(00,03) "%c" HEX(00) "%c"</deviceData>
   **    <deviceData name="setImageParm1" type="integer">1</deviceData>
   **    <deviceData name="setImageParm2" type="integer">0</deviceData>
   */
   if (pData)
   {
      bool fHackCmdSetImage = false;
      byte bParm1           = 0;
      byte bParm2           = 0;

      // search for: <deviceData name="hackCmdSetImage" type="boolean">true</deviceData>
      pData->getBooleanData ("hackCmdSetImage", &fHackCmdSetImage);

      // search for: <deviceData name="cmdSetImage" type="binary">_ESC_ "(t" HEX2S(00,03) "%c" HEX(00) "%c"</deviceData>
      if (  pData->getBinaryData ("cmdSetImage", &pCmd)
         && pData->getByteData ("setImageParm1", &bParm1)
         && pData->getByteData ("setImageParm2", &bParm2)
         )
      {
         if (DevicePrintMode::COLOR_TECH_CMYK == pDPM->getColorTech ())
         {
             bParm2 = 0x01;
         }
         else
         {
             if (DevicePrintMode::COLOR_TECH_CcMmYK == pDPM->getColorTech ())
                 bParm2 = 0x04;
         }

         pDevice_d->sendPrintfToDevice (pCmd,
                                        bParm1,
                                        bParm2);
      }
      else
      {
         if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::setupPrinter: Error: could not find cmdSetImage or ..." << endl;
      }
   }

   DeviceTray *pDT = getCurrentTray ();

   /* The default is:
   **    The command attached to the tray
   */
   if (pData)
   {
      // search for: <deviceData name="cmdSetTray" type="binary">_ESC_ "(l" HEX2S(00,02) "%c%c"</deviceData>
      if (pData->getBinaryData ("cmdSetTray", &pCmd))
      {
         BinaryData  *pBD      = pDT->getData ();
         DeviceMedia *pDM      = getCurrentMedia ();
         int          iMediaID = pDM->getID ();
         byte         bParm1   = 32 + pBD->getData ()[5]; // @TBD UGLY HACK
         byte         bParm2   = 0;

         switch (iMediaID)
         {
         case DeviceMedia::MEDIA_PLAIN: bParm2 = CANON_MEDIA_ID_PLAIN; break;
         case DeviceMedia::MEDIA_COATED: bParm2 = CANON_MEDIA_ID_COATED; break;
         case DeviceMedia::MEDIA_TRANSPARENCY: bParm2 = CANON_MEDIA_ID_TRANS; break;
         case DeviceMedia::MEDIA_BACKPRINT: bParm2 = CANON_MEDIA_ID_BACKPRINT; break;
         case DeviceMedia::MEDIA_CLOTH: bParm2 = CANON_MEDIA_ID_CLOTH; break;
         case DeviceMedia::MEDIA_GLOSSY: bParm2 = CANON_MEDIA_ID_GLOSSY; break;
         case DeviceMedia::MEDIA_HIGH_GLOSS_FILM: bParm2 = CANON_MEDIA_ID_GLOSSFILM; break;
         case DeviceMedia::MEDIA_HIGH_RESOLUTION: bParm2 = CANON_MEDIA_ID_HIGHRES; break;
         case DeviceMedia::MEDIA_ENVELOPE: bParm2 = CANON_MEDIA_ID_ENVELOPE; break;
         case DeviceMedia::MEDIA_POSTCARD: bParm2 = CANON_MEDIA_ID_POSTCARD; break;
         case DeviceMedia::MEDIA_THICK: bParm2 = CANON_MEDIA_ID_THICK; break;
////////////case DeviceMedia::: bParm2 = CANON_MEDIA_ID_FULLXXXXX; break;
////////////case DeviceMedia::: bParm2 = CANON_MEDIA_ID_BANNER; break;
         case DeviceMedia::MEDIA_OTHER: bParm2 = CANON_MEDIA_ID_OTHER; break;
         }

         pDevice_d->sendPrintfToDevice (pCmd,
                                        bParm1,
                                        bParm2);
      }
      else
      {
         pDevice_d->sendBinaryDataToDevice (pDT);
      }
   }
   else
   {
      pDevice_d->sendBinaryDataToDevice (pDT);
   }
}

void Canon_BJC_8500_Instance::
setPrintColor ()
{
   DevicePrintMode *pDPM  = getCurrentPrintMode ();
   DeviceData      *pData = getDeviceData ();
   BinaryData      *pCmd  = 0;

   if (!pData)
   {
      if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::setPrintColor () Error: There is no device data for this device!" << endl;
      return;
   }

   bool fHackCmdSetMono = false;

   // search for: <deviceData name="hackCmdSetMono" type="boolean">true</deviceData>
   pData->getBooleanData ("hackCmdSetMono", &fHackCmdSetMono);

   if (DevicePrintMode::COLOR_TECH_K == pDPM->getColorTech ())
   {
      // Monochrome default
      // search for: <deviceData name="cmdSetMono" type="binary">_ESC_ "(c" HEX2S(00,01) HEX(01)</deviceData>
      if (pData->getBinaryData ("cmdSetMono", &pCmd))
      {
         pDevice_d->sendBinaryDataToDevice (pCmd);
      }
      // search for: <deviceData name="cmdSetMono2" type="binary">_ESC_ "(c" HEX2S(00,03) "%c%c%c"</deviceData>
      else if (pData->getBinaryData ("cmdSetMono2", &pCmd))
      {
         DeviceResolution *pDR      = getCurrentResolution ();
         DeviceMedia      *pDM      = getCurrentMedia ();
         int               iMediaID = pDM->getID ();
         byte              bParm1   = 49;
         byte              bParm2   = 0;
         byte              bParm3   = 0;

         switch (iMediaID)
         {
         case DeviceMedia::MEDIA_PLAIN: bParm2 = CANON_MEDIA_ID_PLAIN; break;
         case DeviceMedia::MEDIA_COATED: bParm2 = CANON_MEDIA_ID_COATED; break;
         case DeviceMedia::MEDIA_TRANSPARENCY: bParm2 = CANON_MEDIA_ID_TRANS; break;
         case DeviceMedia::MEDIA_BACKPRINT: bParm2 = CANON_MEDIA_ID_BACKPRINT; break;
         case DeviceMedia::MEDIA_CLOTH: bParm2 = CANON_MEDIA_ID_CLOTH; break;
         case DeviceMedia::MEDIA_GLOSSY: bParm2 = CANON_MEDIA_ID_GLOSSY; break;
         case DeviceMedia::MEDIA_HIGH_GLOSS_FILM: bParm2 = CANON_MEDIA_ID_GLOSSFILM; break;
         case DeviceMedia::MEDIA_HIGH_RESOLUTION: bParm2 = CANON_MEDIA_ID_HIGHRES; break;
         case DeviceMedia::MEDIA_ENVELOPE: bParm2 = CANON_MEDIA_ID_ENVELOPE; break;
         case DeviceMedia::MEDIA_POSTCARD: bParm2 = CANON_MEDIA_ID_POSTCARD; break;
         case DeviceMedia::MEDIA_THICK: bParm2 = CANON_MEDIA_ID_THICK; break;
////////////case DeviceMedia::: bParm2 = CANON_MEDIA_ID_FULLXXXXX; break;
////////////case DeviceMedia::: bParm2 = CANON_MEDIA_ID_BANNER; break;
         case DeviceMedia::MEDIA_OTHER: bParm2 = CANON_MEDIA_ID_OTHER; break;
         }

         if (300 > pDR->getYRes ())
         {
            bParm3 = 0;
         }
         else if (600 < pDR->getYRes ())
         {
            bParm3 = 0x12;
         }
         else
         {
            bParm3 = 0x11;
         }
         if (fHackCmdSetMono)
         {
            bParm3 = 0;
         }

         pDevice_d->sendPrintfToDevice (pCmd,
                                        bParm1,
                                        bParm2,
                                        bParm3);
      }
      // search for: <deviceData name="cmdSetMono3" type="binary">_ESC_ "(c" HEX2S(00,02) HEX(11) HEX(00)</deviceData>
      else if (pData->getBinaryData ("cmdSetMono3", &pCmd))
      {
         pDevice_d->sendBinaryDataToDevice (pCmd);
      }
      else
      {
         if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::setPrintColor () Error: There is no cmdSetMono or cmdSetMono2 for this device!" << endl;
      }
   }
   else
   {
      // Color
      // search for: <deviceData name="cmdSetColor" type="binary">_ESC_ "(c" HEX2S(00,03) HEX(10) HEX(11) HEX(10)</deviceData>
      if (pData->getBinaryData ("cmdSetColor", &pCmd))
      {
         pDevice_d->sendBinaryDataToDevice (pCmd);
      }
      // search for: <deviceData name="cmdSetColor2" type="binary">_ESC_ "(c" HEX2S(00,03) HEX(30) "%c%c"</deviceData>
      else if (pData->getBinaryData ("cmdSetColor2", &pCmd))
      {
         DeviceResolution *pDR      = getCurrentResolution ();
         DeviceMedia      *pDM      = getCurrentMedia ();
         int               iMediaID = pDM->getID ();
         byte              bParm1   = 0;
         byte              bParm2   = 0;

         switch (iMediaID)
         {
         case DeviceMedia::MEDIA_PLAIN: bParm1 = CANON_MEDIA_ID_PLAIN; break;
         case DeviceMedia::MEDIA_COATED: bParm1 = CANON_MEDIA_ID_COATED; break;
         case DeviceMedia::MEDIA_TRANSPARENCY: bParm1 = CANON_MEDIA_ID_TRANS; break;
         case DeviceMedia::MEDIA_BACKPRINT: bParm1 = CANON_MEDIA_ID_BACKPRINT; break;
         case DeviceMedia::MEDIA_CLOTH: bParm1 = CANON_MEDIA_ID_CLOTH; break;
         case DeviceMedia::MEDIA_GLOSSY: bParm1 = CANON_MEDIA_ID_GLOSSY; break;
         case DeviceMedia::MEDIA_HIGH_GLOSS_FILM: bParm1 = CANON_MEDIA_ID_GLOSSFILM; break;
         case DeviceMedia::MEDIA_HIGH_RESOLUTION: bParm1 = CANON_MEDIA_ID_HIGHRES; break;
         case DeviceMedia::MEDIA_ENVELOPE: bParm1 = CANON_MEDIA_ID_ENVELOPE; break;
         case DeviceMedia::MEDIA_POSTCARD: bParm1 = CANON_MEDIA_ID_POSTCARD; break;
         case DeviceMedia::MEDIA_THICK: bParm1 = CANON_MEDIA_ID_THICK; break;
////////////case DeviceMedia::: bParm1 = CANON_MEDIA_ID_FULLXXXXX; break;
////////////case DeviceMedia::: bParm1 = CANON_MEDIA_ID_BANNER; break;
         case DeviceMedia::MEDIA_OTHER: bParm1 = CANON_MEDIA_ID_OTHER; break;
         }

         if (300 > pDR->getYRes ())
         {
            bParm2 = 0;
         }
         else if (600 > pDR->getYRes ())
         {
            bParm2 = 0x03;
         }
         else
         {
            bParm2 = 0x04;
         }

         pDevice_d->sendPrintfToDevice (pCmd,
                                        bParm1,
                                        bParm2);
      }
      // search for: <deviceData name="cmdSetColor3" type="binary">_ESC_ "(c" HEX2S(00,02) HEX(10) HEX(00)</deviceData>
      else if (pData->getBinaryData ("cmdSetColor3", &pCmd))
      {
         pDevice_d->sendBinaryDataToDevice (pCmd);
      }
   }
}

bool Canon_BJC_8500_Instance::
beginJob ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

   if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::beginJob ()" << endl;

   pCmd = pCommands->getCommandData ("cmdInit");
   if (pCmd)
   {
      if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::beginJob () cmdInit = " << *pCmd << endl;

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }
   else
   {
      if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::beginJob () Error: There is no cmdInit defined for this device!" << endl;
   }

   // @TBD
   // Japanese models cannot receive "Compression Mode" command after control
   // commands.

   return true;
}

bool Canon_BJC_8500_Instance::
beginJob (bool fJobPropertiesChanged)
{
   if (DebugOutput::shouldOutputInstance ()) cerr << "Canon_BJC_8500_Instance::beginJob (with props)" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The job properties are as follows:" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The orientation is = " << getCurrentOrientation() << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "iDitherID = " << getCurrentDitherID () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The form is = " << getCurrentForm () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The tray is = " << getCurrentTray () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The media is = " << getCurrentMedia () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The resolution is = " << getCurrentResolution () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The print mode is = " << getCurrentPrintMode () << endl;

   // @TBD - reinitialize with new job properties

   // Call common code
   return beginJob ();
}

bool Canon_BJC_8500_Instance::
newFrame ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

   if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::newFrame ()" << endl;

   pCmd = pCommands->getCommandData ("cmdPageEject");
   if (pCmd)
   {
      if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::newFrame () cmdPageEject = " << *pCmd << endl;

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   ditherNewFrame ();

   ptlPrintHead_d.x = 0;
   ptlPrintHead_d.y = 0;

   return true;
}

bool Canon_BJC_8500_Instance::
newFrame (bool fJobPropertiesChanged)
{
   if (DebugOutput::shouldOutputInstance ()) cerr << "Canon_BJC_8500_Instance::newFrame (with props)" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The job properties are as follows:" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The orientation is = " << getCurrentOrientation() << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "iDitherID = " << getCurrentDitherID () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The form is = " << getCurrentForm () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The tray is = " << getCurrentTray () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The media is = " << getCurrentMedia () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The resolution is = " << getCurrentResolution () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The print mode is = " << getCurrentPrintMode () << endl;

   // @TBD - reinitialize with new job properties

   // Call common code
   return newFrame ();
}

bool Canon_BJC_8500_Instance::
endJob ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

   if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::endJob ()" << endl;

   pCmd = pCommands->getCommandData ("cmdPageEject");
   if (pCmd)
   {
      if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::endJob () cmdPageEject = " << *pCmd << endl;

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   pCmd = pCommands->getCommandData ("cmdTerm");
   if (pCmd)
   {
      if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::endJob () cmdTerm = " << *pCmd << endl;

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   return true;
}

bool Canon_BJC_8500_Instance::
abortJob ()
{
   if (DebugOutput::shouldOutputBlitter ()) cerr << "Canon_BJC_8500_Instance::endJob ()" << endl;

   int cx = getCurrentForm ()->getHardCopyCap ()->getXPels ();

   // The largest size is 24 lines
   int iSize = 24 * (cx + 7) / 8;

   PBYTE abData = new BYTE [iSize];

   memset (abData, 0, sizeof (abData));

   BinaryData data (abData, sizeof (abData));

   pDevice_d->sendBinaryDataToDevice (&data);

   delete []abData;

   return true;
}

string Canon_BJC_8500_Instance::
toString (ostrstream& oss)
{
   oss << hex
       << "{ "
       << DeviceInstance::toString (oss) // Add parent's output as well
       << " }" << ends;

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const Canon_BJC_8500_Instance& const_self)
{
   Canon_BJC_8500_Instance& self = const_cast<Canon_BJC_8500_Instance&>(const_self);
   ostrstream          oss;

   os << self.toString (oss);

   return os;
}
