/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "DevicePrintMode.hpp"
#include "StringResource.hpp"
#include "DebugOutput.hpp"
#include <stdio.h>
#include <string.h>

DevicePrintMode::
DevicePrintMode (int id,
                 int iPhysicalCount,
                 int iLogicalCount,
                 int iPlanes)
{
   id_d             = id;
   iPhysicalCount_d = iPhysicalCount;
   iLogicalCount_d  = iLogicalCount;
   iPlanes_d        = iPlanes;
}

char * DevicePrintMode::
getName (int id)
{
   return StringResource::getString (StringResource::STRINGGROUP_PRINTMODES, id);
}

char * DevicePrintMode::
getName ()
{
   return StringResource::getString (StringResource::STRINGGROUP_PRINTMODES, id_d);
}

bool DevicePrintMode::
isID (int id)
{
   return id == id_d;
}

int DevicePrintMode::
getID ()
{
   return id_d;
}

int DevicePrintMode::
getColorTech ()
{
   switch (id_d)
   {
   case PRINT_MODE_UNLISTED:       return -1;
   case PRINT_MODE_NONE:           return -1;
   case PRINT_MODE_1_ANY:          return COLOR_TECH_K;
   case PRINT_MODE_8_K:            return COLOR_TECH_K;
   case PRINT_MODE_8_CMY:          return COLOR_TECH_CMY;
   case PRINT_MODE_8_CMYK:         return COLOR_TECH_CMYK;
   case PRINT_MODE_8_CcMmYK:       return COLOR_TECH_CcMmYK;
   case PRINT_MODE_8_CcMmYyK:      return COLOR_TECH_CcMmYyK;
   case PRINT_MODE_8_RGB:          return COLOR_TECH_RGB;
   case PRINT_MODE_24_K:           return COLOR_TECH_K;
   case PRINT_MODE_24_CMY:         return COLOR_TECH_CMY;
   case PRINT_MODE_24_CMYK:        return COLOR_TECH_CMYK;
   case PRINT_MODE_24_CcMmYK:      return COLOR_TECH_CcMmYK;
   case PRINT_MODE_24_CcMmYyK:     return COLOR_TECH_CcMmYyK;
   case PRINT_MODE_24_RGB:         return COLOR_TECH_RGB;
   default:
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputDevicePrintMode ())
         cerr << "DevicePrintMode::" << __FUNCTION__ << ": Error: getColorTech: unhandled id_d = " << dec << id_d << endl;
#endif
      return -1;
   }
   }
}

int DevicePrintMode::
getPhysicalCount ()
{
   return iPhysicalCount_d;
}

int DevicePrintMode::
getLogicalCount ()
{
   return iLogicalCount_d;
}

int DevicePrintMode::
getNumPlanes ()
{
   return iPlanes_d;
}

DevicePrintMode * DevicePrintMode::
create (char *pszId)
{
   return this->createV (nameToID (pszId));
}

#define STRING_OF(x) (apszPrintModeNames[x - PRINT_MODE_UNLISTED])

static char *apszPrintModeNames[] = {
   "PRINT_MODE_UNLISTED",
   "PRINT_MODE_NONE",
   "PRINT_MODE_1_ANY",
   "PRINT_MODE_8_K",
   "PRINT_MODE_8_CMY",
   "PRINT_MODE_8_CMYK",
   "PRINT_MODE_8_CcMmYK",
   "PRINT_MODE_8_CcMmYyK",
   "PRINT_MODE_8_RGB",
   "PRINT_MODE_24_K",
   "PRINT_MODE_24_CMY",
   "PRINT_MODE_24_CMYK",
   "PRINT_MODE_24_CcMmYK",
   "PRINT_MODE_24_CcMmYyK",
   "PRINT_MODE_24_RGB"
};

int DevicePrintMode::
nameToID (char *pszId)
{
   static short int asiEntries[] = {
      PRINT_MODE_1_ANY,
      PRINT_MODE_24_CMY,
      PRINT_MODE_24_CMYK,
      PRINT_MODE_24_CcMmYK,
      PRINT_MODE_24_CcMmYyK,
      PRINT_MODE_24_K,
      PRINT_MODE_24_RGB,
      PRINT_MODE_8_CMY,
      PRINT_MODE_8_CMYK,
      PRINT_MODE_8_CcMmYK,
      PRINT_MODE_8_CcMmYyK,
      PRINT_MODE_8_K,
      PRINT_MODE_8_RGB,
      PRINT_MODE_NONE
   };

   int iLow  = 0;
   int iMid  = (int)dimof (asiEntries) / 2;
   int iHigh = (int)dimof (asiEntries) - 1;
   int iResult;

   while (iLow <= iHigh)
   {
      iResult = strcmp (pszId, STRING_OF (asiEntries[iMid]));

      if (0 == iResult)
      {
         return asiEntries[iMid];
      }
      else if (0 > iResult)
      {
         // str1 < str2
         iHigh = iMid - 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
      else // (0 < iResult)
      {
         // str1 > str2
         iLow  = iMid + 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
   }

   return PRINT_MODE_UNLISTED;
}

char * DevicePrintMode::
IDToName (int id)
{
   id -= PRINT_MODE_UNLISTED;

   if (  0 <= id
      && id < (int)dimof (apszPrintModeNames)
      )
      return apszPrintModeNames[id];
   else
   {
      static char achUnknown[21];

      sprintf (achUnknown, "Unknown (%d)", id + PRINT_MODE_UNLISTED);

      return achUnknown;
   }
}

string DevicePrintMode::
toString (ostrstream& oss)
{
   oss << dec
       << "{"
       << STRING_OF (id_d)
       << ", iPhysicalCount_d = " << iPhysicalCount_d
       << ", iLogicalCount_d = " << iLogicalCount_d
       << ", iPlanes_d = " << iPlanes_d
       << "}"
       << ends;

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const DevicePrintMode& const_self)
{
   DevicePrintMode& self = const_cast<DevicePrintMode&>(const_self);
   ostrstream       oss;

   os << self.toString (oss);

   return os;
}
