/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "OkiPos_Instance.hpp"
#include "defines.hpp"

#include <iostream>
#include <strstream>

const static bool fTestNoCompression = true;

OkiPos_Instance::
OkiPos_Instance (Device *pDevice)
   : DeviceInstance (pDevice)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "OkiPos_Instance::OkiPos_Instance ()" << endl;
#endif

   fHaveInitialized_d  = false;
   fHaveSetupPrinter_d = false;
   fUseJournal_d       = false;
   eCutMode_d          = CUT_MODE_NONE;
   ptlPrintHead_d.x    = 0;
   ptlPrintHead_d.y    = 0;
}

OkiPos_Instance::
~OkiPos_Instance ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "OkiPos_Instance::~OkiPos_Instance ()" << endl;
#endif
}

void OkiPos_Instance::
initializeInstance ()
{
   if (fHaveInitialized_d)
      return;

   fHaveInitialized_d = true;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "OkiPos_Instance::initializeInstance ()" << endl;
#endif
}

static char *apszDeviceJobPropertyKeysD[] = {
   "journal",
   "cutmode"
};
static char *apszDeviceJobPropertyKeysS[] = {
   "cutmode"
};
#define JOBPROP_USEJOURNAL apszDeviceJobPropertyKeysD[0]
#define JOBPROP_USECUTMODE apszDeviceJobPropertyKeysD[1]

Enumeration * OkiPos_Instance::
listDeviceJobPropertyKeys ()
{
   class DeviceJobPropertyKeyEnumerator : public Enumeration
   {
   public:
      DeviceJobPropertyKeyEnumerator (char **apszKeys, int iDimOfKeys)
      {
         iWhichKey_d = iDimOfKeys;
         apszKeys_d  = apszKeys;
      }

      virtual bool hasMoreElements ()
      {
         return iWhichKey_d > 0;
      }

      virtual void * nextElement ()
      {
         if (0 < iWhichKey_d)
         {
            iWhichKey_d--;
            return apszKeys_d[iWhichKey_d];
         }

         return 0;
      }

   private:
      int    iWhichKey_d;
      char **apszKeys_d;
   };

   if (hasDeviceOption (MODEL_S))
   {
      return new DeviceJobPropertyKeyEnumerator (apszDeviceJobPropertyKeysS,
                                                 dimof (apszDeviceJobPropertyKeysS));
   }
   else if (hasDeviceOption (MODEL_D))
   {
      return new DeviceJobPropertyKeyEnumerator (apszDeviceJobPropertyKeysD,
                                                 dimof (apszDeviceJobPropertyKeysD));
   }
   else
   {
      return DeviceInstance::listDeviceJobPropertyKeys ();
   }
}

bool OkiPos_Instance::
handleDeviceJobProperty (char *pszKey,
                         char *pszValue)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "OkiPos_Instance::handleDeviceJobProperty ("
                          << pszKey << ", "
                          << pszValue << ")"
                          << endl;
#endif

   if (0 == strcasecmp (pszKey, JOBPROP_USEJOURNAL))
   {
      if (0 == strcmp (pszValue, "true"))
      {
         fUseJournal_d = true;
         return true;
      }
      else if (0 == strcmp (pszValue, "false"))
      {
         fUseJournal_d = false;
         return true;
      }
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_USECUTMODE))
   {
      if (0 == strcmp (pszValue, "none"))
      {
         eCutMode_d = CUT_MODE_NONE;
         return true;
      }
      else if (0 == strcmp (pszValue, "full"))
      {
         eCutMode_d = CUT_MODE_FULL;
         return true;
      }
      else if (0 == strcmp (pszValue, "partial"))
      {
         eCutMode_d = CUT_MODE_PARTIAL;
         return true;
      }
   }

   return false;
}

string * OkiPos_Instance::
getJobPropertyType (char *pszKey)
{
   if (0 == strcasecmp (pszKey, JOBPROP_USEJOURNAL))
   {
      return new string ("boolean false");
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_USECUTMODE))
   {
      return new string ("string none");
   }

   return 0;
}

string * OkiPos_Instance::
getJobProperty (char *pszKey)
{
   if (0 == strcasecmp (pszKey, JOBPROP_USEJOURNAL))
   {
      if (fUseJournal_d)
      {
         return new string ("true");
      }
      else
      {
         return new string ("false");
      }
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_USECUTMODE))
   {
      switch (eCutMode_d)
      {
      case CUT_MODE_NONE:    return new string ("none");
      case CUT_MODE_FULL:    return new string ("full");
      case CUT_MODE_PARTIAL: return new string ("partial");
      }
   }

   return 0;
}

void OkiPos_Instance::
setupPrinter ()
{
   if (fHaveSetupPrinter_d)
      return;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "OkiPos_Instance::setupPrinter ()" << endl;
#endif

   fHaveSetupPrinter_d = true;

   DeviceResolution *pRes      = getCurrentResolution ();
   DeviceForm       *pDF       = getCurrentForm ();
   HardCopyCap      *pHCC      = pDF->getHardCopyCap ();
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

   // @TBD
///pDevice_d->sendBinaryDataToDevice (getCurrentTray ());

   pCmd = pCommands->getCommandData ("cmdSetLineSpacing144inch");
   if (pCmd)
   {
      int iSpacing = 144 * pRes->getScanlineMultiple () / pRes->getYRes ();

      pDevice_d->sendPrintfToDevice (pCmd, iSpacing);
   }

   int iLines = (int)(6.0 * (float)pHCC->getCy () / 25400.0);

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "iLines = " << iLines << endl;
#endif

   pCmd = pCommands->getCommandData ("cmdSetPageLengthInLines");
   if (pCmd)
   {
      pDevice_d->sendPrintfToDevice (pCmd, iLines);
   }

   if (fUseJournal_d)
   {
      pCmd = pCommands->getCommandData ("cmdSetJournalMode");
      if (pCmd)
      {
         pDevice_d->sendPrintfToDevice (pCmd, 1);
      }
   }

   cerr << "******* " << pHCC->getXPels () << endl;
}

bool OkiPos_Instance::
beginJob ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "OkiPos_Instance::beginJob ()" << endl;
#endif

   pCmd = pCommands->getCommandData ("cmdInit");
   if (pCmd)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdInit = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   return true;
}

bool OkiPos_Instance::
beginJob (bool fJobPropertiesChanged)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "OkiPos_Instance::beginJob (with props)" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The job properties are as follows:" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The orientation is = " << getCurrentOrientation() << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "iDitherID = " << getCurrentDitherID () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The form is = " << getCurrentForm () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The tray is = " << getCurrentTray () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The media is = " << getCurrentMedia () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The resolution is = " << getCurrentResolution () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The print mode is = " << getCurrentPrintMode () << endl;
#endif

   // @TBD - reinitialize with new job properties

   // Call common code
   return beginJob ();
}

bool OkiPos_Instance::
newFrame ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "OkiPos_Instance::newFrame ()" << endl;
#endif

   pCmd = pCommands->getCommandData ("cmdPageEject");
   if (pCmd)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdPageEject = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   ditherNewFrame ();

   ptlPrintHead_d.x = 0;
   ptlPrintHead_d.y = 0;

   return true;
}

bool OkiPos_Instance::
newFrame (bool fJobPropertiesChanged)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "OkiPos_Instance::newFrame (with props)" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The job properties are as follows:" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The orientation is = " << getCurrentOrientation() << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "iDitherID = " << getCurrentDitherID () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The form is = " << getCurrentForm () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The tray is = " << getCurrentTray () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The media is = " << getCurrentMedia () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The resolution is = " << getCurrentResolution () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The print mode is = " << getCurrentPrintMode () << endl;
#endif

   // @TBD - reinitialize with new job properties

   // Call common code
   return newFrame ();
}

bool OkiPos_Instance::
endJob ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "OkiPos_Instance::endJob ()" << endl;
#endif

   pCmd = pCommands->getCommandData ("cmdPageEject");
   if (pCmd)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdPageEject = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   // If the user wants to cut the receipt, then advance the roll and cut.
   pCmd = 0;
   switch (eCutMode_d)
   {
   case CUT_MODE_FULL:    pCmd = pCommands->getCommandData ("cmdFullCut"); break;
   case CUT_MODE_PARTIAL: pCmd = pCommands->getCommandData ("cmdPartialCut"); break;
   }
   if (pCmd)
   {
      DeviceForm  *pDF       = getCurrentForm ();
      HardCopyCap *pHCC      = pDF->getHardCopyCap ();
      int          iLength   = 0;
      BinaryData  *pCmdSpace = 0;

#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdXXXCut = " << *pCmd << endl;
#endif

      iLength = (pHCC->getTopClip () + pHCC->getBottomClip ()) * 144 / 25400;

      cerr << "iLength = " << iLength << endl;

      pCmdSpace = pCommands->getCommandData ("cmdSetLineSpacing144inch");
      if (pCmdSpace)
      {
         pDevice_d->sendPrintfToDevice (pCmdSpace, iLength);
      }
      pCmdSpace = pCommands->getCommandData ("cmdMoveToNextRasterGraphicsLine");
      if (pCmdSpace)
      {
         pDevice_d->sendBinaryDataToDevice (pCmdSpace);
      }
      pCmdSpace = pCommands->getCommandData ("cmdEndRasterGraphicsLine");
      if (pCmdSpace)
      {
         pDevice_d->sendBinaryDataToDevice (pCmdSpace);
      }

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   pCmd = pCommands->getCommandData ("cmdTerm");
   if (pCmd)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdTerm = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   return true;
}

bool OkiPos_Instance::
abortJob ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "OkiPos_Instance::endJob ()" << endl;
#endif

   int cx = getCurrentForm ()->getHardCopyCap ()->getXPels ();

   // The largest size is 24 lines
   int iSize = 24 * (cx + 7) / 8;

   PBYTE abData = new BYTE [iSize];

   memset (abData, 0, sizeof (abData));

   BinaryData data (abData, sizeof (abData));

   pDevice_d->sendBinaryDataToDevice (&data);

   delete[] abData;

   return true;
}

string OkiPos_Instance::
toString (ostrstream& oss)
{
   oss << "{ "
       << DeviceInstance::toString (oss) // Add parent's output as well
       << " }"
       << ends;

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const OkiPos_Instance& const_self)
{
   OkiPos_Instance& self = const_cast<OkiPos_Instance&>(const_self);
   ostrstream       oss;

   os << self.toString (oss);

   return os;
}
