/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Idea borrowed from http://hpinkjet.sourceforge.net
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/shm.h>
#include <errno.h>

#include "OmniIJSProxy.hpp"

static char *vpszServerToClient = "IJS_SRV_TO_CLIENT";
static char *vpszClientToServer = "IJS_CLIENT_TO_SRV";

char *
setupName (char *pszName, char *pszFormat)
{
#define ENVIRONMENT_SIZE 64
   char   *pszRet = 0;
   size_t  rc;

   pszRet = (char *)malloc (ENVIRONMENT_SIZE);

   if (!pszRet)
      return 0;

   rc = sprintf (pszRet, pszFormat, pszName, getpid ());
   if (rc > ENVIRONMENT_SIZE)
   {
      exit (EXIT_FAILURE);
   }

   if (-1 == putenv (pszRet))
      return 0;

   return pszRet;
}

OmniIJSProxy::
OmniIJSProxy (char *pszDeviceName,
              char *pszJobProperties,
              bool  fAdvanced,
              int   fdStdOut,
              int   fdStdErr)
{
   char        *pszS2C            = 0;
   char        *pszC2S            = 0;
   bool         fSucceeded        = false;
   int          pid;

   // Assign fifo names for environment space
   pszS2C_d        = setupName (vpszServerToClient, "%s=/tmp/ijs_s2c_%d");
   pszC2S_d        = setupName (vpszClientToServer, "%s=/tmp/ijs_c2s_%d");
   pszS2C          = getenv (vpszServerToClient);
   pszC2S          = getenv (vpszClientToServer);

   fAdvanced_d     = fAdvanced;

   fdS2C_d         = -1;
   fdC2S_d         = -1;

   idBuffer1_d     = -1;
   cbBuffer1_d     = 0;
   pbBuffer1_d     = 0;
   idBuffer2_d     = -1;
   cbBuffer2_d     = 0;
   pbBuffer2_d     = 0;

   pCmd_d          = new PrinterCommand ("client");

   pszVersion_d    = 0;
   pszDriverName_d = 0;
   pszDeviceName_d = 0;
   pszShortName_d  = 0;

   pDO_d           = 0;
   pszDitherID_d   = 0;
   pDF_d           = 0;
   pDT_d           = 0;
   pDM_d           = 0;
   pDR_d           = 0;
   pDPM_d          = 0;

   if (  !pszS2C_d
      || !pszC2S_d
      || !pCmd_d
      )
   {
      goto BUGOUT;
   }

   /* Create actual fifos in filesystem. */
   if ((mkfifo (pszS2C, 0666)) < 0)
   {
      goto BUGOUT;
   }

   if ((mkfifo (pszC2S, 0666)) < 0)
   {
      goto BUGOUT;
   }

   /* Spawn the server. */
   if ((pid = fork ()) < 0)
   {
      goto BUGOUT;
   }
   else if (pid > 0)
   {
      /* parent */
      /* Open fifos. */
      if ((fdS2C_d = open (pszS2C, O_RDONLY)) < 0)
      {
         goto BUGOUT;
      }
      if ((fdC2S_d = open (pszC2S, O_WRONLY)) < 0)
      {
         goto BUGOUT;
      }
   }
   else
   {
      /* child */
      /* Redirect print_stream to stdout. */
      if (fdStdOut != STDOUT_FILENO)
      {
         dup2 (fdStdOut, STDOUT_FILENO);
      }
      if (fdStdErr != STDERR_FILENO)
      {
         dup2 (fdStdErr, STDERR_FILENO);
      }

      if (execlp ("omni", "omni", (char *)0, (char *)0) < 0)
      {
         cerr << "client: execlp failed." << endl;

	      goto BUGOUT2;
      }

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniIJSProxy ()) cerr << "client: execlp succeded." << endl;
#endif

      exit (0);

BUGOUT2:
      exit (EXIT_FAILURE);
   }

   // Tell the server what the device name will be
   if (  !pCmd_d->setCommand (COMMAND_TYPE_SET_DEVICE_NAME, pszDeviceName)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || COMMAND_TYPE_ACK != pCmd_d->getCommandType ()
      )
      goto BUGOUT;

   if (fAdvanced_d)
   {
      // Tell the server to used advanced bitblt mode.  This is optional
      pCmd_d->setCommand (COMMAND_TYPE_OMNI_SET_ADVANCED, 1);
      pCmd_d->sendCommand (fdC2S_d);
      pCmd_d->readCommand (fdS2C_d);
   }

   if (  pszJobProperties
      && *pszJobProperties
      )
   {
      // If there are job properties, pass them in
      if (  !pCmd_d->setCommand (COMMAND_TYPE_SET_JOB_PROPERTIES, pszJobProperties)
         || !pCmd_d->sendCommand (fdC2S_d)
         || !pCmd_d->readCommand (fdS2C_d)
         || COMMAND_TYPE_ACK != pCmd_d->getCommandType ()
         )
         goto BUGOUT;
   }

   // Tell the server to create a device instance
   if (  !pCmd_d->setCommand (COMMAND_TYPE_NEW_DEVICE, (char *)0)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || COMMAND_TYPE_ACK != pCmd_d->getCommandType ()
      )
      goto BUGOUT;

   fSucceeded = true;

BUGOUT:
   if (!fSucceeded)
   {
      // @TBD throw error?
      return;
   }
}

OmniIJSProxy::
~OmniIJSProxy ()
{
   char *pszS2C = 0;
   char *pszC2S = 0;
   int   rc;

   // End the connection
   if (pCmd_d->setCommand (COMMAND_TYPE_TERMINATE, (char *)0))
   {
      pCmd_d->sendCommand (fdC2S_d);
   }

   pszS2C = getenv (vpszServerToClient);
   pszC2S = getenv (vpszClientToServer);

   // Clean up
   rc = close (fdS2C_d);
   rc = close (fdC2S_d);

   rc = remove (pszS2C);
   if (-1 == rc)
   {
      cerr << "remove (" << pszS2C << ") = " << errno << endl;
   }
   rc = remove (pszC2S);
   if (-1 == rc)
   {
      cerr << "remove (" << pszC2S << ") = " << errno << endl;
   }

   if (pszS2C_d)
   {
      free (pszS2C_d);
   }
   if (pszC2S_d)
   {
      free (pszC2S_d);
   }

   if (pbBuffer1_d)
   {
      shmdt (pbBuffer1_d);
      shmctl (idBuffer1_d, IPC_RMID, 0);
   }
   if (pbBuffer2_d)
   {
      shmdt (pbBuffer2_d);
      shmctl (idBuffer2_d, IPC_RMID, 0);
   }

   delete pCmd_d;

   if (pszVersion_d)
   {
      free (pszVersion_d);
   }
   if (pszDriverName_d)
   {
      free (pszDriverName_d);
   }
   if (pszDeviceName_d)
   {
      free (pszDeviceName_d);
   }
   if (pszShortName_d)
   {
      free (pszShortName_d);
   }

   delete pDO_d;
   if (pszDitherID_d)
   {
      free (pszDitherID_d);
   }
   delete pDF_d;
   delete pDT_d;
   delete pDM_d;
   delete pDR_d;
   delete pDPM_d;
}

char * OmniIJSProxy::
getVersion ()
{
   if (!pszVersion_d)
   {
      // Query the orientation of that device
      if (  pCmd_d->setCommand (COMMAND_TYPE_GET_VERSION, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
         )
      {
         char *pszResponse = 0;

         pszResponse = pCmd_d->getCommandString ();

         if (  pszResponse
            && *pszResponse
            )
         {
            pszVersion_d = (char *)malloc (strlen (pszResponse) + 1);

            if (pszVersion_d)
            {
               strcpy (pszVersion_d, pszResponse);
            }
         }
      }
   }

   return pszVersion_d;
}

char * OmniIJSProxy::
getDriverName ()
{
   if (!pszDriverName_d)
   {
      // Query the orientation of that device
      if (  pCmd_d->setCommand (COMMAND_TYPE_GET_DRIVER_NAME, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
         )
      {
         char *pszResponse = 0;

         pszResponse = pCmd_d->getCommandString ();

         if (  pszResponse
            && *pszResponse
            )
         {
            pszDriverName_d = (char *)malloc (strlen (pszResponse) + 1);

            if (pszDriverName_d)
            {
               strcpy (pszDriverName_d, pszResponse);
            }
         }
      }
   }

   return pszDriverName_d;
}

char * OmniIJSProxy::
getDeviceName ()
{
   if (!pszDeviceName_d)
   {
      // Query the orientation of that device
      if (  pCmd_d->setCommand (COMMAND_TYPE_GET_DEVICE_NAME, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
         )
      {
         char *pszResponse = 0;

         pszResponse = pCmd_d->getCommandString ();

         if (  pszResponse
            && *pszResponse
            )
         {
            pszDeviceName_d = (char *)malloc (strlen (pszResponse) + 1);

            if (pszDeviceName_d)
            {
               strcpy (pszDeviceName_d, pszResponse);
            }
         }
      }
   }

   return pszDeviceName_d;
}

char * OmniIJSProxy::
getShortName ()
{
   if (!pszShortName_d)
   {
      // Query the orientation of that device
      if (  pCmd_d->setCommand (COMMAND_TYPE_GET_SHORT_NAME, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
         )
      {
         char *pszResponse = 0;

         pszResponse = pCmd_d->getCommandString ();

         if (  pszResponse
            && *pszResponse
            )
         {
            pszDeviceName_d = (char *)malloc (strlen (pszResponse) + 1);

            if (pszDeviceName_d)
            {
               strcpy (pszDeviceName_d, pszResponse);
            }
         }
      }
   }

   return pszShortName_d;
}

DeviceOrientation * OmniIJSProxy::
getCurrentOrientation ()
{
   class OmniIJSProxyOrientation : public DeviceOrientation
   {
   public:
      OmniIJSProxyOrientation (int               id,
                               PrinterCommand   *pCmd,
                               int               fdC2S,
                               int               fdS2C)
         : DeviceOrientation (id)
      {
         pCmd_d  = pCmd;
         fdC2S_d = fdC2S;
         fdS2C_d = fdS2C;
      }

      bool
      isOrientationSupported (int id)
      {
         if (  pCmd_d->setCommand (COMMAND_TYPE_IS_ORIENTATION_SUPPORTED, id)
            && pCmd_d->sendCommand (fdC2S_d)
            && pCmd_d->readCommand (fdS2C_d)
            && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
            )
            return true;

         return false;
      }

      Enumeration *
      getEnumeration ()
      {
         class OmniIJSOrientationEnumeration : public Enumeration
         {
         public:
            OmniIJSOrientationEnumeration (char             *pszOrientations,
                                           PrinterCommand   *pCmd,
                                           int               fdC2S,
                                           int               fdS2C)
            {
               pszOrientations_d       = 0;
               pszOrientationCurrent_d = 0;
               pCmd_d                  = pCmd;
               fdC2S_d                 = fdC2S;
               fdS2C_d                 = fdS2C;

               if (  pszOrientations
                  && *pszOrientations
                  )
               {
                  pszOrientations_d = (char *)malloc (strlen (pszOrientations) + 1);
                  if (pszOrientations_d)
                  {
                     strcpy (pszOrientations_d, pszOrientations);

                     pszOrientationCurrent_d = pszOrientations_d;
                  }
               }
            }

            virtual
            ~OmniIJSOrientationEnumeration ()
            {
               if (pszOrientations_d)
               {
                  free (pszOrientations_d);
               }
            }

            bool
            hasMoreElements ()
            {
               if (  !pszOrientationCurrent_d
                  || !*pszOrientationCurrent_d
                  )
                  return false;

               return true;
            }

            void *
            nextElement ()
            {
               void *pvRet = 0;

               if (  pszOrientationCurrent_d
                  && *pszOrientationCurrent_d
                  )
               {
                  char *pszSpace;

                  pszSpace = strchr (pszOrientationCurrent_d, ' ');

                  if (pszSpace)
                  {
                     *pszSpace = '\0';
                  }

                  pvRet = create (pszOrientationCurrent_d,
                                  pCmd_d,
                                  fdC2S_d,
                                  fdS2C_d);

                  if (pszSpace)
                     pszOrientationCurrent_d = pszSpace + 1;
                  else
                     pszOrientationCurrent_d += strlen (pszOrientationCurrent_d);
               }

               return pvRet;
            }

         private:
            char             *pszOrientations_d;
            char             *pszOrientationCurrent_d;
            PrinterCommand   *pCmd_d;
            int               fdC2S_d;
            int               fdS2C_d;
         };

         // Enumerate the orientations of that device
         if (  pCmd_d->setCommand (COMMAND_TYPE_ENUM_ORIENTATIONS, (char *)0)
            && pCmd_d->sendCommand (fdC2S_d)
            && pCmd_d->readCommand (fdS2C_d)
            && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
            )
            return new OmniIJSOrientationEnumeration (pCmd_d->getCommandString (),
                                                      pCmd_d,
                                                      fdC2S_d,
                                                      fdS2C_d);

         return new OmniIJSOrientationEnumeration (0,
                                                   pCmd_d,
                                                   fdC2S_d,
                                                   fdS2C_d);
      }

      static DeviceOrientation *
      create (char             *pszName,
              PrinterCommand   *pCmd,
              int               fdC2S,
              int               fdS2C)
      {
         // Query the orientation of that device
         if (  pCmd->setCommand (COMMAND_TYPE_QUERY_ORIENTATION, pszName)
            && pCmd->sendCommand (fdC2S)
            && pCmd->readCommand (fdS2C)
            && COMMAND_TYPE_ACK == pCmd->getCommandType ()
            )
         {
            char *pszResponse = 0;

            pszResponse = pCmd->getCommandString ();

#ifndef RETAIL
            if (DebugOutput::shouldOutputOmniIJSProxy ())
            {
               cerr << __FUNCTION__ << ": id            = " << pszResponse << endl;
            }
#endif

            DeviceOrientation *pDO;

            pDO = new OmniIJSProxyOrientation (DeviceOrientation::nameToID (pszResponse),
                                               pCmd,
                                               fdC2S,
                                               fdS2C);

            return pDO;
         }

         return 0;
      }

   protected:
      DeviceOrientation *
      createV (int id)
      {
         return 0;
      }

   private:
      PrinterCommand   *pCmd_d;
      int               fdC2S_d;
      int               fdS2C_d;
   };

   if (pDO_d)
      return pDO_d;

   pDO_d = OmniIJSProxyOrientation::create ((char *)0,
                                            pCmd_d,
                                            fdC2S_d,
                                            fdS2C_d);

   return pDO_d;
}

char * OmniIJSProxy::
getCurrentDitherID ()
{
   if (!pszDitherID_d)
   {
      // Query the orientation of that device
      if (  pCmd_d->setCommand (COMMAND_TYPE_QUERY_DITHER_ID, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
         )
      {
         char *pszResponse = 0;

         pszResponse = pCmd_d->getCommandString ();

         if (  pszResponse
            && *pszResponse
            )
         {
            pszDitherID_d = (char *)malloc (strlen (pszResponse) + 1);

            if (pszDitherID_d)
            {
               strcpy (pszDitherID_d, pszResponse);
            }
         }
      }
   }

   return pszDitherID_d;
}

DeviceForm * OmniIJSProxy::
getCurrentForm ()
{
   class OmniIJSProxyForm : public DeviceForm
   {
   public:
      OmniIJSProxyForm (int               id,
                        int               iCapabilities,
                        BinaryData       *data,
                        HardCopyCap      *hcInfo,
                        DeviceResolution *pDR,
                        PrinterCommand   *pCmd,
                        int               fdC2S,
                        int               fdS2C)
         : DeviceForm (id,
                       iCapabilities,
                       data,
                       hcInfo)
      {
         pDR_d   = pDR;
         pCmd_d  = pCmd;
         fdC2S_d = fdC2S;
         fdS2C_d = fdS2C;
      }

      bool
      isFormSupported (int id)
      {
         if (  pCmd_d->setCommand (COMMAND_TYPE_IS_FORM_SUPPORTED, id)
            && pCmd_d->sendCommand (fdC2S_d)
            && pCmd_d->readCommand (fdS2C_d)
            && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
            )
            return true;

         return false;
      }

      Enumeration *
      getEnumeration ()
      {
         class OmniIJSFormEnumeration : public Enumeration
         {
         public:
            OmniIJSFormEnumeration (char             *pszForms,
                                    DeviceResolution *pDR,
                                    PrinterCommand   *pCmd,
                                    int               fdC2S,
                                    int               fdS2C)
            {
               pszForms_d       = 0;
               pszFormCurrent_d = 0;
               pDR_d            = pDR;
               pCmd_d           = pCmd;
               fdC2S_d          = fdC2S;
               fdS2C_d          = fdS2C;

               if (  pszForms
                  && *pszForms
                  )
               {
                  pszForms_d = (char *)malloc (strlen (pszForms) + 1);
                  if (pszForms_d)
                  {
                     strcpy (pszForms_d, pszForms);

                     pszFormCurrent_d = pszForms_d;
                  }
               }
            }

            virtual
            ~OmniIJSFormEnumeration ()
            {
               if (pszForms_d)
               {
                  free (pszForms_d);
               }
            }

            bool
            hasMoreElements ()
            {
               if (  !pszFormCurrent_d
                  || !*pszFormCurrent_d
                  )
                  return false;

               return true;
            }

            void *
            nextElement ()
            {
               void *pvRet = 0;

               if (  pszFormCurrent_d
                  && *pszFormCurrent_d
                  )
               {
                  char *pszSpace;

                  pszSpace = strchr (pszFormCurrent_d, ' ');

                  if (pszSpace)
                  {
                     *pszSpace = '\0';
                  }

                  pvRet = create (pszFormCurrent_d,
                                  pDR_d,
                                  pCmd_d,
                                  fdC2S_d,
                                  fdS2C_d);

                  if (pszSpace)
                     pszFormCurrent_d = pszSpace + 1;
                  else
                     pszFormCurrent_d += strlen (pszFormCurrent_d);
               }

               return pvRet;
            }

         private:
            char             *pszForms_d;
            char             *pszFormCurrent_d;
            DeviceResolution *pDR_d;
            PrinterCommand   *pCmd_d;
            int               fdC2S_d;
            int               fdS2C_d;
         };

         // Enumerate the forms of that device
         if (  pCmd_d->setCommand (COMMAND_TYPE_ENUM_FORMS, (char *)0)
            && pCmd_d->sendCommand (fdC2S_d)
            && pCmd_d->readCommand (fdS2C_d)
            && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
            )
            return new OmniIJSFormEnumeration (pCmd_d->getCommandString (),
                                               pDR_d,
                                               pCmd_d,
                                               fdC2S_d,
                                               fdS2C_d);

         return new OmniIJSFormEnumeration (0,
                                            pDR_d,
                                            pCmd_d,
                                            fdC2S_d,
                                            fdS2C_d);
      }

      static DeviceForm *
      create (char             *pszName,
              DeviceResolution *pDR,
              PrinterCommand   *pCmd,
              int               fdC2S,
              int               fdS2C)
      {
         // Query the form of that device
         if (  pCmd->setCommand (COMMAND_TYPE_QUERY_FORM, pszName)
            && pCmd->sendCommand (fdC2S)
            && pCmd->readCommand (fdS2C)
            && COMMAND_TYPE_ACK == pCmd->getCommandType ()
            )
         {
            char *pszResponse   = 0;
            char *pszID         = 0;
            int   iCapabilities = 0;
            int   iLeftClip     = 0;
            int   iTopClip      = 0;
            int   iRightClip    = 0;
            int   iBottomClip   = 0;

            pszID = pCmd->getCommandString ();

            pszResponse = strchr (pszID, ' ');

            if (!pszResponse)
               return 0;

            *pszResponse++ = '\0';

            sscanf (pszResponse,
                    "%d %d %d %d %d",
                    &iCapabilities,
                    &iLeftClip,
                    &iTopClip,
                    &iRightClip,
                    &iBottomClip);

#ifndef RETAIL
            if (DebugOutput::shouldOutputOmniIJSProxy ())
            {
               cerr << __FUNCTION__ << ": id            = " << pszID         << endl;
               cerr << __FUNCTION__ << ": iCapabilities = " << iCapabilities << endl;
               cerr << __FUNCTION__ << ": iLeftClip     = " << iLeftClip     << endl;
               cerr << __FUNCTION__ << ": iTopClip      = " << iTopClip      << endl;
               cerr << __FUNCTION__ << ": iRightClip    = " << iRightClip    << endl;
               cerr << __FUNCTION__ << ": iBottomClip   = " << iBottomClip   << endl;
            }
#endif

            DeviceForm *pDF;
            int         id  = DeviceForm::nameToID (pszID);

            pDF = new OmniIJSProxyForm (id,
                                        iCapabilities,
                                        0,
                                        new HardCopyCap (iLeftClip,
                                                         iTopClip,
                                                         iRightClip,
                                                         iBottomClip),
                                        pDR,
                                        pCmd,
                                        fdC2S,
                                        fdS2C);

            if (pDF)
            {
               pDF->associateWith (pDR);
            }

            return pDF;
         }

         return 0;
      }

   protected:
      DeviceForm *
      createV (int id)
      {
         return 0;
      }

   private:
      DeviceResolution *pDR_d;
      PrinterCommand   *pCmd_d;
      int               fdC2S_d;
      int               fdS2C_d;
   };

   if (pDF_d)
      return pDF_d;

   pDF_d = OmniIJSProxyForm::create ((char *)0,
                                     getCurrentResolution (),
                                     pCmd_d,
                                     fdC2S_d,
                                     fdS2C_d);

   return pDF_d;
}

DeviceTray * OmniIJSProxy::
getCurrentTray ()
{
   class OmniIJSProxyTray : public DeviceTray
   {
   public:
      OmniIJSProxyTray (int             id,
                        int             iType,
                        BinaryData     *data,
                        PrinterCommand *pCmd,
                        int             fdC2S,
                        int             fdS2C)
         : DeviceTray (id,
                       iType,
                       data)
      {
         pCmd_d  = pCmd;
         fdC2S_d = fdC2S;
         fdS2C_d = fdS2C;
      }

      bool
      isTraySupported (int id)
      {
         if (  pCmd_d->setCommand (COMMAND_TYPE_IS_TRAY_SUPPORTED, id)
            && pCmd_d->sendCommand (fdC2S_d)
            && pCmd_d->readCommand (fdS2C_d)
            && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
            )
            return true;

         return false;
      }

      Enumeration *
      getEnumeration ()
      {
         class OmniIJSTrayEnumeration : public Enumeration
         {
         public:
            OmniIJSTrayEnumeration (char           *pszTrays,
                                    PrinterCommand *pCmd,
                                    int             fdC2S,
                                    int             fdS2C)
            {
               pszTrays_d       = 0;
               pszTrayCurrent_d = 0;
               pCmd_d           = pCmd;
               fdC2S_d          = fdC2S;
               fdS2C_d          = fdS2C;

               if (  pszTrays
                  && *pszTrays
                  )
               {
                  pszTrays_d = (char *)malloc (strlen (pszTrays) + 1);
                  if (pszTrays_d)
                  {
                     strcpy (pszTrays_d, pszTrays);

                     pszTrayCurrent_d = pszTrays_d;
                  }
               }
            }

            virtual
            ~OmniIJSTrayEnumeration ()
            {
               if (pszTrays_d)
               {
                  free (pszTrays_d);
               }
            }

            bool
            hasMoreElements ()
            {
               if (  !pszTrayCurrent_d
                  || !*pszTrayCurrent_d
                  )
                  return false;

               return true;
            }

            void *
            nextElement ()
            {
               void *pvRet = 0;

               if (  pszTrayCurrent_d
                  && *pszTrayCurrent_d
                  )
               {
                  char *pszSpace;

                  pszSpace = strchr (pszTrayCurrent_d, ' ');

                  if (pszSpace)
                  {
                     *pszSpace = '\0';
                  }

                  pvRet = create (pszTrayCurrent_d,
                                  pCmd_d,
                                  fdC2S_d,
                                  fdS2C_d);

                  if (pszSpace)
                     pszTrayCurrent_d = pszSpace + 1;
                  else
                     pszTrayCurrent_d += strlen (pszTrayCurrent_d);
               }

               return pvRet;
            }

         private:
            char           *pszTrays_d;
            char           *pszTrayCurrent_d;
            PrinterCommand *pCmd_d;
            int             fdC2S_d;
            int             fdS2C_d;
         };

         // Enumerate the trays of that device
         if (  pCmd_d->setCommand (COMMAND_TYPE_ENUM_TRAYS, (char *)0)
            && pCmd_d->sendCommand (fdC2S_d)
            && pCmd_d->readCommand (fdS2C_d)
            && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
            )
            return new OmniIJSTrayEnumeration (pCmd_d->getCommandString (),
                                               pCmd_d,
                                               fdC2S_d,
                                               fdS2C_d);

         return new OmniIJSTrayEnumeration (0,
                                            pCmd_d,
                                            fdC2S_d,
                                            fdS2C_d);
      }

      static DeviceTray *
      create (char           *pszName,
              PrinterCommand *pCmd,
              int             fdC2S,
              int             fdS2C)
      {
         // Query the tray of that device
         if (  pCmd->setCommand (COMMAND_TYPE_QUERY_TRAY, pszName)
            && pCmd->sendCommand (fdC2S)
            && pCmd->readCommand (fdS2C)
            && COMMAND_TYPE_ACK == pCmd->getCommandType ()
            )
         {
            char *pszResponse = 0;
            char *pszID       = 0;
            int   iType       = 0;

            pszID = pCmd->getCommandString ();

            pszResponse = strchr (pszID, ' ');

            if (!pszResponse)
               return 0;

            *pszResponse++ = '\0';

            sscanf (pszResponse,
                    "%d",
                    &iType);

#ifndef RETAIL
            if (DebugOutput::shouldOutputOmniIJSProxy ())
            {
               cerr << __FUNCTION__ << ": id            = " << pszID << endl;
               cerr << __FUNCTION__ << ": iType         = " << iType << endl;
            }
#endif

            return new OmniIJSProxyTray (DeviceTray::nameToID (pszID),
                                         iType,
                                         0,
                                         pCmd,
                                         fdC2S,
                                         fdS2C);
         }

         return 0;
      }

   protected:
      DeviceTray *
      createV (int id)
      {
         return 0;
      }

   private:
      PrinterCommand *pCmd_d;
      int             fdC2S_d;
      int             fdS2C_d;
   };

   if (pDT_d)
      return pDT_d;

   pDT_d = OmniIJSProxyTray::create ((char *)0,
                                     pCmd_d,
                                     fdC2S_d,
                                     fdS2C_d);

   return pDT_d;
}

DeviceMedia * OmniIJSProxy::
getCurrentMedia ()
{
   class OmniIJSProxyMedia : public DeviceMedia
   {
   public:
      OmniIJSProxyMedia (int             id,
                         BinaryData     *data,
                         int             iColorAdjustRequired,
                         int             iAbsorption,
                         PrinterCommand *pCmd,
                         int             fdC2S,
                         int             fdS2C)
         : DeviceMedia (id,
                        data,
                        iColorAdjustRequired,
                        iAbsorption)
      {
         pCmd_d  = pCmd;
         fdC2S_d = fdC2S;
         fdS2C_d = fdS2C;
      }

      bool
      isMediaSupported (int id)
      {
         if (  pCmd_d->setCommand (COMMAND_TYPE_IS_MEDIA_SUPPORTED, id)
            && pCmd_d->sendCommand (fdC2S_d)
            && pCmd_d->readCommand (fdS2C_d)
            && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
            )
            return true;

         return false;
      }

      Enumeration *
      getEnumeration ()
      {
         class OmniIJSMediaEnumeration : public Enumeration
         {
         public:
            OmniIJSMediaEnumeration (char           *pszMedias,
                                     PrinterCommand *pCmd,
                                     int             fdC2S,
                                     int             fdS2C)
            {
               pszMedias_d       = 0;
               pszMediaCurrent_d = 0;
               pCmd_d            = pCmd;
               fdC2S_d           = fdC2S;
               fdS2C_d           = fdS2C;

               if (  pszMedias
                  && *pszMedias
                  )
               {
                  pszMedias_d = (char *)malloc (strlen (pszMedias) + 1);
                  if (pszMedias_d)
                  {
                     strcpy (pszMedias_d, pszMedias);

                     pszMediaCurrent_d = pszMedias_d;
                  }
               }
            }

            virtual
            ~OmniIJSMediaEnumeration ()
            {
               if (pszMedias_d)
               {
                  free (pszMedias_d);
               }
            }

            bool
            hasMoreElements ()
            {
               if (  !pszMediaCurrent_d
                  || !*pszMediaCurrent_d
                  )
                  return false;

               return true;
            }

            void *
            nextElement ()
            {
               void *pvRet = 0;

               if (  pszMediaCurrent_d
                  && *pszMediaCurrent_d
                  )
               {
                  char *pszSpace;

                  pszSpace = strchr (pszMediaCurrent_d, ' ');

                  if (pszSpace)
                  {
                     *pszSpace = '\0';
                  }

                  pvRet = create (pszMediaCurrent_d,
                                  pCmd_d,
                                  fdC2S_d,
                                  fdS2C_d);

                  if (pszSpace)
                     pszMediaCurrent_d = pszSpace + 1;
                  else
                     pszMediaCurrent_d += strlen (pszMediaCurrent_d);
               }

               return pvRet;
            }

         private:
            char           *pszMedias_d;
            char           *pszMediaCurrent_d;
            PrinterCommand *pCmd_d;
            int             fdC2S_d;
            int             fdS2C_d;
         };

         // Enumerate the medias of that device
         if (  pCmd_d->setCommand (COMMAND_TYPE_ENUM_MEDIAS, (char *)0)
            && pCmd_d->sendCommand (fdC2S_d)
            && pCmd_d->readCommand (fdS2C_d)
            && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
            )
            return new OmniIJSMediaEnumeration (pCmd_d->getCommandString (),
                                                pCmd_d,
                                                fdC2S_d,
                                                fdS2C_d);

         return new OmniIJSMediaEnumeration (0,
                                             pCmd_d,
                                             fdC2S_d,
                                             fdS2C_d);
      }

      static DeviceMedia *
      create (char           *pszName,
              PrinterCommand *pCmd,
              int             fdC2S,
              int             fdS2C)
      {
         // Query the media of that device
         if (  pCmd->setCommand (COMMAND_TYPE_QUERY_MEDIA, pszName)
            && pCmd->sendCommand (fdC2S)
            && pCmd->readCommand (fdS2C)
            && COMMAND_TYPE_ACK == pCmd->getCommandType ()
            )
         {
            char *pszResponse          = 0;
            char *pszID                = 0;
            int   iColorAdjustRequired = 0;
            int   iAbsorption          = 0;

            pszID = pCmd->getCommandString ();

            pszResponse = strchr (pszID, ' ');

            if (!pszResponse)
               return 0;

            *pszResponse++ = '\0';

            sscanf (pszResponse,
                    "%d %d",
                    &iColorAdjustRequired,
                    &iAbsorption);

#ifndef RETAIL
            if (DebugOutput::shouldOutputOmniIJSProxy ())
            {
               cerr << __FUNCTION__ << ": id                   = " << pszID                << endl;
               cerr << __FUNCTION__ << ": iColorAdjustRequired = " << iColorAdjustRequired << endl;
               cerr << __FUNCTION__ << ": iAbsorption          = " << iAbsorption          << endl;
            }
#endif

            return new OmniIJSProxyMedia (DeviceMedia::nameToID (pszID),
                                          0,
                                          iColorAdjustRequired,
                                          iAbsorption,
                                          pCmd,
                                          fdC2S,
                                          fdS2C);
         }

         return 0;
      }

   protected:
      DeviceMedia *
      createV (int id)
      {
         return 0;
      }

   private:
      PrinterCommand *pCmd_d;
      int             fdC2S_d;
      int             fdS2C_d;
   };

   if (pDM_d)
      return pDM_d;

   pDM_d = OmniIJSProxyMedia::create ((char *)0,
                                      pCmd_d,
                                      fdC2S_d,
                                      fdS2C_d);

   return pDM_d;
}

DeviceResolution * OmniIJSProxy::
getCurrentResolution ()
{
   class OmniIJSProxyResolution : public DeviceResolution
   {
   public:
      OmniIJSProxyResolution (int             id,
                              int             iXRes,
                              int             iYRes,
                              int             iXInternalRes,
                              int             iYInternalRes,
                              BinaryData     *pbdData,
                              int             iCapabilities,
                              int             iDestinationBitsPerPel,
                              int             iScanlineMultiple,
                              PrinterCommand *pCmd,
                              int             fdC2S,
                              int             fdS2C)
         : DeviceResolution (id,
                             iXRes,
                             iYRes,
                             iXInternalRes,
                             iYInternalRes,
                             pbdData,
                             iCapabilities,
                             iDestinationBitsPerPel,
                             iScanlineMultiple)
      {
         pCmd_d  = pCmd;
         fdC2S_d = fdC2S;
         fdS2C_d = fdS2C;
      }

      bool
      isResolutionSupported (int id)
      {
         if (  pCmd_d->setCommand (COMMAND_TYPE_IS_RESOLUTION_SUPPORTED, id)
            && pCmd_d->sendCommand (fdC2S_d)
            && pCmd_d->readCommand (fdS2C_d)
            && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
            )
            return true;

         return false;
      }

      Enumeration *
      getEnumeration ()
      {
         class OmniIJSResolutionEnumeration : public Enumeration
         {
         public:
            OmniIJSResolutionEnumeration (char           *pszResolutions,
                                          PrinterCommand *pCmd,
                                          int             fdC2S,
                                          int             fdS2C)
            {
               pszResolutions_d       = 0;
               pszResolutionCurrent_d = 0;
               pCmd_d                 = pCmd;
               fdC2S_d                = fdC2S;
               fdS2C_d                = fdS2C;

               if (  pszResolutions
                  && *pszResolutions
                  )
               {
                  pszResolutions_d = (char *)malloc (strlen (pszResolutions) + 1);
                  if (pszResolutions_d)
                  {
                     strcpy (pszResolutions_d, pszResolutions);

                     pszResolutionCurrent_d = pszResolutions_d;
                  }
               }
            }

            virtual
            ~OmniIJSResolutionEnumeration ()
            {
               if (pszResolutions_d)
               {
                  free (pszResolutions_d);
               }
            }

            bool
            hasMoreElements ()
            {
               if (  !pszResolutionCurrent_d
                  || !*pszResolutionCurrent_d
                  )
                  return false;

               return true;
            }

            void *
            nextElement ()
            {
               void *pvRet = 0;

               if (  pszResolutionCurrent_d
                  && *pszResolutionCurrent_d
                  )
               {
                  char *pszSpace;

                  pszSpace = strchr (pszResolutionCurrent_d, ' ');

                  if (pszSpace)
                  {
                     *pszSpace = '\0';
                  }

                  pvRet = create (pszResolutionCurrent_d,
                                  pCmd_d,
                                  fdC2S_d,
                                  fdS2C_d);

                  if (pszSpace)
                     pszResolutionCurrent_d = pszSpace + 1;
                  else
                     pszResolutionCurrent_d += strlen (pszResolutionCurrent_d);
               }

               return pvRet;
            }

         private:
            char           *pszResolutions_d;
            char           *pszResolutionCurrent_d;
            PrinterCommand *pCmd_d;
            int             fdC2S_d;
            int             fdS2C_d;
         };

         // Enumerate the resolutions of that device
         if (  pCmd_d->setCommand (COMMAND_TYPE_ENUM_RESOLUTIONS, (char *)0)
            && pCmd_d->sendCommand (fdC2S_d)
            && pCmd_d->readCommand (fdS2C_d)
            && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
            )
            return new OmniIJSResolutionEnumeration (pCmd_d->getCommandString (),
                                                     pCmd_d,
                                                     fdC2S_d,
                                                     fdS2C_d);

         return new OmniIJSResolutionEnumeration (0,
                                                  pCmd_d,
                                                  fdC2S_d,
                                                  fdS2C_d);
      }

      static DeviceResolution *
      create (char           *pszName,
              PrinterCommand *pCmd,
              int             fdC2S,
              int             fdS2C)
      {
         // Query the resolution of that device
         if (  pCmd->setCommand (COMMAND_TYPE_QUERY_RESOLUTION, pszName)
            && pCmd->sendCommand (fdC2S)
            && pCmd->readCommand (fdS2C)
            && COMMAND_TYPE_ACK == pCmd->getCommandType ()
            )
         {
            char *pszResponse            = 0;
            char *pszID                  = 0;
            int   iXRes                  = 0;
            int   iYRes                  = 0;
            int   iXInternalRes          = 0;
            int   iYInternalRes          = 0;
            int   iCapabilities          = 0;
            int   iDestinationBitsPerPel = 0;
            int   iScanlineMultiple      = 0;

            pszID = pCmd->getCommandString ();

            pszResponse = strchr (pszID, ' ');

            if (!pszResponse)
               return 0;

            *pszResponse++ = '\0';

            sscanf (pszResponse,
                    "%d %d %d %d %d %d %d",
                    &iXRes,
                    &iYRes,
                    &iXInternalRes,
                    &iYInternalRes,
                    &iCapabilities,
                    &iDestinationBitsPerPel,
                    &iScanlineMultiple);

#ifndef RETAIL
            if (DebugOutput::shouldOutputOmniIJSProxy ())
            {
               cerr << __FUNCTION__ << ": id                     = " << pszID                  << endl;
               cerr << __FUNCTION__ << ": iXRes                  = " << iXRes                  << endl;
               cerr << __FUNCTION__ << ": iYRes                  = " << iYRes                  << endl;
               cerr << __FUNCTION__ << ": iXInternalRes          = " << iXInternalRes          << endl;
               cerr << __FUNCTION__ << ": iYInternalRes          = " << iYInternalRes          << endl;
               cerr << __FUNCTION__ << ": iCapabilities          = " << iCapabilities          << endl;
               cerr << __FUNCTION__ << ": iDestinationBitsPerPel = " << iDestinationBitsPerPel << endl;
               cerr << __FUNCTION__ << ": iScanlineMultiple      = " << iScanlineMultiple      << endl;
            }
#endif

            return new OmniIJSProxyResolution (DeviceResolution::nameToID (pszID),
                                               iXRes,
                                               iYRes,
                                               iXInternalRes,
                                               iYInternalRes,
                                               0,
                                               iCapabilities,
                                               iDestinationBitsPerPel,
                                               iScanlineMultiple,
                                               pCmd,
                                               fdC2S,
                                               fdS2C);
         }

         return 0;
      }

   protected:
      DeviceResolution *
      createV (int id)
      {
         return 0;
      }

   private:
      PrinterCommand *pCmd_d;
      int             fdC2S_d;
      int             fdS2C_d;
   };

   if (pDR_d)
      return pDR_d;

   pDR_d = OmniIJSProxyResolution::create ((char *)0,
                                           pCmd_d,
                                           fdC2S_d,
                                           fdS2C_d);

   return pDR_d;
}

DevicePrintMode * OmniIJSProxy::
getCurrentPrintMode ()
{
   class OmniIJSProxyPrintMode : public DevicePrintMode
   {
   public:
      OmniIJSProxyPrintMode (int             id,
                             int             iPhysicalCount,
                             int             iLogicalCount,
                             int             iPlanes,
                             PrinterCommand *pCmd,
                             int             fdC2S,
                             int             fdS2C)
         : DevicePrintMode (id,
                            iPhysicalCount,
                            iLogicalCount,
                            iPlanes)
      {
         pCmd_d  = pCmd;
         fdC2S_d = fdC2S;
         fdS2C_d = fdS2C;
      }

      bool
      isPrintModeSupported (int id)
      {
         if (  pCmd_d->setCommand (COMMAND_TYPE_IS_PRINT_MODE_SUPPORTED, id)
            && pCmd_d->sendCommand (fdC2S_d)
            && pCmd_d->readCommand (fdS2C_d)
            && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
            )
            return true;

         return false;
      }

      Enumeration *
      getEnumeration ()
      {
         class OmniIJSPrintModeEnumeration : public Enumeration
         {
         public:
            OmniIJSPrintModeEnumeration (char           *pszPrintModes,
                                         PrinterCommand *pCmd,
                                         int             fdC2S,
                                         int             fdS2C)
            {
               pszPrintModes_d       = 0;
               pszPrintModeCurrent_d = 0;
               pCmd_d                = pCmd;
               fdC2S_d               = fdC2S;
               fdS2C_d               = fdS2C;

               if (  pszPrintModes
                  && *pszPrintModes
                  )
               {
                  pszPrintModes_d = (char *)malloc (strlen (pszPrintModes) + 1);
                  if (pszPrintModes_d)
                  {
                     strcpy (pszPrintModes_d, pszPrintModes);

                     pszPrintModeCurrent_d = pszPrintModes_d;
                  }
               }
            }

            virtual
            ~OmniIJSPrintModeEnumeration ()
            {
               if (pszPrintModes_d)
               {
                  free (pszPrintModes_d);
               }
            }

            bool
            hasMoreElements ()
            {
               if (  !pszPrintModeCurrent_d
                  || !*pszPrintModeCurrent_d
                  )
                  return false;

               return true;
            }

            void *
            nextElement ()
            {
               void *pvRet = 0;

               if (  pszPrintModeCurrent_d
                  && *pszPrintModeCurrent_d
                  )
               {
                  char *pszSpace;

                  pszSpace = strchr (pszPrintModeCurrent_d, ' ');

                  if (pszSpace)
                  {
                     *pszSpace = '\0';
                  }

                  pvRet = create (pszPrintModeCurrent_d,
                                  pCmd_d,
                                  fdC2S_d,
                                  fdS2C_d);

                  if (pszSpace)
                     pszPrintModeCurrent_d = pszSpace + 1;
                  else
                     pszPrintModeCurrent_d += strlen (pszPrintModeCurrent_d);
               }

               return pvRet;
            }

         private:
            char           *pszPrintModes_d;
            char           *pszPrintModeCurrent_d;
            PrinterCommand *pCmd_d;
            int             fdC2S_d;
            int             fdS2C_d;
         };

         // Enumerate the print modes of that device
         if (  pCmd_d->setCommand (COMMAND_TYPE_ENUM_PRINT_MODES, (char *)0)
            && pCmd_d->sendCommand (fdC2S_d)
            && pCmd_d->readCommand (fdS2C_d)
            && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
            )
            return new OmniIJSPrintModeEnumeration (pCmd_d->getCommandString (),
                                                    pCmd_d,
                                                    fdC2S_d,
                                                    fdS2C_d);

         return new OmniIJSPrintModeEnumeration (0,
                                                 pCmd_d,
                                                 fdC2S_d,
                                                 fdS2C_d);
      }

      static DevicePrintMode *
      create (char           *pszName,
              PrinterCommand *pCmd,
              int             fdC2S,
              int             fdS2C)
      {
         // Query the print mode of that device
         if (  pCmd->setCommand (COMMAND_TYPE_QUERY_PRINT_MODE, pszName)
            && pCmd->sendCommand (fdC2S)
            && pCmd->readCommand (fdS2C)
            && COMMAND_TYPE_ACK == pCmd->getCommandType ()
            )
         {
            char *pszResponse    = 0;
            char *pszID          = 0;
            int   iPhysicalCount = 0;
            int   iLogicalCount  = 0;
            int   iPlanes        = 0;

            pszID = pCmd->getCommandString ();

            pszResponse = strchr (pszID, ' ');

            if (!pszResponse)
               return 0;

            *pszResponse++ = '\0';

            sscanf (pszResponse,
                    "%d %d %d",
                    &iPhysicalCount,
                    &iLogicalCount,
                    &iPlanes);

#ifndef RETAIL
            if (DebugOutput::shouldOutputOmniIJSProxy ())
            {
               cerr << __FUNCTION__ << ": id                     = " << pszID          << endl;
               cerr << __FUNCTION__ << ": iPhysicalCount         = " << iPhysicalCount << endl;
               cerr << __FUNCTION__ << ": iLogicalCount          = " << iLogicalCount  << endl;
               cerr << __FUNCTION__ << ": iPlanes                = " << iPlanes        << endl;
            }
#endif

            return new OmniIJSProxyPrintMode (DevicePrintMode::nameToID (pszID),
                                              iPhysicalCount,
                                              iLogicalCount,
                                              iPlanes,
                                              pCmd,
                                              fdC2S,
                                              fdS2C);
         }

         return 0;
      }

   protected:
      DevicePrintMode *
      createV (int id)
      {
         return 0;
      }

   private:
      PrinterCommand *pCmd_d;
      int             fdC2S_d;
      int             fdS2C_d;
   };

   if (pDPM_d)
      return pDPM_d;

   pDPM_d = OmniIJSProxyPrintMode::create ((char *)0,
                                           pCmd_d,
                                           fdC2S_d,
                                           fdS2C_d);

   return pDPM_d;
}

int OmniIJSProxy::
getPDLLevel ()
{
   if (  pCmd_d->setCommand (COMMAND_TYPE_GET_PDL_LEVEL, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return pCmd_d->getCommandInt ();

   return 0;
}

int OmniIJSProxy::
getPDLSubLevel ()
{
   if (  pCmd_d->setCommand (COMMAND_TYPE_GET_PDL_SUBLEVEL, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return pCmd_d->getCommandInt ();

   return 0;
}

int OmniIJSProxy::
getPDLMajorRevisionLevel ()
{
   if (  pCmd_d->setCommand (COMMAND_TYPE_GET_PDL_MAJOR_REVISION_LEVEL, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return pCmd_d->getCommandInt ();

   return 0;
}

int OmniIJSProxy::
getPDLMinorRevisionLevel ()
{
   if (  pCmd_d->setCommand (COMMAND_TYPE_GET_PDL_MINOR_REVISION_LEVEL, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return pCmd_d->getCommandInt ();

   return 0;
}

bool OmniIJSProxy::
hasCapability (long lMask)
{
   if (  pCmd_d->setCommand (COMMAND_TYPE_HAS_CAPABILITY, lMask)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return pCmd_d->getCommandLong ();

   return 0;
}

bool OmniIJSProxy::
hasRasterCapability (long lMask)
{
   if (  pCmd_d->setCommand (COMMAND_TYPE_HAS_RASTER_CAPABILITY, lMask)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return pCmd_d->getCommandLong ();

   return 0;
}

bool OmniIJSProxy::
hasDeviceOption (long lMask)
{
   if (  pCmd_d->setCommand (COMMAND_TYPE_HAS_DEVICE_OPTION, lMask)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return pCmd_d->getCommandLong ();

   return 0;
}

int OmniIJSProxy::
getScanlineMultiple ()
{
   if (!pDR_d)
   {
      pDR_d = getCurrentResolution ();
   }

   if (pDR_d)
      return pDR_d->getScanlineMultiple ();
   else
      return 1;
}

bool OmniIJSProxy::
beginJob ()
{
   // Start the job on the device
   if (  pCmd_d->setCommand (COMMAND_TYPE_BEGIN_JOB, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return true;

   return false;
}

bool OmniIJSProxy::
newFrame ()
{
   // Send a new page to the device
   if (  pCmd_d->setCommand (COMMAND_TYPE_NEW_FRAME, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return true;

   return false;
}

bool OmniIJSProxy::
newFrame (char *pszJobProperties)
{
   if (  pszJobProperties
      && *pszJobProperties
      )
   {
      // If there are job properties, pass them in
      if (  !pCmd_d->setCommand (COMMAND_TYPE_SET_JOB_PROPERTIES, pszJobProperties)
         || !pCmd_d->sendCommand (fdC2S_d)
         || !pCmd_d->readCommand (fdS2C_d)
         || COMMAND_TYPE_ACK != pCmd_d->getCommandType ()
         )
         return false;
   }

   // Send a new page to the device
   if (  pCmd_d->setCommand (COMMAND_TYPE_NEW_FRAME, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return true;

   return false;
}

bool OmniIJSProxy::
endJob ()
{
   // End the job on the device
   if (  pCmd_d->setCommand (COMMAND_TYPE_END_JOB, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return true;

   return false;
}

bool OmniIJSProxy::
abortJob ()
{
   // Abort the job on the device
   if (  pCmd_d->setCommand (COMMAND_TYPE_ABORT_JOB, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return true;

   return false;
}

bool OmniIJSProxy::
rasterize (PBYTE        pbBits,
           PBITMAPINFO2 pbmi,
           PRECTL       prectlPageLocation,
           BITBLT_TYPE  eType)
{
   if (pbmi->cbFix > cbBuffer1_d)
   {
      // Release the previous buffer
      if (pbBuffer1_d)
      {
         if (  !pCmd_d->setCommand (COMMAND_TYPE_DETACH_BUFFER1, idBuffer1_d)
            || !pCmd_d->sendCommand (fdC2S_d)
            || !pCmd_d->readCommand (fdS2C_d)
            || COMMAND_TYPE_ACK != pCmd_d->getCommandType ()
            )
            return false;

         shmdt (pbBuffer1_d);
         shmctl (idBuffer1_d, IPC_RMID, 0);

         idBuffer1_d = -1;
         cbBuffer1_d = 0;
         pbBuffer1_d = 0;
      }

      idBuffer1_d = shmget (IPC_PRIVATE, pbmi->cbFix, 0666);

      if (0 < idBuffer1_d)
      {
         cbBuffer1_d = pbmi->cbFix;
         pbBuffer1_d = (byte *)shmat (idBuffer1_d, 0, 0);

         if (-1 == (int)pbBuffer1_d)
         {
            return false;
         }

         if (  !pCmd_d->setCommand (COMMAND_TYPE_ATTACH_BUFFER1, idBuffer1_d)
            || !pCmd_d->sendCommand (fdC2S_d)
            || !pCmd_d->readCommand (fdS2C_d)
            || COMMAND_TYPE_ACK != pCmd_d->getCommandType ()
            )
            return false;
      }
   }

   int cbBuffer2;

   cbBuffer2 = pbmi->cy * (((pbmi->cx * pbmi->cBitCount + 31) >> 5) << 2);

   if (cbBuffer2 > cbBuffer1_d)
   {
      // Release the previous buffer
      if (pbBuffer2_d)
      {
         if (  !pCmd_d->setCommand (COMMAND_TYPE_DETACH_BUFFER2, idBuffer2_d)
            || !pCmd_d->sendCommand (fdC2S_d)
            || !pCmd_d->readCommand (fdS2C_d)
            || COMMAND_TYPE_ACK != pCmd_d->getCommandType ()
            )
            return false;

         shmdt (pbBuffer2_d);
         shmctl (idBuffer2_d, IPC_RMID, 0);

         idBuffer2_d = -1;
         cbBuffer2_d = 0;
         pbBuffer2_d = 0;
      }

      idBuffer2_d = shmget (IPC_PRIVATE, cbBuffer2, 0666);

      if (0 < idBuffer2_d)
      {
         cbBuffer2_d = cbBuffer2;
         pbBuffer2_d = (byte *)shmat (idBuffer2_d, 0, 0);

         if (-1 == (int)pbBuffer2_d)
         {
            return false;
         }

         if (  !pCmd_d->setCommand (COMMAND_TYPE_ATTACH_BUFFER2, idBuffer2_d)
            || !pCmd_d->sendCommand (fdC2S_d)
            || !pCmd_d->readCommand (fdS2C_d)
            || COMMAND_TYPE_ACK != pCmd_d->getCommandType ()
            )
            return false;
      }
   }

   memcpy (pbBuffer1_d, pbmi,   pbmi->cbFix);
   memcpy (pbBuffer2_d, pbBits, cbBuffer2);

   char achBuffer[64]; // @TBD

   sprintf (achBuffer,
            "%d %d %d %d %d",
            eType,
            prectlPageLocation->xLeft,
            prectlPageLocation->yBottom,
            prectlPageLocation->xRight,
            prectlPageLocation->yTop);

   if (  pCmd_d->setCommand (COMMAND_TYPE_RASTERIZE, achBuffer)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return true;

   return false;
}

string * OmniIJSProxy::
queryCurrentJobProperties ()
{
   // Query the current job properties
   if (  pCmd_d->setCommand (COMMAND_TYPE_QUERY_CURRENT_JOB_PROPERTIES, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return new string (pCmd_d->getCommandString ());

   return 0;
}

Enumeration * OmniIJSProxy::
listJobPropertyKeys ()
{
   class OmniIJSJobPropertyKeyEnumeration : public Enumeration
   {
   public:
      OmniIJSJobPropertyKeyEnumeration (char *pszJobPropertyKeyIDs)
      {
         pszJobPropertyKeyIDs_d       = 0;
         pszJobPropertyKeyIDCurrent_d = 0;

         if (  pszJobPropertyKeyIDs
            && *pszJobPropertyKeyIDs
            )
         {
            pszJobPropertyKeyIDs_d = (char *)malloc (strlen (pszJobPropertyKeyIDs) + 1);
            if (pszJobPropertyKeyIDs_d)
            {
               strcpy (pszJobPropertyKeyIDs_d, pszJobPropertyKeyIDs);

               pszJobPropertyKeyIDCurrent_d = pszJobPropertyKeyIDs_d;
            }
         }
      }

      virtual
      ~OmniIJSJobPropertyKeyEnumeration ()
      {
         if (pszJobPropertyKeyIDs_d)
         {
            free (pszJobPropertyKeyIDs_d);
         }
      }

      bool
      hasMoreElements ()
      {
         if (  !pszJobPropertyKeyIDCurrent_d
            || !*pszJobPropertyKeyIDCurrent_d
            )
            return false;

         return true;
      }

      void *
      nextElement ()
      {
         void *pvRet = 0;

         if (  pszJobPropertyKeyIDCurrent_d
            && *pszJobPropertyKeyIDCurrent_d
            )
         {
            char *pszSpace;

            pszSpace = strchr (pszJobPropertyKeyIDCurrent_d, ' ');

            if (pszSpace)
            {
               *pszSpace = '\0';
            }

            pvRet = pszJobPropertyKeyIDCurrent_d;

            if (pszSpace)
               pszJobPropertyKeyIDCurrent_d = pszSpace + 1;
            else
               pszJobPropertyKeyIDCurrent_d += strlen (pszJobPropertyKeyIDCurrent_d);
         }

         return pvRet;
      }

   private:
      char *pszJobPropertyKeyIDs_d;
      char *pszJobPropertyKeyIDCurrent_d;
   };

   // Enumerate the job property keys of omni
   if (  pCmd_d->setCommand (COMMAND_TYPE_LIST_JOB_PROPERTY_KEYS, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return new OmniIJSJobPropertyKeyEnumeration (pCmd_d->getCommandString ());

   return new OmniIJSJobPropertyKeyEnumeration (0);
}

Enumeration * OmniIJSProxy::
listDeviceJobPropertyKeys ()
{
   class OmniIJSDeviceJobPropertyKeyEnumeration : public Enumeration
   {
   public:
      OmniIJSDeviceJobPropertyKeyEnumeration (char *pszDeviceJobPropertyKeyIDs)
      {
         pszDeviceJobPropertyKeyIDs_d       = 0;
         pszDeviceJobPropertyKeyIDCurrent_d = 0;

         if (  pszDeviceJobPropertyKeyIDs
            && *pszDeviceJobPropertyKeyIDs
            )
         {
            pszDeviceJobPropertyKeyIDs_d = (char *)malloc (strlen (pszDeviceJobPropertyKeyIDs) + 1);
            if (pszDeviceJobPropertyKeyIDs_d)
            {
               strcpy (pszDeviceJobPropertyKeyIDs_d, pszDeviceJobPropertyKeyIDs);

               pszDeviceJobPropertyKeyIDCurrent_d = pszDeviceJobPropertyKeyIDs_d;
            }
         }
      }

      virtual
      ~OmniIJSDeviceJobPropertyKeyEnumeration ()
      {
         if (pszDeviceJobPropertyKeyIDs_d)
         {
            free (pszDeviceJobPropertyKeyIDs_d);
         }
      }

      bool
      hasMoreElements ()
      {
         if (  !pszDeviceJobPropertyKeyIDCurrent_d
            || !*pszDeviceJobPropertyKeyIDCurrent_d
            )
            return false;

         return true;
      }

      void *
      nextElement ()
      {
         void *pvRet = 0;

         if (  pszDeviceJobPropertyKeyIDCurrent_d
            && *pszDeviceJobPropertyKeyIDCurrent_d
            )
         {
            char *pszSpace;

            pszSpace = strchr (pszDeviceJobPropertyKeyIDCurrent_d, ' ');

            if (pszSpace)
            {
               *pszSpace = '\0';
            }

            pvRet = pszDeviceJobPropertyKeyIDCurrent_d;

            if (pszSpace)
               pszDeviceJobPropertyKeyIDCurrent_d = pszSpace + 1;
            else
               pszDeviceJobPropertyKeyIDCurrent_d += strlen (pszDeviceJobPropertyKeyIDCurrent_d);
         }

         return pvRet;
      }

   private:
      char *pszDeviceJobPropertyKeyIDs_d;
      char *pszDeviceJobPropertyKeyIDCurrent_d;
   };

   // Enumerate the job property keys of that device
   if (  pCmd_d->setCommand (COMMAND_TYPE_LIST_DEVICE_JOB_PROPERTY_KEYS, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return new OmniIJSDeviceJobPropertyKeyEnumeration (pCmd_d->getCommandString ());

   return new OmniIJSDeviceJobPropertyKeyEnumeration (0);
}

string * OmniIJSProxy::
getJobProperty (char *pszKey)
{
   // Query the current job properties
   if (  pCmd_d->setCommand (COMMAND_TYPE_GET_JOB_PROPERTY, pszKey)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return new string (pCmd_d->getCommandString ());

   return 0;
}

string * OmniIJSProxy::
getJobPropertyType (char *pszKey)
{
   // Query the current job properties
   if (  pCmd_d->setCommand (COMMAND_TYPE_GET_JOB_PROPERTY_TYPE, pszKey)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return new string (pCmd_d->getCommandString ());

   return 0;
}

Enumeration * OmniIJSProxy::
getDitherEnumeration ()
{
   class OmniIJSDitherEnumeration : public Enumeration
   {
   public:
      OmniIJSDitherEnumeration (char *pszDitherIDs)
      {
         pszDitherIDs_d       = 0;
         pszDitherIDCurrent_d = 0;

         if (  pszDitherIDs
            && *pszDitherIDs
            )
         {
            pszDitherIDs_d = (char *)malloc (strlen (pszDitherIDs) + 1);
            if (pszDitherIDs_d)
            {
               strcpy (pszDitherIDs_d, pszDitherIDs);

               pszDitherIDCurrent_d = pszDitherIDs_d;
            }
         }
      }

      virtual
      ~OmniIJSDitherEnumeration ()
      {
         if (pszDitherIDs_d)
         {
            free (pszDitherIDs_d);
         }
      }

      bool
      hasMoreElements ()
      {
         if (  !pszDitherIDCurrent_d
            || !*pszDitherIDCurrent_d
            )
            return false;

         return true;
      }

      void *
      nextElement ()
      {
         void *pvRet = 0;

         if (  pszDitherIDCurrent_d
            && *pszDitherIDCurrent_d
            )
         {
            char *pszSpace;

            pszSpace = strchr (pszDitherIDCurrent_d, ' ');

            if (pszSpace)
            {
               *pszSpace = '\0';
            }

            pvRet = pszDitherIDCurrent_d;

            if (pszSpace)
               pszDitherIDCurrent_d = pszSpace + 1;
            else
               pszDitherIDCurrent_d += strlen (pszDitherIDCurrent_d);
         }

         return pvRet;
      }

   private:
      char *pszDitherIDs_d;
      char *pszDitherIDCurrent_d;
   };

   // Enumerate the dither ids of that device
   if (  pCmd_d->setCommand (COMMAND_TYPE_ENUM_DITHER_IDS, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && COMMAND_TYPE_ACK == pCmd_d->getCommandType ()
      )
      return new OmniIJSDitherEnumeration (pCmd_d->getCommandString ());

   return new OmniIJSDitherEnumeration (0);
}
