/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdlib.h>
#include <string.h>
#include <fstream.h>
#include <unistd.h>

#include "PrinterCommand.hpp"
#include "DebugOutput.hpp"

PrinterCommand::
PrinterCommand (char *pszProgram)
{
   pCmd_d       = 0;
   iCmdLength_d = 0;
   pszProgram_d = 0;

   if (  pszProgram
      && *pszProgram
      )
   {
      pszProgram_d = (char *)malloc (strlen (pszProgram) + 1);
      if (pszProgram_d)
      {
         strcpy (pszProgram_d, pszProgram);
      }
   }
}

PrinterCommand::
~PrinterCommand ()
{
   if (pCmd_d)
   {
      free (pCmd_d);
   }
   if (pszProgram_d)
   {
      free (pszProgram_d);
   }
}

bool PrinterCommand::
setCommand (COMMAND_TYPE eCommand, char *pszLine)
{
   size_t iNewCmdLength = sizeof (COMMAND_PACKET);

   if (  pszLine
      && *pszLine
      )
   {
      iNewCmdLength += strlen (pszLine);
   }

   if (iNewCmdLength > iCmdLength_d)
   {
      pCmd_d = (PCOMMAND_PACKET)realloc (pCmd_d, iNewCmdLength);
   }

   iCmdLength_d = iNewCmdLength;

   if (!pCmd_d)
   {
      cerr << "Command allocation failed" << endl;

      return false;
   }

   pCmd_d->eCommand          = eCommand;
   pCmd_d->cbLength          = iCmdLength_d;
   pCmd_d->eFormat           = COMMAND_FORMAT_STRING;
   pCmd_d->achCommandLine[0] = '\0';

   if (  pszLine
      && *pszLine
      )
   {
      strcpy (pCmd_d->achCommandLine, pszLine);
   }

   return true;
}

bool PrinterCommand::
setCommand (COMMAND_TYPE eCommand, int iValue)
{
   size_t iNewCmdLength = sizeof (COMMAND_PACKET) - 1 + sizeof (int);

   if (iNewCmdLength > iCmdLength_d)
   {
      pCmd_d = (PCOMMAND_PACKET)realloc (pCmd_d, iNewCmdLength);
   }

   iCmdLength_d = iNewCmdLength;

   if (!pCmd_d)
   {
      cerr << "Command allocation failed" << endl;

      return false;
   }

   pCmd_d->eCommand = eCommand;
   pCmd_d->cbLength = iCmdLength_d;
   pCmd_d->eFormat  = COMMAND_FORMAT_INTEGER;

   unsigned char *pbBuffer = (unsigned char *)pCmd_d->achCommandLine;

   for (size_t i = 0; i < sizeof (int); i++)
   {
      *pbBuffer = iValue & 0xFF;

      iValue >>= 8;
      pbBuffer++;
   }

   return true;
}

bool PrinterCommand::
setCommand (COMMAND_TYPE eCommand, long int iValue)
{
   size_t iNewCmdLength = sizeof (COMMAND_PACKET) - 1 + sizeof (long int);

   if (iNewCmdLength > iCmdLength_d)
   {
      pCmd_d = (PCOMMAND_PACKET)realloc (pCmd_d, iNewCmdLength);
   }

   iCmdLength_d = iNewCmdLength;

   if (!pCmd_d)
   {
      cerr << "Command allocation failed" << endl;

      return false;
   }

   pCmd_d->eCommand = eCommand;
   pCmd_d->cbLength = iCmdLength_d;
   pCmd_d->eFormat  = COMMAND_FORMAT_LONG;

   unsigned char *pbBuffer = (unsigned char *)pCmd_d->achCommandLine;

   for (size_t i = 0; i < sizeof (long int); i++)
   {
      *pbBuffer = iValue & 0xFF;

      iValue >>= 8;
      pbBuffer++;
   }

   return true;
}

PCOMMAND_PACKET PrinterCommand::
getCommand ()
{
   return pCmd_d;
}

char * PrinterCommand::
getCommandString ()
{
   return pCmd_d->achCommandLine;
}

int PrinterCommand::
getCommandInt ()
{
   return *(int*)pCmd_d->achCommandLine;
}

long int PrinterCommand::
getCommandLong ()
{
   return *(long int*)pCmd_d->achCommandLine;
}

COMMAND_TYPE PrinterCommand::
getCommandType ()
{
   return pCmd_d->eCommand;
}

size_t PrinterCommand::
getCommandLength ()
{
   return iCmdLength_d;
}

bool PrinterCommand::
readCommand (int fd)
{
   typedef unsigned char byte;

   byte    *pbBuffer;
   ssize_t  cbBytesToRead;
   ssize_t  rc;

   if (!pCmd_d)
   {
      // Strange, can't use alloc...
      pCmd_d = (PCOMMAND_PACKET)realloc (pCmd_d, sizeof (COMMAND_PACKET));

      if (!pCmd_d)
         return false;
   }

   cbBytesToRead = sizeof (COMMAND_PACKET);
   pbBuffer      = (byte *)pCmd_d;

   while (0 < cbBytesToRead)
   {
      rc = read (fd, pbBuffer, cbBytesToRead);

      if (   0 == rc
         || -1 == rc
         )
         break;

      cbBytesToRead -= rc;
      pbBuffer      += rc;
   }

   if (0 < cbBytesToRead)
   {
      cerr << "client: read "
           << sizeof (COMMAND_PACKET) - cbBytesToRead
           << " bytes, expecting "
           << sizeof (COMMAND_PACKET)
           << " @ line "
           << __LINE__
           << endl;

      return false;
   }

   if (pCmd_d->cbLength > sizeof (COMMAND_PACKET))
   {
      if (pCmd_d->cbLength > iCmdLength_d)
      {
         pCmd_d = (PCOMMAND_PACKET)realloc (pCmd_d, pCmd_d->cbLength);

         if (!pCmd_d)
            return false;

         iCmdLength_d = pCmd_d->cbLength;
      }

      size_t iBytesLeft = pCmd_d->cbLength - sizeof (COMMAND_PACKET);

      cbBytesToRead = iBytesLeft;
      pbBuffer      = (byte *)(pCmd_d->achCommandLine + 1);

      while (0 < cbBytesToRead)
      {
         rc = read (fd, pbBuffer, cbBytesToRead);

         if (   0 == rc
            || -1 == rc
            )
            break;

         cbBytesToRead -= rc;
         pbBuffer      += rc;
      }

      if (0 < cbBytesToRead)
      {
         cerr << " client: read "
              << iBytesLeft - cbBytesToRead
              << " bytes, expecting "
              << iBytesLeft
              << " @ line "
              << __LINE__
              << endl;

         return false;
      }
   }

   printCommand (": read {");

   return true;
}

bool PrinterCommand::
sendCommand (int fd)
{
   size_t rc;

   rc = write (fd, pCmd_d, iCmdLength_d);

   if (rc != iCmdLength_d)
   {
      cerr << "client: write "
           << rc
           << " bytes, expecting "
           << iCmdLength_d
           << " @ line "
           << __LINE__
           << endl;

      return false;
   }
   else
   {
      printCommand (": write {");

      return true;
   }
}

char * PrinterCommand::
commandToString (COMMAND_TYPE eCommand)
{
   switch (eCommand)
   {
   case COMMAND_TYPE_ACK:                           return "<Acknowledge>";
   case COMMAND_TYPE_NACK:                          return "<Not Acknowledge>";
   case COMMAND_TYPE_UNSUPPORTED:                   return "<Unsupported>";
   case COMMAND_TYPE_TERMINATE:                     return "<Terminate>";
   case COMMAND_TYPE_SET_DEVICE_NAME:               return "<Set Device Name>";
   case COMMAND_TYPE_OMNI_SET_ADVANCED:             return "<Omni Set Advanced Mode>";
   case COMMAND_TYPE_SET_JOB_PROPERTIES:            return "<Set Job Properties>";
   case COMMAND_TYPE_NEW_DEVICE:                    return "<New Device>";
   case COMMAND_TYPE_GET_VERSION:                   return "<Get Version>";
   case COMMAND_TYPE_GET_DRIVER_NAME:               return "<Get Driver Name>";
   case COMMAND_TYPE_GET_DEVICE_NAME:               return "<Get Device Name>";
   case COMMAND_TYPE_GET_SHORT_NAME:                return "<Get Device Short Name>";
   case COMMAND_TYPE_QUERY_ORIENTATION:             return "<Query Orientation>";
   case COMMAND_TYPE_QUERY_DITHER_ID:               return "<Query Dither ID>";
   case COMMAND_TYPE_QUERY_FORM:                    return "<Query Form>";
   case COMMAND_TYPE_QUERY_TRAY:                    return "<Query Tray>";
   case COMMAND_TYPE_QUERY_MEDIA:                   return "<Query Media>";
   case COMMAND_TYPE_QUERY_RESOLUTION:              return "<Query Resolution>";
   case COMMAND_TYPE_QUERY_PRINT_MODE:              return "<Query Print Mode>";
   case COMMAND_TYPE_GET_PDL_LEVEL:                 return "<Get PDL Level>";
   case COMMAND_TYPE_GET_PDL_SUBLEVEL:              return "<Get PDL Sublevel>";
   case COMMAND_TYPE_GET_PDL_MAJOR_REVISION_LEVEL:  return "<Get PDL Major Revision Level>";
   case COMMAND_TYPE_GET_PDL_MINOR_REVISION_LEVEL:  return "<Get PDL Minor Revision Level>";
   case COMMAND_TYPE_HAS_CAPABILITY:                return "<Has Capability>";
   case COMMAND_TYPE_HAS_RASTER_CAPABILITY:         return "<Has Raster Capability>";
   case COMMAND_TYPE_HAS_DEVICE_OPTION:             return "<Has Device Option>";
   case COMMAND_TYPE_ENUM_ORIENTATIONS:             return "<Enumerate Orientations>";
   case COMMAND_TYPE_ENUM_DITHER_IDS:               return "<Enumerate Dither IDs>";
   case COMMAND_TYPE_ENUM_FORMS:                    return "<Enumerate Forms>";
   case COMMAND_TYPE_ENUM_TRAYS:                    return "<Enumerate Trays>";
   case COMMAND_TYPE_ENUM_MEDIAS:                   return "<Enumerate Medias>";
   case COMMAND_TYPE_ENUM_RESOLUTIONS:              return "<Enumerate Resolutions>";
   case COMMAND_TYPE_ENUM_PRINT_MODES:              return "<Enumerate Print Modes>";
   case COMMAND_TYPE_IS_ORIENTATION_SUPPORTED:      return "<Is Orientation Supported>";
   case COMMAND_TYPE_IS_FORM_SUPPORTED:             return "<Is Form Supported>";
   case COMMAND_TYPE_IS_TRAY_SUPPORTED:             return "<Is Tray Supported>";
   case COMMAND_TYPE_IS_MEDIA_SUPPORTED:            return "<Is Media Supported>";
   case COMMAND_TYPE_IS_RESOLUTION_SUPPORTED:       return "<Is Resolution Supported>";
   case COMMAND_TYPE_IS_PRINT_MODE_SUPPORTED:       return "<Is Print Mode Supported>";
   case COMMAND_TYPE_BEGIN_JOB:                     return "<Begin Job>";
   case COMMAND_TYPE_NEW_FRAME:                     return "<New Frame>";
   case COMMAND_TYPE_END_JOB:                       return "<End Job>";
   case COMMAND_TYPE_ABORT_JOB:                     return "<Abort Job>";
   case COMMAND_TYPE_ATTACH_BUFFER1:                return "<Attach Buffer #1>";
   case COMMAND_TYPE_ATTACH_BUFFER2:                return "<Attach Buffer #2>";
   case COMMAND_TYPE_DETACH_BUFFER1:                return "<Detach Buffer #1>";
   case COMMAND_TYPE_DETACH_BUFFER2:                return "<Detach Buffer #2>";
   case COMMAND_TYPE_RASTERIZE:                     return "<Rasterize>";
   case COMMAND_TYPE_QUERY_CURRENT_JOB_PROPERTIES:  return "<Query Current JobProperties>";
   case COMMAND_TYPE_LIST_JOB_PROPERTY_KEYS:        return "<List Job Property Keys>";
   case COMMAND_TYPE_LIST_DEVICE_JOB_PROPERTY_KEYS: return "<List Device Job Property Keys>";
   case COMMAND_TYPE_GET_JOB_PROPERTY_TYPE:         return "<Get Job Property Type>";
   case COMMAND_TYPE_GET_JOB_PROPERTY:              return "<Get Job Property>";
   default:                                         return "<\?\?\?>";
   }
}

void PrinterCommand::
printCommand (char *pszHeader)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPrinterCommand ())
   {
      if (pszProgram_d)
         cerr << pszProgram_d;

      cerr << pszHeader
           << commandToString (pCmd_d->eCommand)
           << ", "
           << pCmd_d->cbLength
           << ", "
           << (long unsigned int)pCmd_d->eFormat;

      if (COMMAND_FORMAT_STRING == pCmd_d->eFormat)
      {
         cerr << ", \"" << pCmd_d->achCommandLine << "\"";
      }
      else if (COMMAND_FORMAT_BINARY == pCmd_d->eFormat)
      {
         size_t iBytesLeft = pCmd_d->cbLength - sizeof (COMMAND_PACKET);

         cerr << ", {" << hex;

         for (size_t i = 0; i <= iBytesLeft; i++)
         {
            cerr << "0x" << (int)pCmd_d->achCommandLine[i];

            if (i != iBytesLeft)
               cerr << ",";
         }

         cerr << dec << "}";
      }
      else if (COMMAND_FORMAT_INTEGER == pCmd_d->eFormat)
      {
         cerr << ", " << *(int *)pCmd_d->achCommandLine;
      }
      else if (COMMAND_FORMAT_LONG == pCmd_d->eFormat)
      {
         cerr << ", " << *(long int *)pCmd_d->achCommandLine;
      }
      else
      {
         cerr << ", Error: unhandled case for format " << (long int)pCmd_d->eFormat;
      }

      cerr << "}" << endl;
   }
#endif
}
