/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "Star_ESCP2_Instance.hpp"
#include "defines.hpp"

#include <iostream>
#include <strstream>

const static bool fTestNoCompression = true;

Star_ESCP2_Instance::
Star_ESCP2_Instance (Device *pDevice)
   : DeviceInstance (pDevice)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "Star_ESCP2_Instance::Star_ESCP2_Instance ()" << endl;
#endif

   fHaveInitialized_d  = false;
   fHaveSetupPrinter_d = false;
   fUseBidirectional_d = true;
   ptlPrintHead_d.x    = 0;
   ptlPrintHead_d.y    = 0;
}

Star_ESCP2_Instance::
~Star_ESCP2_Instance ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "Star_ESCP2_Instance::~Star_ESCP2_Instance ()" << endl;
#endif
}

void Star_ESCP2_Instance::
initializeInstance ()
{
   if (fHaveInitialized_d)
      return;

   fHaveInitialized_d = true;
}

static char *apszDeviceJobPropertyKeys[] = {
   "bidirectional"
};
#define JOBPROP_BIDIRECTIONAL apszDeviceJobPropertyKeys[0]

Enumeration * Star_ESCP2_Instance::
listDeviceJobPropertyKeys ()
{
   class DeviceJobPropertyKeyEnumerator : public Enumeration
   {
   public:
      DeviceJobPropertyKeyEnumerator (char **apszKeys, int iDimOfKeys)
      {
         iWhichKey_d = iDimOfKeys;
         apszKeys_d  = apszKeys;
      }

      virtual bool hasMoreElements ()
      {
         return iWhichKey_d > 0;
      }

      virtual void * nextElement ()
      {
         if (0 < iWhichKey_d)
         {
            iWhichKey_d--;
            return apszKeys_d[iWhichKey_d];
         }

         return 0;
      }

   private:
      int   iWhichKey_d;
      char **apszKeys_d;
   };

   return new DeviceJobPropertyKeyEnumerator (apszDeviceJobPropertyKeys, dimof (apszDeviceJobPropertyKeys));
}

bool Star_ESCP2_Instance::
handleDeviceJobProperty (char *pszKey,
                         char *pszValue)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "Star_ESCP2_Instance::handleDeviceJobProperty ("
                          << pszKey << ", "
                          << pszValue << ")"
                          << endl;
#endif

   if (0 == strcasecmp (pszKey, JOBPROP_BIDIRECTIONAL))
   {
      if (0 == strcmp (pszValue, "true"))
      {
         fUseBidirectional_d = true;
         return true;
      }
      else if (0 == strcmp (pszValue, "false"))
      {
         fUseBidirectional_d = false;
         return true;
      }
   }

   return false;
}

string * Star_ESCP2_Instance::
getJobPropertyType (char *pszKey)
{
   if (0 == strcasecmp (pszKey, JOBPROP_BIDIRECTIONAL))
   {
      if (fUseBidirectional_d)
      {
         return new string ("boolean true");
      }
      else
      {
         return new string ("boolean false");
      }
   }

   return 0;
}

string * Star_ESCP2_Instance::
getJobProperty (char *pszKey)
{
   if (0 == strcasecmp (pszKey, JOBPROP_BIDIRECTIONAL))
   {
      if (fUseBidirectional_d)
      {
         return new string ("true");
      }
      else
      {
         return new string ("false");
      }
   }

   return 0;
}

void Star_ESCP2_Instance::
setupPrinter ()
{
   if (fHaveSetupPrinter_d)
      return;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "Star_ESCP2_Instance::setupPrinter ()" << endl;
#endif

   fHaveSetupPrinter_d = true;

   DeviceResolution *pRes      = getCurrentResolution ();
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

   pCmd = pCommands->getCommandData ("cmdBeginRasterGraphics");
   if (pCmd)
   {
      pDevice_d->sendBinaryDataToDevice (pCmd);
   }
   else
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "Error: There is no cmdBeginRasterGraphics defined for this device!" << endl;
#endif
   }

   pCmd = pCommands->getCommandData ("cmdSetResolution");
   if (pCmd)
   {
      pDevice_d->sendPrintfToDevice (pCmd, 3600 / pRes->getYRes ());
   }
   else
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "Error: There is no cmdSetResolution defined for this device!" << endl;
#endif
   }

   DeviceForm *pForm = getCurrentForm ();

   pCmd = pCommands->getCommandData ("cmdSetFormSize");
   if (pCmd)
   {
      int iUnits = pForm->getHardCopyCap ()->getYPels ();

#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "iUnits = " << iUnits << endl;
#endif

      pDevice_d->sendPrintfToDevice (pCmd, iUnits);
   }
   else
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "Error: There is no cmdSetFormSize defined for this device!" << endl;
#endif
   }

   // @TBD
   fUseMicroweave_d = false;

   if (pRes->isID (DeviceResolution::RESOLUTION_720_X_720))
   {
      fUseMicroweave_d = true;
   }

   pCmd = pCommands->getCommandData ("cmdBidi");
   if (pCmd)
   {
      char chState = '0';

      if (fUseBidirectional_d)
      {
         chState = '1';
      }

      if (fUseMicroweave_d)
         // Turn off bidirectional for microweaving
         chState = '0';

      pDevice_d->sendPrintfToDevice (pCmd, chState);
   }
   else
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "Error: There is no cmdBidi defined for this device!" << endl;
#endif
   }

   // @TBD
///pDevice_d->sendBinaryDataToDevice (getCurrentTray ());

   pCmd = pCommands->getCommandData ("cmdSetHardwareMicroweave");
   if (pCmd)
   {
      if (fUseMicroweave_d)
         pDevice_d->sendPrintfToDevice (pCmd, '1');
   }
   else
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "Error: There is no cmdSetHardwareMicroweave defined for this device!" << endl;
#endif
   }

   // @TBD
   pCmd = pCommands->getCommandData ("cmdSetDotSize");
   if (pCmd)
   {
   }
   else
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "Error: There is no cmdSetDotSize defined for this device!" << endl;
#endif
   }
}

bool Star_ESCP2_Instance::
beginJob ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "Star_ESCP2_Instance::beginJob ()" << endl;
#endif

   pCmd = pCommands->getCommandData ("cmdUSBInit");

   if (  hasDeviceOption (USB_PORT_SUPPORT)
      && pCmd
      )
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdUSBInit = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   pCmd = pCommands->getCommandData ("cmdInit");
   if (pCmd)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdInit = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   return true;
}

bool Star_ESCP2_Instance::
beginJob (bool fJobPropertiesChanged)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "Star_ESCP2_Instance::beginJob (with props)" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The job properties are as follows:" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The orientation is = " << getCurrentOrientation() << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "iDitherID = " << getCurrentDitherID () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The form is = " << getCurrentForm () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The tray is = " << getCurrentTray () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The media is = " << getCurrentMedia () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The resolution is = " << getCurrentResolution () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The print mode is = " << getCurrentPrintMode () << endl;
#endif

   // @TBD - reinitialize with new job properties

   // Call common code
   return beginJob ();
}

bool Star_ESCP2_Instance::
newFrame ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "Star_ESCP2_Instance::newFrame ()" << endl;
#endif

   pCmd = pCommands->getCommandData ("cmdPageEject");
   if (pCmd)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdPageEject = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   ditherNewFrame ();

   ptlPrintHead_d.x = 0;
   ptlPrintHead_d.y = 0;

   return true;
}

bool Star_ESCP2_Instance::
newFrame (bool fJobPropertiesChanged)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "Star_ESCP2_Instance::newFrame (with props)" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The job properties are as follows:" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The orientation is = " << getCurrentOrientation() << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "iDitherID = " << getCurrentDitherID () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The form is = " << getCurrentForm () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The tray is = " << getCurrentTray () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The media is = " << getCurrentMedia () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The resolution is = " << getCurrentResolution () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The print mode is = " << getCurrentPrintMode () << endl;
#endif

   // @TBD - reinitialize with new job properties

   // Call common code
   return newFrame ();
}

bool Star_ESCP2_Instance::
endJob ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "Star_ESCP2_Instance::endJob ()" << endl;
#endif

   pCmd = pCommands->getCommandData ("cmdPageEject");
   if (pCmd)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdPageEject = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   pCmd = pCommands->getCommandData ("cmdTerm");
   if (pCmd)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdTerm = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   return true;
}

bool Star_ESCP2_Instance::
abortJob ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "Star_ESCP2_Instance::endJob ()" << endl;
#endif

   int cx = getCurrentForm ()->getHardCopyCap ()->getXPels ();

   // The largest size is 24 lines
   int iSize = 24 * (cx + 7) / 8;

   PBYTE abData = new BYTE [iSize];

   memset (abData, 0, sizeof (abData));

   BinaryData data (abData, sizeof (abData));

   pDevice_d->sendBinaryDataToDevice (&data);

   delete []abData;

   return true;
}

string Star_ESCP2_Instance::
toString (ostrstream& oss)
{
   oss << hex
       << "{ "
       << DeviceInstance::toString (oss) // Add parent's output as well
       << " }"
       << dec
       << ends;

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const Star_ESCP2_Instance& const_self)
{
   Star_ESCP2_Instance& self = const_cast<Star_ESCP2_Instance&>(const_self);
   ostrstream            oss;

   os << self.toString (oss);

   return os;
}
