/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "UPDF_Blitter.hpp"
#include "UPDF_Instance.hpp"
#include "DeviceOrientation.hpp"
#include <CMYKbitmap.hpp>

UPDF_Blitter::
UPDF_Blitter (Device *pDevice)
   : DeviceBlitter (pDevice)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputBlitter ()) cerr << "UPDF_Blitter::" << __FUNCTION__ << " ()" << endl;
#endif

   fHaveInitialized_d      = false;
   fGraphicsHaveBeenSent_d = false;
}

UPDF_Blitter::
~UPDF_Blitter ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputBlitter ()) cerr << "UPDF_Blitter::~" << __FUNCTION__ << " () enter" << endl;
#endif

   fGraphicsHaveBeenSent_d = false;

#ifndef RETAIL
   if (DebugOutput::shouldOutputBlitter ()) cerr << "UPDF_Blitter::~" << __FUNCTION__ << " () exit" << endl;
#endif
}

void UPDF_Blitter::
initializeInstance ()
{
   HardCopyCap      *pHCC = getCurrentForm ()->getHardCopyCap ();
   DeviceResolution *pDR  = getCurrentResolution ();
   DevicePrintMode  *pDPM = getCurrentPrintMode ();

   char *pszDitherID = getCurrentDitherID ();

   if (  pDPM->isID (DevicePrintMode::PRINT_MODE_24_CMY)
      || pDPM->isID (DevicePrintMode::PRINT_MODE_24_CMYK)
      || pDPM->isID (DevicePrintMode::PRINT_MODE_8_CMY)
      || pDPM->isID (DevicePrintMode::PRINT_MODE_8_CMYK)
      )
   {
      int  iNumDstRowBytes8      = (pHCC->getXPels () + 7) >> 3;
      char achDitherOptions[512]; // @TBD

      sprintf (achDitherOptions,
               "fDataInRGB=true "
               "iBlackReduction=%d "
               "iColorTech=%d "
               "iNumDitherRows=%d "
               "iSrcRowPels=%d "
               "iNumDestRowBytes=%d "
               "iDestBitsPerPel=%d",
               /* @TBD iBlackReduction*/0,
               pDPM->getColorTech (),
               pDR->getScanlineMultiple (),
               pHCC->getXPels (),
               iNumDstRowBytes8,
               pDR->getDstBitsPerPel ());

      setDitherInstance (DeviceDither::createDitherInstance (pszDitherID,                 // iDitherType
                                                             pDevice_d,                   // pDevice
                                                             achDitherOptions));          // pszOptions
   }
}

bool UPDF_Blitter::
rasterize (PBYTE        pbBits,
           PBITMAPINFO2 pbmi2,
           PRECTL       prectlPageLocation,
           BITBLT_TYPE  eType)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputBlitter ())
      cerr << hex
           << "UPDF_Blitter::rasterize (0x"
           << (int)pbBits << ", {"
           << dec << pbmi2->cx << ", "
           << pbmi2->cy << ", "
           << pbmi2->cPlanes << ", "
           << pbmi2->cBitCount << "}, "
           << "{" << prectlPageLocation->xLeft << ", " << prectlPageLocation->yBottom << ", " << prectlPageLocation->xRight << ", " << prectlPageLocation->yTop << "})"
           << endl;
#endif

   UPDF_Instance *pInstance = dynamic_cast <UPDF_Instance *>(getInstance ());
   if (!pInstance)
      return false;

   pInstance->setupPrinter ();

   switch (pDevice_d->getCurrentPrintMode ()->getColorTech ())
   {
   case DevicePrintMode::COLOR_TECH_K:
   {
      return updfMonoRasterize (pbBits,
                                 pbmi2,
                                 prectlPageLocation,
                                 eType);
      break;
   }

   case DevicePrintMode::COLOR_TECH_CMYK:
   case DevicePrintMode::COLOR_TECH_CMY:
   {
      return updfColorRasterize (pbBits,
                                  pbmi2,
                                  prectlPageLocation,
                                  eType);
      break;
   }

   default:
#ifndef RETAIL
      if (DebugOutput::shouldOutputBlitter ()) cerr << "UPDF_Blitter::rasterize Error: unknown color tech " << pDevice_d->getCurrentPrintMode ()->getColorTech () << endl;
#endif
      break;
   }

   return true;
}

bool UPDF_Blitter::
updfMonoRasterize (PBYTE        pbBits,
                    PBITMAPINFO2 pbmi2,
                    PRECTL       prectlPageLocation,
                    BITBLT_TYPE  eType)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputBlitter ()) cerr << "UPDF_Blitter::" << __FUNCTION__ << " ()" << endl;
#endif

   UPDF_Instance *pInstance = dynamic_cast <UPDF_Instance *>(getInstance ());
   if (!pInstance)
      return false;

   char      *pszDumpEnvironmentVar = getenv ("DUMP_OUTGOING_BITMAPS");
   bool       fDumpOutgoingBitmaps  = false;
   static int iNum = 0;
   char       achName[4 + 1 + 3 + 1];

   sprintf (achName, "%04dOUT.bmp", iNum++);

   CMYKBitmap outgoingBitmap (achName, pbmi2->cx, pbmi2->cy);

   if (pszDumpEnvironmentVar)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputBlitter ()) cerr << hex << "UPDFP2_Blitter::updfMonoRasterize (out)pszDumpEnvironmentVar = " << (int)pszDumpEnvironmentVar << dec << endl;
#endif

      if (*pszDumpEnvironmentVar)
         fDumpOutgoingBitmaps = true;
   }

   pInstance->executeEvent ("Event_RasterGraphicBand", true);

   return true;
}

bool UPDF_Blitter::
updfColorRasterize (PBYTE        pbBits,
                     PBITMAPINFO2 pbmi2,
                     PRECTL       prectlPageLocation,
                     BITBLT_TYPE  eType)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputBlitter ()) cerr << "UPDF_Blitter::" << __FUNCTION__ << " ()" << endl;
#endif

   UPDF_Instance *pInstance = dynamic_cast <UPDF_Instance *>(getInstance ());
   if (!pInstance)
      return false;

   char      *pszDumpEnvironmentVar = getenv ("DUMP_OUTGOING_BITMAPS");
   bool       fDumpOutgoingBitmaps  = false;
   static int iNum = 0;
   char       achName[4 + 1 + 3 + 1];

   sprintf (achName, "%04dOUT.bmp", iNum++);

   CMYKBitmap outgoingBitmap (achName, pbmi2->cx, pbmi2->cy);

   if (pszDumpEnvironmentVar)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputBlitter ()) cerr << hex << "UPDF_Blitter::updfColorRasterize (out)pszDumpEnvironmentVar = " << (int)pszDumpEnvironmentVar << dec << endl;
#endif

      if (*pszDumpEnvironmentVar)
         fDumpOutgoingBitmaps = true;
   }

   return true;
}

/****************************************************************************/
/* NOTE: assumes iWorldY increases                                          */
/****************************************************************************/
bool UPDF_Blitter::
moveToYPosition (int  iWorldY,
                 bool fAbsolute)
{
   UPDF_Instance *pInstance = dynamic_cast <UPDF_Instance *>(getInstance ());
   if (!pInstance)
      return false;

   return true;
}

string UPDF_Blitter::
toString (ostrstream& oss)
{
   oss << "{ "
       << DeviceBlitter::toString (oss) // Add parent's output as well
       << " }"
       << ends;

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const UPDF_Blitter& const_self)
{
   UPDF_Blitter& self = const_cast<UPDF_Blitter&>(const_self);
   ostrstream    oss;

   os << self.toString (oss);

   return os;
}
