/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "UPDF_Instance.hpp"
#include "defines.hpp"

#include <stdio.h>
#include <string.h>

#include <iostream>
#include <strstream>

const static bool fTestNoCompression = true;

UPDF_Instance::
UPDF_Instance (Device *pDevice)
   : DeviceInstance (pDevice)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "UPDF_Instance::UPDF_Instance ()" << endl;
#endif

   fHaveInitialized_d  = false;
   fHaveSetupPrinter_d = false;
   ptlPrintHead_d.x    = 0;
   ptlPrintHead_d.y    = 0;
}

UPDF_Instance::
~UPDF_Instance ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "UPDF_Instance::~UPDF_Instance ()" << endl;
#endif

   for ( SYMBOLTABLE::iterator next = symbolTable.begin () ;
         next != symbolTable.end () ;
         next++ )
   {
      delete (*next).second;
   }
}

void UPDF_Instance::
initializeInstance ()
{
   if (fHaveInitialized_d)
      return;

   fHaveInitialized_d = true;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "UPDF_Instance::initializeInstance ()" << endl;
#endif
}

static char *apszDeviceJobPropertyKeys[] = {
   "RandomAccessMemory.RAM_KB",
   "DeviceResolution.HorizontalResolution",
   "DeviceResolution.VerticalResolution",
   "RasterGraphic.HorizontalResolution",
   "RasterGraphic.VerticalResolution",
   "Positioning.HorizontalResolution",
   "Positioning.VerticalResolution",
   "MediaCopy.Media_Copy_ID",
   "MediaCollation.MediaCollation_Global_ID",
   "MediaCollation.Media_Proprietary_ID",
   "MediaPageOrientation.MediaPageOrientation_Global_ID",
   "MediaSize.MediaSize_Global_ID",
   "MediaSize.Media_Proprietary_ID",
   "MediaType.MediaType_Global_ID",
   "MediaType.Media_Proprietary_ID",
   "MediaSource.MediaSource_Global_ID",
   "MediaSource.Media_Proprietary_ID",
   "MediaDestination.MediaDestination_Global_ID",
   "MediaDestination.Media_Proprietary_ID",
   "MediaCoating.MediaCoating_Global_ID",
   "MediaCoating.Media_Proprietary_ID",
   "MediaColor.MediaColor_Global_ID",
   "MediaColor.Media_Proprietary_ID",
   "MediaWeight.MediaWeight_Global_ID",
   "MediaWeight.Media_Proprietary_ID",
   "MediaStaplingSet.Media_Proprietary_ID",
   "MediaPunchingSet.Media_Proprietary_ID",
   "MediaJogging.MediaJogging_Global_ID",
   "MediaJogging.Media_Proprietary_ID",
   "MediaDuplex.MediaDuplex_Global_ID",
   "MediaDuplex.Media_Proprietary_ID",
   "MediaBooklet.MediaBooklet_Global_ID",
   "MediaBooklet.Media_Proprietary_ID",
   "MediaPageOrder.MediaPageOrder_Global_ID",
   "MediaPageOrder.Media_Proprietary_ID",
   "MediaNUp.MediaNUp_HorizontalPages",
   "MediaNUp.MediaNUp_VerticalPages",
   "MediaNUpBorder.MediaNUpBorder_Global_ID",
   "MediaNUpBorder.Media_Proprietary_ID",
   "MediaNUpPageOrder.MediaNUpPageOrder_Global_ID",
   "MediaNUpPageOrder.Media_Proprietary_ID",
   "MediaTargetSize.MediaTargetSize_Global_ID",
   "MediaTargetSize.Media_Proprietary_ID",
   "MediaScaling.MediaScaling_Global_ID",
   "MediaScaling.Media_Proprietary_ID",
   "MediaPercentage.Media_Percentage_ID"
};

Enumeration * UPDF_Instance::
listDeviceJobPropertyKeys ()
{
   class DeviceJobPropertyKeyEnumerator : public Enumeration
   {
   public:
      DeviceJobPropertyKeyEnumerator (char **apszKeys, int iDimOfKeys)
      {
         iWhichKey_d = iDimOfKeys;
         apszKeys_d  = apszKeys;
      }

      virtual bool hasMoreElements ()
      {
         return iWhichKey_d > 0;
      }

      virtual void * nextElement ()
      {
         if (0 < iWhichKey_d)
         {
            iWhichKey_d--;
            return apszKeys_d[iWhichKey_d];
         }

         return 0;
      }

   private:
      int    iWhichKey_d;
      char **apszKeys_d;
   };

   return new DeviceJobPropertyKeyEnumerator (apszDeviceJobPropertyKeys, dimof (apszDeviceJobPropertyKeys));
}

bool UPDF_Instance::
handleDeviceJobProperty (char *pszKey,
                         char *pszValue)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "UPDF_Instance::handleDeviceJobProperty ("
                          << pszKey << ", "
                          << pszValue << ")"
                          << endl;
#endif

   for (int i = 0; i < (int)dimof (apszDeviceJobPropertyKeys); i++)
   {
      if (0 == strcmp (pszKey, apszDeviceJobPropertyKeys[i]))
      {
         symbolTable[string (pszKey)] = new Expression (pszValue, true);
      }
   }

   return false;
}

string * UPDF_Instance::
getJobPropertyType (char *pszKey)
{
   // @TBD
   return 0;
}

string * UPDF_Instance::
getJobProperty (char *pszKey)
{
   Expression *pExpressionValue = 0;
   string     *pStringReturn    = 0;

   pExpressionValue = symbolTable[string (pszKey)];

   if (pExpressionValue)
   {
      switch (pExpressionValue->getType ())
      {
      case TYPE_STRING:
      {
         pStringReturn = new string (pExpressionValue->getString ());
         break;
      }

      case TYPE_HEX:
      {
         // @TBD
/////////pExpressionValue->getBinaryData ()
         break;
      }
      }
   }

   return pStringReturn;
}

void UPDF_Instance::
setupPrinter ()
{
   if (fHaveSetupPrinter_d)
      return;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "UPDF_Instance::setupPrinter ()" << endl;
#endif

   fHaveSetupPrinter_d = true;
}

bool UPDF_Instance::
executeEvent (char *pszPrefix, bool fPreEvent)
{
   char        achParameterName[128]; // @TBD
   DeviceData *pData                 = getDeviceData ();
   bool        fExecute              = fPreEvent;

   if (!pData)
      return true;

   int iCount = 0;

   sprintf (achParameterName, "%s_count", pszPrefix);

   if (  !pData->getIntData (achParameterName, &iCount)
      || 0 == iCount
      )
      return true;

   for (int i = 0; i < iCount; i++)
   {
      BinaryData *pbdExpression = 0;
      char       *pszEventName  = 0;

      sprintf (achParameterName, "%s_binary_%d", pszPrefix, i);

      if (pData->getBinaryData (achParameterName, &pbdExpression))
      {
      }
      else if (pData->getStringData (achParameterName, &pszEventName))
      {
         fExecute = !fExecute;
      }
      else
      {
      }
   }

   return true;
}

bool UPDF_Instance::
beginJob ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "UPDF_Instance::beginJob ()" << endl;
#endif

   return executeEvent ("Event_PDL_Start", true);
}

bool UPDF_Instance::
beginJob (bool fJobPropertiesChanged)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "UPDF_Instance::beginJob (with props)" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The job properties are as follows:" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The orientation is = " << getCurrentOrientation() << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "iDitherID = " << getCurrentDitherID () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The form is = " << getCurrentForm () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The tray is = " << getCurrentTray () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The media is = " << getCurrentMedia () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The resolution is = " << getCurrentResolution () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The print mode is = " << getCurrentPrintMode () << endl;
#endif

   // @TBD - reinitialize with new job properties

   // Call common code
   return beginJob ();
}

bool UPDF_Instance::
newFrame ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "UPDF_Instance::newFrame ()" << endl;
#endif

   ditherNewFrame ();

   ptlPrintHead_d.x = 0;
   ptlPrintHead_d.y = 0;

   return true;
}

bool UPDF_Instance::
newFrame (bool fJobPropertiesChanged)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "UPDF_Instance::newFrame (with props)" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The job properties are as follows:" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The orientation is = " << getCurrentOrientation() << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "iDitherID = " << getCurrentDitherID () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The form is = " << getCurrentForm () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The tray is = " << getCurrentTray () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The media is = " << getCurrentMedia () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The resolution is = " << getCurrentResolution () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The print mode is = " << getCurrentPrintMode () << endl;
#endif

   // @TBD - reinitialize with new job properties

   // Call common code
   return newFrame ();
}

bool UPDF_Instance::
endJob ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "UPDF_Instance::endJob ()" << endl;
#endif

   return true;
}

bool UPDF_Instance::
abortJob ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "UPDF_Instance::endJob ()" << endl;
#endif

   return true;
}

string UPDF_Instance::
toString (ostrstream& oss)
{
   oss << "{ "
       << DeviceInstance::toString (oss) // Add parent's output as well
       << " }"
       << ends;

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const UPDF_Instance& const_self)
{
   UPDF_Instance& self = const_cast<UPDF_Instance&>(const_self);
   ostrstream     oss;

   os << self.toString (oss);

   return os;
}
