/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <fstream.h>
#include <ctype.h>
#include <stdio.h>
#include <sys/stat.h>

#include "OmniDomParser.hpp"
#include "MyErrorHandler.hpp"

const static bool fDebugOutput = false;

void convertFilename (char *pszFileName);

extern char *vpszExeName;

OmniDomParser::
OmniDomParser (const char *pszXMLFile, DeviceInfo *pdi)
{
   pdi_d             = pdi;
   pstrXMLFile_d     = new string (pszXMLFile);
   fErrorCondition_d = false;
   doc_d             = 0;
}

OmniDomParser::
~OmniDomParser ()
{
   delete pstrXMLFile_d;

   if (doc_d)
   {
      xmlFreeDoc (doc_d);
   }
}

void OmniDomParser::
setErrorCondition ()
{
   fErrorCondition_d = true;
}

void
convertFilename (char *pszFileName)
{
   while (*pszFileName)
   {
      switch (*pszFileName)
      {
      case '-':
      case '+':
      case '(':
      case ')':
      case ' ':
      case '/':
      case '\\':
         *pszFileName = '_';
      }

      pszFileName++;
   }
}

string * OmniDomParser::
getNameOut ()
{
   int     iDotPos    = pstrXMLFile_d->find_last_of ('.');
   string *pstringRet = new string (pstrXMLFile_d->substr (0, iDotPos));

   convertFilename (const_cast<char*>(pstringRet->c_str ())); // @HACK?

   return pstringRet;
}

string *
converToClassName (const char *pszFileName)
{
   string  stringFileName (pszFileName);

   int     iDotPos    = stringFileName.find_last_of ('.');
   string *pstringRet = new string (stringFileName.substr (0, iDotPos));

   convertFilename (const_cast<char*>(pstringRet->c_str ())); // @HACK?

   return pstringRet;
}

bool OmniDomParser::
shouldCreateFile (string strFileName)
{
   string stringFileNameHPP (strFileName.c_str ());
   string stringFileNameCPP (strFileName.c_str ());

   convertFilename (const_cast<char*>(stringFileNameHPP.c_str ())); // @HACK?
   convertFilename (const_cast<char*>(stringFileNameCPP.c_str ())); // @HACK?

   int iDotPos = stringFileNameHPP.find_last_of ('.');

   stringFileNameHPP.replace (iDotPos, 4, ".hpp");

   iDotPos = stringFileNameCPP.find_last_of ('.');

   stringFileNameCPP.replace (iDotPos, 4, ".cpp");

   struct stat statExe;
   struct stat statIn;
   struct stat statOutHPP;
   struct stat statOutCPP;

   if (-1 == stat (vpszExeName, &statExe))
   {
      cerr << "Error: File not found \"" << vpszExeName << "\" (could not stat it)!" << endl;
      setErrorCondition ();
      abort ();
   }

   if (-1 == stat (strFileName.c_str (), &statIn))
   {
      cerr << "Error: File not found \"" << strFileName << "\" (could not stat it)!" << endl;
      setErrorCondition ();
      abort ();
   }

   if (fDebugOutput)
   {
      cerr << "stats for \"" << strFileName << "\"" << endl;
      cerr << dec << "   atime = " << statIn.st_atime << endl;
      cerr        << "   mtime = " << statIn.st_mtime << endl;
      cerr        << "   ctime = " << statIn.st_ctime << endl;
   }

   if (-1 == stat (stringFileNameHPP.c_str (), &statOutHPP))
      return true;

   if (fDebugOutput)
   {
      cerr << "stats for \"" << stringFileNameHPP << "\"" << endl;
      cerr << dec << "   atime = " << statOutHPP.st_atime << endl;
      cerr        << "   mtime = " << statOutHPP.st_mtime << endl;
      cerr        << "   ctime = " << statOutHPP.st_ctime << endl;
   }

   if (-1 == stat (stringFileNameCPP.c_str (), &statOutCPP))
      return true;

   if (fDebugOutput)
   {
      cerr << "stats for \"" << stringFileNameCPP << "\"" << endl;
      cerr << dec << "   atime = " << statOutCPP.st_atime << endl;
      cerr        << "   mtime = " << statOutCPP.st_mtime << endl;
      cerr        << "   ctime = " << statOutCPP.st_ctime << endl;
   }

   if (  statExe.st_ctime > statOutHPP.st_ctime
      || statExe.st_ctime > statOutCPP.st_ctime
      || statIn.st_ctime > statOutHPP.st_ctime
      || statIn.st_ctime > statOutCPP.st_ctime
      || 0 == statOutHPP.st_size
      || 0 == statOutCPP.st_size
      )
      return true;
   else
      return false;
}

void OmniDomParser::
copyFile (const char *pszFileNameIn)
{
   string stringFileNameOut (pszFileNameIn);

   convertFilename (const_cast<char*>(stringFileNameOut.c_str ())); // @HACK?

   struct stat statIn;
   struct stat statOut;
   bool        fShouldCopy = false;

   if (-1 == stat (pszFileNameIn, &statIn))
   {
      cerr << "Error: File not found \"" << pszFileNameIn << "\" (could not stat it)!" << endl;
      setErrorCondition ();

      return;
   }

   if (-1 == stat (stringFileNameOut.c_str (), &statOut))
      fShouldCopy = true;

   if (  statIn.st_ctime > statOut.st_ctime
       || 0 == statOut.st_size
      )
      fShouldCopy = true;

   if (fShouldCopy)
   {
      cerr << "copying \"" << pszFileNameIn << "\" to \"" << stringFileNameOut << "\"" << endl;

      ifstream  ifIn  (pszFileNameIn, ios::in | ios::binary);
      ofstream  ofOut (stringFileNameOut.c_str (), ios::out | ios::binary);

      // copy the file in one line o' code!
      ofOut << ifIn.rdbuf ();
   }
}

xmlNodePtr
nextNode (xmlNodePtr elm)
{
   do
   {
      // Move to the next node
      elm = elm->next;

   } while (  0 !=  elm
           && XML_COMMENT_NODE == elm->type
           );

   return elm;
}

void
printXMLDocument (xmlDocPtr doc, int iLevel)
{
   xmlDocDump (stdout, doc);
}

void
printXMLNode (xmlNodePtr elm)
{
   xmlElemDump (stdout, elm->doc, elm);
   fprintf (stdout, "\n");
}

bool
openOutputFiles (string    *pstringName,
                 ofstream **ppofstreamHPP,
                 ofstream **ppofstreamCPP)
{
   char achFilename[512]; //@TBD

   strcpy (achFilename, pstringName->c_str ());
   strcat (achFilename, ".hpp");
   convertFilename (achFilename);

   *ppofstreamHPP = new ofstream (achFilename);

   strcpy (achFilename, pstringName->c_str ());
   strcat (achFilename, ".cpp");
   convertFilename (achFilename);

   *ppofstreamCPP = new ofstream (achFilename);

   return true;
}

bool
openMakeFile (string    *pstringName,
              ofstream **ppofstreamMAK)
{
   char achFilename[512]; //@TBD

   strcpy (achFilename, pstringName->c_str ());
   strcat (achFilename, ".mak");
   convertFilename (achFilename);

   *ppofstreamMAK = new ofstream (achFilename);

   return true;
}

void
outputHeader (ofstream *pofstream)
{
   *pofstream << "/*" << endl;
   *pofstream << " *   IBM Omni driver" << endl;
   *pofstream << " *   Copyright (c) International Business Machines Corp., 2000" << endl;
   *pofstream << " *" << endl;
   *pofstream << " *   This library is free software; you can redistribute it and/or modify" << endl;
   *pofstream << " *   it under the terms of the GNU Lesser General Public License as published" << endl;
   *pofstream << " *   by the Free Software Foundation; either version 2.1 of the License, or" << endl;
   *pofstream << " *   (at your option) any later version." << endl;
   *pofstream << " *" << endl;
   *pofstream << " *   This library is distributed in the hope that it will be useful," << endl;
   *pofstream << " *   but WITHOUT ANY WARRANTY; without even the implied warranty of" << endl;
   *pofstream << " *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See" << endl;
   *pofstream << " *   the GNU Lesser General Public License for more details." << endl;
   *pofstream << " *" << endl;
   *pofstream << " *   You should have received a copy of the GNU Lesser General Public License" << endl;
   *pofstream << " *   along with this library; if not, write to the Free Software" << endl;
   *pofstream << " *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA" << endl;
   *pofstream << " */" << endl;
}

char *
bundleStringData (xmlNodePtr nodeItem)
{
   return (char *)xmlNodeListGetString (nodeItem->doc, nodeItem->childs, 1);
}

int
bundleIntegerData (xmlNodePtr nodeItem)
{
   char *pszData = (char *)xmlNodeListGetString (nodeItem->doc, nodeItem->childs, 1);
   int   iRet    = 0;

   if (1 != sscanf (pszData, "%d", &iRet))
   {
      iRet = 0;
   }

   free (pszData);

   return iRet;
}

void
printNodeName (xmlNodePtr cur)
{
   const char *pszNodeName = 0;

   if (!fDebugOutput)
      return;

   pszNodeName = bundleStringData (cur);

   cerr << "Node <"
        << cur->name
        << " "
        << pszNodeName
        << ">"
        << endl;

   if (pszNodeName)
   {
      free ((char *)pszNodeName);
   }
}

bool
parseHexGroup (const char   *pszData,
               unsigned int *puiData)
{
   if (  isxdigit (pszData[0])
      && isxdigit (pszData[1])
      )
   {
      return 1 == sscanf (pszData, "%x", puiData);
   }
   else
   {
      return false;
   }
}

bool OmniDomParser::
parseBinaryData (const char  *pszData,
                 byte       **ppbData,
                 int         *pcbData)
{
   int   cbDataAlloc    = 64;
   int   cbDataLeft     = 64;
   byte *pbData         = new byte [cbDataAlloc];
   byte *pbDataCurrent  = pbData;
   bool  fContinue      = true;
   bool  fInString      = false;
   byte  bChar          = 0;
   byte  bNextChar      = 0;
   bool  fCharFound     = false;
   bool  fNextCharFound = false;

   *ppbData = pbData;
   *pcbData = cbDataAlloc - cbDataLeft;

   while (  *pszData
         && fContinue
         )
   {
      if (fNextCharFound)
      {
         bChar          = bNextChar;
         fNextCharFound = false;
         fCharFound     = true;
      }
      else
      {
         if (fInString)
         {
            switch (*pszData)
            {
            case '\\':
            {
               break;
            }

            case '"':
            {
               pszData++;

               fInString = false;
               continue;
            }

            default:
            {
               bChar      = *pszData++;
               fCharFound = true;
               break;
            }
            }
         }
         else
         {
            while (  '\x0a' == *pszData
                  || '\x0d' == *pszData
                  || ' '    == *pszData
                  )
               pszData++;

            switch (*pszData)
            {
            case '_':
            {
               pszData++;

               if (0 == strncmp (pszData, "ESC_", 4))
               {
                  bChar      = 0x1b;
                  fCharFound = true;
                  pszData += 4;
               }
               else if (0 == strncmp (pszData, "NUL_", 4))
               {
                  bChar      = 0;
                  fCharFound = true;
                  pszData += 4;
               }
               else if (0 == strncmp (pszData, "LF_", 3))
               {
                  bChar      = 0x0a;
                  fCharFound = true;
                  pszData += 3;
               }
               else if (0 == strncmp (pszData, "CR_", 3))
               {
                  bChar      = 0x0d;
                  fCharFound = true;
                  pszData += 3;
               }
               else if (0 == strncmp (pszData, "FF_", 3))
               {
                  bChar      = 0x0c;
                  fCharFound = true;
                  pszData += 3;
               }
               else
               {
                  cerr << "Error: Unknown define " << pszData << endl;
                  setErrorCondition ();
                  fContinue = false;
               }
               break;
            }

            case 'H':
            {
               pszData++;

               if (0 == strncmp (pszData, "EX2S", 4))
               {
                  pszData += 4;

                  while (' ' == *pszData)
                     pszData++;

                  if ('(' == *pszData)
                  {
                     pszData++;
                     while (' ' == *pszData)
                        pszData++;

                     unsigned int uiData  = 0;
                     unsigned int uiData2 = 0;

                     if (parseHexGroup (pszData, &uiData))
                     {
                        pszData += 2;

                        while (' ' == *pszData)
                           pszData++;

                        if (',' == *pszData)
                        {
                           pszData++;

                           while (' ' == *pszData)
                              pszData++;

                           if (parseHexGroup (pszData, &uiData2))
                           {
                              pszData += 2;

                              while (' ' == *pszData)
                                 pszData++;

                              if (')' == *pszData)
                              {
                                 pszData++;

                                 bChar          = (byte)uiData2;
                                 bNextChar      = (byte)uiData;
                                 fCharFound     = true;
                                 fNextCharFound = true;
                              }
                              else
                              {
                                 cerr << "Error: Expecting ')', found '" << *pszData << "' @ " << __LINE__ << endl;
                                 setErrorCondition ();
                                 fContinue = false;
                              }
                           }
                           else
                           {
                              cerr << "Error: Expecting ',', found '" << *pszData << "' @ " << __LINE__ << endl;
                              setErrorCondition ();
                              fContinue = false;
                           }
                        }
                        else
                        {
                           cerr << "Error: Expecting ',', found '" << *pszData << "' @ " << __LINE__ << endl;
                           setErrorCondition ();
                           fContinue = false;
                        }
                     }
                     else
                     {
                        cerr << "Error: Expecting 2 hex digits, found '" << pszData[0] << pszData[1] << "' @ " << __LINE__ << endl;
                        setErrorCondition ();
                        fContinue = false;
                     }
                  }
                  else
                  {
                     cerr << "Error: Expecting '(', found '" << *pszData << "' @ " << __LINE__ << endl;
                     setErrorCondition ();
                     fContinue = false;
                  }
               }
               else if (0 == strncmp (pszData, "EX", 2))
               {
                  pszData += 2;

                  while (' ' == *pszData)
                     pszData++;

                  if ('(' == *pszData)
                  {
                     pszData++;
                     while (' ' == *pszData)
                        pszData++;

                     unsigned int uiData = 0;

                     if (parseHexGroup (pszData, &uiData))
                     {
                        pszData += 2;

                        while (' ' == *pszData)
                           pszData++;

                        if (')' == *pszData)
                        {
                           pszData++;

                           bChar      = (byte)uiData;
                           fCharFound = true;
                        }
                        else
                        {
                           cerr << "Error: Expecting ')', found '" << *pszData << "' @ " << __LINE__ << endl;
                           setErrorCondition ();
                           fContinue = false;
                        }
                     }
                     else
                     {
                        cerr << "Error: Expecting 2 hex digits, found '" << pszData[0] << pszData[1] << "' @ " << __LINE__ << endl;
                        setErrorCondition ();
                        fContinue = false;
                     }
                  }
                  else
                  {
                     cerr << "Error: Expecting '(', found '" << *pszData << "' @ " << __LINE__ << endl;
                     setErrorCondition ();
                     fContinue = false;
                  }
               }
               break;
            }

            case 'A':
            {
               pszData++;

               if (0 == strncmp (pszData, "SCII", 4))
               {
                  pszData += 4;

                  while (' ' == *pszData)
                     pszData++;

                  if ('(' == *pszData)
                  {
                     pszData++;
                     while (' ' == *pszData)
                        pszData++;

                     bChar      = (byte)*pszData++;
                     fCharFound = true;

                     if (')' == *pszData)
                     {
                        pszData++;
                     }
                     else
                     {
                        cerr << "Error: Expecting ')', found '" << *pszData << "' @ " << __LINE__ << endl;
                        setErrorCondition ();
                        fContinue = false;
                     }
                  }
                  else
                  {
                     cerr << "Error: Expecting '(', found '" << *pszData << "' @ " << __LINE__ << endl;
                     setErrorCondition ();
                     fContinue = false;
                  }
               }
               break;
            }

            case '"':
            {
               pszData++;
               fInString = true;
               break;
            }
            }
         }
      }

      if (fCharFound)
      {
/////////if (isprint (bChar))
/////////   cerr << "bChar = '" << bChar << "'" << endl;
/////////else
/////////   cerr << "bChar = 0x" << hex << (int)bChar << endl;

         *pbDataCurrent++ = bChar;
         cbDataLeft--;

         bChar      = 0;
         fCharFound = false;

         if (0 == cbDataLeft)
         {
            byte *pbTemp = new byte [cbDataAlloc + 64];

            if (pbTemp)
            {
               memcpy (pbTemp, pbData, cbDataAlloc);

               delete[] pbData;

               pbData        = pbTemp;
               pbDataCurrent = pbData + cbDataAlloc;
               cbDataAlloc  += 64;
               cbDataLeft    = 64;
            }
            else
            {
               cerr << "Error: Allocation of " << (cbDataAlloc + 64) << " bytes failed @ " << __LINE__ << endl;
               setErrorCondition ();
               fContinue = false;
            }
         }
      }
      else if (  !fInString
              && *pszData
              )
      {
         delete[] pbData;
         return false;
      }
   }

   if (fContinue)
   {
      *ppbData = pbData;
      *pcbData = cbDataAlloc - cbDataLeft;
   }
   else
   {
      delete[] pbData;
   }

   return fContinue;
}

bool OmniDomParser::
processDeviceOrientations (xmlNodePtr deviceOrientationsNode)
{
   if (deviceOrientationsNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringOrientationClassName = pstringNameOut;

   typedef map <string, string *> OrientationMapElement;

   xmlNodePtr             deviceOrientationNode = deviceOrientationsNode->childs;
   ofstream              *pofstreamHPP          = 0;
   ofstream              *pofstreamCPP          = 0;
   OrientationMapElement  mapElement;
   int                    rc                    = false;

   if (deviceOrientationNode == 0)
      goto done;

   while (deviceOrientationNode != 0)
   {
      xmlNodePtr  nodeItem = deviceOrientationNode->childs;
      const char *pszId    = 0;
      int         id;

      if (nodeItem == 0)
         goto done;

      pszId = bundleStringData (nodeItem);
      if (pszId)
      {
         id = DeviceOrientation::nameToID (const_cast<char*>(pszId));
         if (-1 == id)
         {
            cerr << "Error: Unknown DeviceOrientation \"" << pszId << "\"" << endl;
            setErrorCondition ();
            goto done;
         }
      }

      pdi_d->listOrientations.push_back (string (pszId));

      ostrstream oss;

      oss << "   case "
          << pszId
          << ": return new "
          << *pstringNameOut
          << " ("
          << pszId
          << ");"
          << ends;

      mapElement[string (pszId)] = new string (oss.str ());

      if (fDebugOutput)
      {
         cerr << dec << "id              = " << id              << endl;
      }

      // Move to the next deviceOrientation node
      deviceOrientationNode = nextNode (deviceOrientationNode);

      if (pszId)
      {
         free ((char *)pszId);
      }
   }

   rc = true;

   if (shouldCreateFile (*pstrXMLFile_d))
   {
      cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

      openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

      outputHeader (pofstreamHPP);

      *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
      *pofstreamHPP << "#define _" << *pstringNameOut << endl;
      *pofstreamHPP << "#include <DeviceOrientation.hpp>" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceOrientation" << endl;
      *pofstreamHPP << "{" << endl;
      *pofstreamHPP << "public:" << endl;
      *pofstreamHPP << *pstringNameOut << " (int          id);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   static  DeviceOrientation  *create                 (int   id);" << endl;
      *pofstreamHPP << "   static  DeviceOrientation  *create                 (char *pszId);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual bool                isOrientationSupported (int   id);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual Enumeration        *getEnumeration         ();" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "protected:" << endl;
      *pofstreamHPP << "   virtual DeviceOrientation  *createV                (int   id);" << endl;

      *pofstreamHPP << "};" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#endif" << endl;

      outputHeader (pofstreamCPP);

      *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << *pstringNameOut << "::" << endl;
      *pofstreamCPP << *pstringNameOut << " (int id)" << endl;
      *pofstreamCPP << "      : DeviceOrientation (id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceOrientation * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "createV (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return " << *pstringNameOut << "::create (id);" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceOrientation * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (char *pszId)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return create (DeviceOrientation::nameToID (pszId));" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceOrientation * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   switch (id)" << endl;
      *pofstreamCPP << "   {" << endl;

      for ( OrientationMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++ )
      {
         string *pElement = (string *)next->second;

         *pofstreamCPP << *pElement << endl;
      }

      *pofstreamCPP << "   default: return 0;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << "static int aSupportedOrientations[] = {" << endl;

      int iCount = mapElement.size ();
      for ( OrientationMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << "   DeviceOrientation::" << next->first;
         if (1 < iCount)
            *pofstreamCPP << ",";
         *pofstreamCPP << endl;
      }

      *pofstreamCPP << "};" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "bool " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "isOrientationSupported (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   for (int i = 0; i < (int)dimof (aSupportedOrientations); i++)" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "      if (aSupportedOrientations[i] == id)" << endl;
      *pofstreamCPP << "         return true;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return false;" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "Enumeration * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "getEnumeration ()" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   class OrientationEnumerator : public Enumeration" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "   public:" << endl;
      *pofstreamCPP << "      OrientationEnumerator (int cOrientations, int aOrientations[])" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         iOrientation_d  = 0;" << endl;
      *pofstreamCPP << "         cOrientations_d = cOrientations;" << endl;
      *pofstreamCPP << "         aOrientations_d = aOrientations;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual bool hasMoreElements ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iOrientation_d < cOrientations_d)" << endl;
      *pofstreamCPP << "            return true;" << endl;
      *pofstreamCPP << "         else" << endl;
      *pofstreamCPP << "            return false;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual void *nextElement ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iOrientation_d > cOrientations_d - 1)" << endl;
      *pofstreamCPP << "            return 0;" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "         return create (aOrientations_d[iOrientation_d++]);" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   private:" << endl;
      *pofstreamCPP << "      int  iOrientation_d;" << endl;
      *pofstreamCPP << "      int  cOrientations_d;" << endl;
      *pofstreamCPP << "      int *aOrientations_d;" << endl;
      *pofstreamCPP << "   };" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return new OrientationEnumerator (dimof (aSupportedOrientations), aSupportedOrientations);" << endl;
      *pofstreamCPP << "}" << endl;
   }

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printXMLDocument (deviceOrientationsNode->doc, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringOrientationClassName = 0;
   }

   for ( OrientationMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      delete (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDeviceCommands (xmlNodePtr deviceCommandsNode)
{
   if (deviceCommandsNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringCommandClassName = pstringNameOut;

   if (!shouldCreateFile (*pstrXMLFile_d))
      return true;

   cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

   typedef map <string, BinaryData *> CommandMapElement;
   typedef map <string, byte *>       CommandMapData;

   xmlNodePtr        deviceCommandNode = deviceCommandsNode->childs;
   ofstream         *pofstreamHPP      = 0;
   ofstream         *pofstreamCPP      = 0;
   CommandMapElement mapElement;
   CommandMapData    mapData;
   int               rc                = false;

   if (deviceCommandNode == 0)
      goto done;

   while (deviceCommandNode != 0)
   {
      xmlChar *name = xmlGetProp (deviceCommandNode, (xmlChar *)"name");

      if (name == 0)
         goto done;

      const char *pszCmdName  = (char *)name;
      const char *pszCmdValue = 0;

      pszCmdValue = (char *)xmlNodeListGetString (deviceCommandNode->doc, deviceCommandNode->childs, 1);

      if (fDebugOutput) cerr << pszCmdName << " = " << pszCmdValue << endl;

      byte *pbData = 0;
      int   cbData = 0;

      if (parseBinaryData (pszCmdValue, &pbData, &cbData))
      {
         mapData[string (pszCmdName)]    = pbData;
         mapElement[string (pszCmdName)] = new BinaryData (pbData, cbData);
      }
      else
      {
         cerr << "Error: binary data not understood \"" << pszCmdValue << "\" for " << pszCmdName << "." << endl;
         goto done;
      }

      // Move to the next one
      deviceCommandNode = nextNode (deviceCommandNode);

      if (name)
      {
         free (name);
      }
      if (pszCmdValue)
      {
         free ((char *)pszCmdValue);
      }
   }

   rc = true;

   openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

   outputHeader (pofstreamHPP);

   *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
   *pofstreamHPP << "#define _" << *pstringNameOut << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "#include <DeviceCommand.hpp>" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceCommand" << endl;
   *pofstreamHPP << "{" << endl;
   *pofstreamHPP << "public:" << endl;
   *pofstreamHPP << "   " << *pstringNameOut << " ();" << endl;
   *pofstreamHPP << "};" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "#endif" << endl;

   outputHeader (pofstreamCPP);

   *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << *pstringNameOut << "::" << endl;
   *pofstreamCPP << *pstringNameOut << "()" << endl;
   *pofstreamCPP << "{" << endl;

   for ( CommandMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      BinaryData *bdValue = (BinaryData *)next->second;
      PBYTE       pbValue = bdValue->getData ();
      int         cbValue = bdValue->getLength ();

      *pofstreamCPP << "   static byte " << next->first << "[] = {" << dec;

      for (int i = 0; i < cbValue; i++)
      {
         *pofstreamCPP << (int)pbValue[i];
         if (i < cbValue - 1)
            *pofstreamCPP << ",";
      }

      *pofstreamCPP << "};" << endl;

//////if (fDebugOutput) cerr << next->first << " = " << *(next->second) << endl;
   }

   *pofstreamCPP << endl;

   for ( CommandMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      *pofstreamCPP << "   add (\""
                    << next->first
                    << "\", new BinaryData ("
                    << next->first
                    << ", dimof ("
                    << next->first << ")));"
                    << endl;
   }

   *pofstreamCPP << "}" << endl;

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printXMLDocument (deviceCommandsNode->doc, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringCommandClassName = 0;
   }

   for ( CommandMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( CommandMapData::iterator next = mapData.begin () ;
         next != mapData.end () ;
         next++ )
   {
      delete[] (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDeviceResolutions (xmlNodePtr deviceResolutionsNode)
{
   if (deviceResolutionsNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringResolutionClassName = pstringNameOut;

   typedef map <string, string *>     ResolutionMapElement;
   typedef map <string, BinaryData *> ResolutionMapData1;
   typedef map <string, byte *>       ResolutionMapData2;

   xmlNodePtr           deviceResolutionNode = deviceResolutionsNode->childs;
   ofstream            *pofstreamHPP         = 0;
   ofstream            *pofstreamCPP         = 0;
   ResolutionMapElement mapResolution;
   ResolutionMapData1   mapData1;
   ResolutionMapData2   mapData2;
   int                  iCount               = 0;
   int                  rc                   = false;

   if (deviceResolutionNode == 0)
      goto done;

   while (deviceResolutionNode != 0)
   {
      xmlNodePtr  nodeItem               = deviceResolutionNode->childs;
      const char *pszId                  = 0;
      int         id;
      int         iXRes;
      int         iYRes;
      int         iXInternalRes          = 0;
      int         iYInternalRes          = 0;
      const char *pszBinaryData          = 0;
      byte       *pbData                 = 0;
      int         cbData                 = 0;
      BinaryData *pdbData;
      int         iCapabilities;
      int         iDestinationBitsPerPel;
      int         iScanlineMultiple;

      if (nodeItem == 0)
         goto done;

      pszId = bundleStringData (nodeItem);
      if (pszId)
      {
         id = DeviceResolution::nameToID (const_cast<char*>(pszId));
         if (-1 == id)
         {
            cerr << "Error: Unknown DeviceResolution \"" << pszId << "\"" << endl;
            setErrorCondition ();
            goto done;
         }
      }

      pdi_d->listResolutions.push_back (string (pszId));

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iXRes = bundleIntegerData (nodeItem);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iYRes = bundleIntegerData (nodeItem);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      if (0 == strcmp ((const char *)nodeItem->name, "xInternalRes"))
      {
         iXInternalRes = bundleIntegerData (nodeItem);

         // Move to the next one
         nodeItem = nextNode (nodeItem);
         if (nodeItem == 0)
            goto done;

         iYInternalRes = bundleIntegerData (nodeItem);

         // Move to the next one
         nodeItem = nextNode (nodeItem);
         if (nodeItem == 0)
            goto done;
      }

      pszBinaryData = bundleStringData (nodeItem);
      if (parseBinaryData (pszBinaryData, &pbData, &cbData))
      {
         pdbData = new BinaryData (pbData, cbData);
      }
      else
      {
         cerr << "Error: binary data not understood \"" << pszBinaryData << "\" for " << pszId << "." << endl;
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iCapabilities = bundleIntegerData (nodeItem);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iDestinationBitsPerPel = bundleIntegerData (nodeItem);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iScanlineMultiple = bundleIntegerData (nodeItem);

      ostrstream oss;

      oss << "   case "
          << pszId
          << ": return new "
          << *pstringNameOut
          << " ("
          << pszId
          << ","
          << iXRes
          << ","
          << iYRes
          << ","
          << iXInternalRes
          << ","
          << iYInternalRes
          << ",new BinaryData (abData"
          << pszId
          << ",dimof (abData"
          << pszId
          << ")),"
          << iCapabilities
          << ","
          << iDestinationBitsPerPel
          << ","
          << iScanlineMultiple
          << ");"
          << ends;

      mapData2[string (pszId)]      = pbData;
      mapData1[string (pszId)]      = new BinaryData (pbData, cbData);
      mapResolution[string (pszId)] = new string (oss.str ());

      if (fDebugOutput)
      {
         cerr << dec << "id                     = " << id                     << endl;
         cerr        << "iXRes                  = " << iXRes                  << endl;
         cerr        << "iYRes                  = " << iYRes                  << endl;
         cerr        << "iXInternalRes          = " << iXInternalRes          << endl;
         cerr        << "iYInternalRes          = " << iYInternalRes          << endl;
         cerr        << "pszBinaryData          = " << pszBinaryData          << endl;
         cerr        << "pdbData                = " << *pdbData               << endl;
         cerr << dec << "iCapabilities          = " << iCapabilities          << endl;
         cerr        << "iDestinationBitsPerPel = " << iDestinationBitsPerPel << endl;
         cerr        << "iScanlineMultiple      = " << iScanlineMultiple      << endl;
      }

      // Move to the next deviceResolution node
      deviceResolutionNode = nextNode (deviceResolutionNode);

      if (pszId)
      {
         free ((char *)pszId);
      }
      if (pszBinaryData)
      {
         free ((char *)pszBinaryData);
      }
   }

   pdi_d->listResolutions.sort ();
   pdi_d->listResolutions.unique ();

   rc = true;

   if (shouldCreateFile (*pstrXMLFile_d))
   {
      cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

      openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

      outputHeader (pofstreamHPP);

      *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
      *pofstreamHPP << "#define _" << *pstringNameOut << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#include <BinaryData.hpp>" << endl;
      *pofstreamHPP << "#include <DeviceResolution.hpp>" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceResolution" << endl;
      *pofstreamHPP << "{" << endl;
      *pofstreamHPP << "public:" << endl;
      *pofstreamHPP << *pstringNameOut << " (int          id," << endl;
      *pofstreamHPP << "                          int         iXRes," << endl;
      *pofstreamHPP << "                          int         iYRes," << endl;
      *pofstreamHPP << "                          int         iXInternalRes," << endl;
      *pofstreamHPP << "                          int         iYInternalRes," << endl;
      *pofstreamHPP << "                          BinaryData *data," << endl;
      *pofstreamHPP << "                          int         iCapabilities," << endl;
      *pofstreamHPP << "                          int         iDestinationBitsPerPel," << endl;
      *pofstreamHPP << "                          int         iScanlineMultiple);" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "   static  DeviceResolution *create                (int   id);" << endl;
      *pofstreamHPP << "   static  DeviceResolution *create                (char *pszId);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual bool              isResolutionSupported (int   id);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual Enumeration      *getEnumeration        ();" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "protected:" << endl;
      *pofstreamHPP << "   virtual DeviceResolution *createV               (int   id);" << endl;

      *pofstreamHPP << "};" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#endif" << endl;

      outputHeader (pofstreamCPP);

      *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << *pstringNameOut << "::" << endl;
      *pofstreamCPP << *pstringNameOut << " (int          id," << endl;
      *pofstreamCPP << "                          int         iXRes," << endl;
      *pofstreamCPP << "                          int         iYRes," << endl;
      *pofstreamCPP << "                          int         iXInternalRes," << endl;
      *pofstreamCPP << "                          int         iYInternalRes," << endl;
      *pofstreamCPP << "                          BinaryData *data," << endl;
      *pofstreamCPP << "                          int         iCapabilities," << endl;
      *pofstreamCPP << "                          int         iDestinationBitsPerPel," << endl;
      *pofstreamCPP << "                          int         iScanlineMultiple)" << endl;
      *pofstreamCPP << "      : DeviceResolution (id," << endl;
      *pofstreamCPP << "                          iXRes," << endl;
      *pofstreamCPP << "                          iYRes," << endl;
      *pofstreamCPP << "                          iXInternalRes," << endl;
      *pofstreamCPP << "                          iYInternalRes," << endl;
      *pofstreamCPP << "                          data," << endl;
      *pofstreamCPP << "                          iCapabilities," << endl;
      *pofstreamCPP << "                          iDestinationBitsPerPel," << endl;
      *pofstreamCPP << "                          iScanlineMultiple)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceResolution * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "createV (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return " << *pstringNameOut << "::create (id);" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceResolution * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (char *pszId)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return create (DeviceResolution::nameToID (pszId));" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceResolution * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (int id)" << endl;
      *pofstreamCPP << "{" << endl;

      for ( ResolutionMapData1::iterator next = mapData1.begin () ;
            next != mapData1.end () ;
            next++ )
      {
         BinaryData *bdValue = (BinaryData *)next->second;
         PBYTE       pbValue = bdValue->getData ();
         int         cbValue = bdValue->getLength ();

         *pofstreamCPP << "   static byte abData" << next->first << "[] = {" << dec;

         for (int i = 0; i < cbValue; i++)
         {
            *pofstreamCPP << (int)pbValue[i];
            if (i < cbValue - 1)
               *pofstreamCPP << ",";
         }

         *pofstreamCPP << "};" << endl;
      }

      *pofstreamCPP << endl;
      *pofstreamCPP << "   switch (id)" << endl;
      *pofstreamCPP << "   {" << endl;

      for ( ResolutionMapElement::iterator next = mapResolution.begin () ;
            next != mapResolution.end () ;
            next++ )
      {
         string *pElement = (string *)next->second;

         *pofstreamCPP << *pElement << endl;
      }

      *pofstreamCPP << "   default: return 0;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << "static int aSupportedResolutions[] = {" << endl;

      iCount = mapResolution.size ();
      for ( ResolutionMapElement::iterator next = mapResolution.begin () ;
            next != mapResolution.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << "   DeviceResolution::" << next->first;
         if (1 < iCount)
            *pofstreamCPP << ",";
         *pofstreamCPP << endl;
      }

      *pofstreamCPP << "};" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "bool " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "isResolutionSupported (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   for (int i = 0; i < (int)dimof (aSupportedResolutions); i++)" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "      if (aSupportedResolutions[i] == id)" << endl;
      *pofstreamCPP << "         return true;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return false;" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "Enumeration * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "getEnumeration ()" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   class ResolutionEnumerator : public Enumeration" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "   public:" << endl;
      *pofstreamCPP << "      ResolutionEnumerator (int cResolutions, int aResolutions[])" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         iResolution_d  = 0;" << endl;
      *pofstreamCPP << "         cResolutions_d = cResolutions;" << endl;
      *pofstreamCPP << "         aResolutions_d = aResolutions;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual bool hasMoreElements ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iResolution_d < cResolutions_d)" << endl;
      *pofstreamCPP << "            return true;" << endl;
      *pofstreamCPP << "         else" << endl;
      *pofstreamCPP << "            return false;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual void *nextElement ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iResolution_d > cResolutions_d - 1)" << endl;
      *pofstreamCPP << "            return 0;" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "         return create (aResolutions_d[iResolution_d++]);" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   private:" << endl;
      *pofstreamCPP << "      int  iResolution_d;" << endl;
      *pofstreamCPP << "      int  cResolutions_d;" << endl;
      *pofstreamCPP << "      int *aResolutions_d;" << endl;
      *pofstreamCPP << "   };" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return new ResolutionEnumerator (dimof (aSupportedResolutions), aSupportedResolutions);" << endl;
      *pofstreamCPP << "}" << endl;
   }

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printXMLDocument (deviceResolutionsNode->doc, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringResolutionClassName = 0;
   }

   for ( ResolutionMapElement::iterator next = mapResolution.begin () ;
         next != mapResolution.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( ResolutionMapData1::iterator next = mapData1.begin () ;
         next != mapData1.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( ResolutionMapData2::iterator next = mapData2.begin () ;
         next != mapData2.end () ;
         next++ )
   {
      delete[] (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDevicePrintModes (xmlNodePtr devicePrintModesNode)
{
   if (devicePrintModesNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringPrintModeClassName = pstringNameOut;

   typedef map <string, string *> PrintModeMapElement;

   xmlNodePtr           devicePrintModeNode = devicePrintModesNode->childs;
   ofstream            *pofstreamHPP        = 0;
   ofstream            *pofstreamCPP        = 0;
   PrintModeMapElement  mapElement;
   int                  iCount              = 0;
   int                  rc                  = false;

   if (devicePrintModeNode == 0)
      goto done;

   while (devicePrintModeNode != 0)
   {
      xmlNodePtr  nodeItem       = devicePrintModeNode->childs;
      const char *pszId          = 0;
      int         id;
      int         iPhysicalCount;
      int         iLogicalCount;
      int         iPlanes;

      if (nodeItem == 0)
         goto done;

      pszId = bundleStringData (nodeItem);
      if (pszId)
      {
         id = DevicePrintMode::nameToID (const_cast<char*>(pszId));
         if (-1 == id)
         {
            cerr << "Error: Unknown DevicePrintMode \"" << pszId << "\"" << endl;
            setErrorCondition ();
            goto done;
         }
      }

      pdi_d->listPrintModes.push_back (string (pszId));

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iPhysicalCount = bundleIntegerData (nodeItem);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iLogicalCount = bundleIntegerData (nodeItem);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iPlanes = bundleIntegerData (nodeItem);

      ostrstream oss;

      oss << "   case "
          << pszId
          << ": return new "
          << *pstringNameOut
          << " ("
          << pszId
          << ","
          << iPhysicalCount
          << ","
          << iLogicalCount
          << ","
          << iPlanes
          << ");"
          << ends;

      mapElement[string (pszId)] = new string (oss.str ());

      if (fDebugOutput)
      {
         cerr << dec << "id             = " << id             << endl;
         cerr        << "iPhysicalCount = " << iPhysicalCount << endl;
         cerr        << "iLogicalCount  = " << iLogicalCount  << endl;
         cerr        << "iPlanes        = " << iPlanes        << endl;
      }

      // Get the next devicePrintMode node
      devicePrintModeNode = nextNode (devicePrintModeNode);

      if (pszId)
      {
         free ((char *)pszId);
      }
   }

   pdi_d->listPrintModes.sort ();
   pdi_d->listPrintModes.unique ();

   rc = true;

   if (shouldCreateFile (*pstrXMLFile_d))
   {
      cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

      openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

      outputHeader (pofstreamHPP);

      *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
      *pofstreamHPP << "#define _" << *pstringNameOut << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#include <BinaryData.hpp>" << endl;
      *pofstreamHPP << "#include <DevicePrintMode.hpp>" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "class " << *pstringNameOut << " : public DevicePrintMode" << endl;
      *pofstreamHPP << "{" << endl;
      *pofstreamHPP << "public:" << endl;
      *pofstreamHPP << *pstringNameOut << " (int         id," << endl;
      *pofstreamHPP << "                          int         iPhysicalCount," << endl;
      *pofstreamHPP << "                          int         iLogicalCount," << endl;
      *pofstreamHPP << "                          int         iPlanes);" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "   static  DevicePrintMode *create               (int   id);" << endl;
      *pofstreamHPP << "   static  DevicePrintMode *create               (char *pszId);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual bool             isPrintModeSupported (int   id);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual Enumeration     *getEnumeration       ();" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "protected:" << endl;
      *pofstreamHPP << "   virtual DevicePrintMode *createV              (int   id);" << endl;

      *pofstreamHPP << "};" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#endif" << endl;

      outputHeader (pofstreamCPP);

      *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << *pstringNameOut << "::" << endl;
      *pofstreamCPP << *pstringNameOut << " (int          id," << endl;
      *pofstreamCPP << "                          int         iPhysicalCount," << endl;
      *pofstreamCPP << "                          int         iLogicalCount," << endl;
      *pofstreamCPP << "                          int         iPlanes)" << endl;
      *pofstreamCPP << "      : DevicePrintMode (id," << endl;
      *pofstreamCPP << "                         iPhysicalCount," << endl;
      *pofstreamCPP << "                         iLogicalCount," << endl;
      *pofstreamCPP << "                         iPlanes)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DevicePrintMode * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "createV (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return " << *pstringNameOut << "::create (id);" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DevicePrintMode * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (char *pszId)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return create (DevicePrintMode::nameToID (pszId));" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DevicePrintMode * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   switch (id)" << endl;
      *pofstreamCPP << "   {" << endl;

      for ( PrintModeMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++ )
      {
         string *pElement = (string *)next->second;

         *pofstreamCPP << *pElement << endl;
      }

      *pofstreamCPP << "   default: return 0;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << "static int aSupportedPrintModes[] = {" << endl;

      iCount = mapElement.size ();
      for ( PrintModeMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << "   DevicePrintMode::" << next->first;
         if (1 < iCount)
            *pofstreamCPP << ",";
         *pofstreamCPP << endl;
      }

      *pofstreamCPP << "};" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "bool " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "isPrintModeSupported (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   for (int i = 0; i < (int)dimof (aSupportedPrintModes); i++)" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "      if (aSupportedPrintModes[i] == id)" << endl;
      *pofstreamCPP << "         return true;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return false;" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "Enumeration * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "getEnumeration ()" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   class PrintModeEnumerator : public Enumeration" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "   public:" << endl;
      *pofstreamCPP << "      PrintModeEnumerator (int cPrintModes, int aPrintModes[])" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         iPrintMode_d  = 0;" << endl;
      *pofstreamCPP << "         cPrintModes_d = cPrintModes;" << endl;
      *pofstreamCPP << "         aPrintModes_d = aPrintModes;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual bool hasMoreElements ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iPrintMode_d < cPrintModes_d)" << endl;
      *pofstreamCPP << "            return true;" << endl;
      *pofstreamCPP << "         else" << endl;
      *pofstreamCPP << "            return false;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual void *nextElement ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iPrintMode_d > cPrintModes_d - 1)" << endl;
      *pofstreamCPP << "            return 0;" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "         return create (aPrintModes_d[iPrintMode_d++]);" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   private:" << endl;
      *pofstreamCPP << "      int  iPrintMode_d;" << endl;
      *pofstreamCPP << "      int  cPrintModes_d;" << endl;
      *pofstreamCPP << "      int *aPrintModes_d;" << endl;
      *pofstreamCPP << "   };" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return new PrintModeEnumerator (dimof (aSupportedPrintModes), aSupportedPrintModes);" << endl;
      *pofstreamCPP << "}" << endl;
   }

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printXMLDocument (devicePrintModesNode->doc, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringPrintModeClassName = 0;
   }

   for ( PrintModeMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      delete (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDeviceTrays (xmlNodePtr deviceTraysNode)
{
   if (deviceTraysNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringTrayClassName = pstringNameOut;

   typedef map <string, string *>     TrayMapElement;
   typedef map <string, BinaryData *> TrayMapData1;
   typedef map <string, byte *>       TrayMapData2;

   xmlNodePtr      deviceTrayNode = deviceTraysNode->childs;
   ofstream       *pofstreamHPP   = 0;
   ofstream       *pofstreamCPP   = 0;
   TrayMapElement  mapElement;
   TrayMapData1    mapData1;
   TrayMapData2    mapData2;
   TrayList        listType;
   int             iCount         = 0;
   int             rc             = false;

   if (deviceTrayNode == 0)
      goto done;

   while (deviceTrayNode != 0)
   {
      xmlNodePtr  nodeItem      = deviceTrayNode->childs;
      const char *pszId         = 0;
      int         id;
      const char *pszType;
      const char *pszBinaryData = 0;
      byte       *pbData        = 0;
      int         cbData        = 0;
      BinaryData *pdbData;

      if (nodeItem == 0)
         goto done;

      pszId = bundleStringData (nodeItem);
      if (pszId)
      {
         id = DeviceTray::nameToID (const_cast<char*>(pszId));
         if (-1 == id)
         {
            cerr << "Error: Unknown DeviceTray \"" << pszId << "\"" << endl;
            setErrorCondition ();
            goto done;
         }
      }

      pdi_d->listTrays.push_back (string (pszId));

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      pszType = bundleStringData (nodeItem);

      listType.push_back (string (pszType));

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      pszBinaryData = bundleStringData (nodeItem);

      if (parseBinaryData (pszBinaryData, &pbData, &cbData))
      {
         pdbData = new BinaryData (pbData, cbData);
      }
      else
      {
         cerr << "Error: binary data not understood \"" << pszBinaryData << "\" for " << pszId << "." << endl;
         goto done;
      }

      ostrstream oss;

      oss << "   case "
          << pszId
          << ": return new "
          << *pstringNameOut
          << " ("
          << pszId
          << ","
          << pszType
          << ",new BinaryData (abData"
          << pszId
          << ",dimof (abData"
          << pszId
          << ")));"
          << ends;

      mapData2[string (pszId)]   = pbData;
      mapData1[string (pszId)]   = new BinaryData (pbData, cbData);
      mapElement[string (pszId)] = new string (oss.str ());

      if (fDebugOutput)
      {
         cerr << dec << "id            = " << id            << endl;
         cerr        << "pszType       = " << pszType       << endl;
         cerr        << "pszBinaryData = " << pszBinaryData << endl;
         cerr        << "pdbData       = " << *pdbData      << endl;
      }

      // Get the next deviceTray node
      deviceTrayNode = nextNode (deviceTrayNode);

      if (pszId)
      {
         free ((char *)pszId);
      }
      if (pszType)
      {
         free ((char *)pszType);
      }
      if (pszBinaryData)
      {
         free ((char *)pszBinaryData);
      }
   }

   pdi_d->listTrays.sort ();
   pdi_d->listTrays.unique ();

   rc = true;

   if (shouldCreateFile (*pstrXMLFile_d))
   {
      cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

      openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

      outputHeader (pofstreamHPP);

      *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
      *pofstreamHPP << "#define _" << *pstringNameOut << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#include <BinaryData.hpp>" << endl;
      *pofstreamHPP << "#include <DeviceTray.hpp>" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceTray" << endl;
      *pofstreamHPP << "{" << endl;
      *pofstreamHPP << "public:" << endl;

      *pofstreamHPP << "   enum {" << endl;
      listType.sort ();
      listType.unique ();
      iCount = listType.size ();
      for ( TrayList::iterator next = listType.begin () ;
            next != listType.end () ;
            next++, iCount-- )
      {
         *pofstreamHPP << "      " << *next;
         if (1 < iCount)
            *pofstreamHPP << ",";
         *pofstreamHPP << endl;
      }
      *pofstreamHPP << "   };" << endl;

      *pofstreamHPP << *pstringNameOut << " (int         id," << endl;
      *pofstreamHPP << "                          int         iType," << endl;
      *pofstreamHPP << "                          BinaryData *data);" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "   static  DeviceTray  *create          (int   id);" << endl;
      *pofstreamHPP << "   static  DeviceTray  *create          (char *pszId);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual bool         isTraySupported (int   id);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual Enumeration *getEnumeration  ();" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "protected:" << endl;
      *pofstreamHPP << "   virtual DeviceTray  *createV         (int   id);" << endl;

      *pofstreamHPP << "};" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#endif" << endl;

      outputHeader (pofstreamCPP);

      *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << *pstringNameOut << "::" << endl;
      *pofstreamCPP << *pstringNameOut << " (int         id," << endl;
      *pofstreamCPP << "                          int         iType," << endl;
      *pofstreamCPP << "                          BinaryData *data)" << endl;
      *pofstreamCPP << "      : DeviceTray (id," << endl;
      *pofstreamCPP << "                    iType," << endl;
      *pofstreamCPP << "                    data)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceTray * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "createV (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return " << *pstringNameOut << "::create (id);" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceTray * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (char *pszId)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return create (DeviceTray::nameToID (pszId));" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceTray * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (int id)" << endl;
      *pofstreamCPP << "{" << endl;

      for ( TrayMapData1::iterator next = mapData1.begin () ;
            next != mapData1.end () ;
            next++ )
      {
         BinaryData *bdValue = (BinaryData *)next->second;
         PBYTE       pbValue = bdValue->getData ();
         int         cbValue = bdValue->getLength ();

         *pofstreamCPP << "   static byte abData" << next->first << "[] = {" << dec;

         for (int i = 0; i < cbValue; i++)
         {
            *pofstreamCPP << (int)pbValue[i];
            if (i < cbValue - 1)
               *pofstreamCPP << ",";
         }

         *pofstreamCPP << "};" << endl;
      }

      *pofstreamCPP << endl;
      *pofstreamCPP << "   switch (id)" << endl;
      *pofstreamCPP << "   {" << endl;

      for ( TrayMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++ )
      {
         string *pElement = (string *)next->second;

         *pofstreamCPP << *pElement << endl;
      }

      *pofstreamCPP << "   default: return 0;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << "" << endl;
      *pofstreamCPP << "static int aSupportedTrays[] = {" << endl;

      iCount = mapElement.size ();
      for ( TrayMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << "   DeviceTray::" << next->first;
         if (1 < iCount)
            *pofstreamCPP << ",";
         *pofstreamCPP << endl;
      }

      *pofstreamCPP << "   };" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "bool " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "isTraySupported (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   for (int i = 0; i < (int)dimof (aSupportedTrays); i++)" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "      if (aSupportedTrays[i] == id)" << endl;
      *pofstreamCPP << "         return true;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return false;" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "Enumeration * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "getEnumeration ()" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   class TrayEnumerator : public Enumeration" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "   public:" << endl;
      *pofstreamCPP << "      TrayEnumerator (int cTrays, int aTrays[])" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         iTray_d  = 0;" << endl;
      *pofstreamCPP << "         cTrays_d = cTrays;" << endl;
      *pofstreamCPP << "         aTrays_d = aTrays;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual bool hasMoreElements ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iTray_d < cTrays_d)" << endl;
      *pofstreamCPP << "            return true;" << endl;
      *pofstreamCPP << "         else" << endl;
      *pofstreamCPP << "            return false;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual void *nextElement ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iTray_d > cTrays_d - 1)" << endl;
      *pofstreamCPP << "            return 0;" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "         return create (aTrays_d[iTray_d++]);" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   private:" << endl;
      *pofstreamCPP << "      int  iTray_d;" << endl;
      *pofstreamCPP << "      int  cTrays_d;" << endl;
      *pofstreamCPP << "      int *aTrays_d;" << endl;
      *pofstreamCPP << "   };" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return new TrayEnumerator (dimof (aSupportedTrays), aSupportedTrays);" << endl;
      *pofstreamCPP << "}" << endl;
   }

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printXMLDocument (deviceTraysNode->doc, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringTrayClassName = 0;
   }

   for ( TrayMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( TrayMapData1::iterator next = mapData1.begin () ;
         next != mapData1.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( TrayMapData2::iterator next = mapData2.begin () ;
         next != mapData2.end () ;
         next++ )
   {
      delete[] (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDeviceForms (xmlNodePtr deviceFormsNode)
{
   if (deviceFormsNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringFormClassName = pstringNameOut;

   typedef map <string, string *>     FormMapElement;
   typedef map <string, BinaryData *> FormMapData1;
   typedef map <string, byte *>       FormMapData2;
   typedef list <string>              FormList;

   xmlNodePtr      deviceFormNode = deviceFormsNode->childs;
   ofstream       *pofstreamHPP   = 0;
   ofstream       *pofstreamCPP   = 0;
   FormMapElement  mapElement;
   FormMapData1    mapData1;
   FormMapData2    mapData2;
   FormList        listType;
   int             iCount         = 0;
   int             rc             = false;

   if (deviceFormNode == 0)
      goto done;

   while (deviceFormNode != 0)
   {
      xmlNodePtr  nodeItem        = deviceFormNode->childs;
      xmlNodePtr  nodeHCC;
      const char *pszId           = 0;
      int         id;
      const char *pszCapabilities = 0;
      const char *pszBinaryData   = 0;
      byte       *pbData          = 0;
      int         cbData          = 0;
      BinaryData *pdbData;
      int         iLeft           = 0;
      int         iTop            = 0;
      int         iRight          = 0;
      int         iBottom         = 0;

      if (nodeItem == 0)
         goto done;

      pszId = bundleStringData (nodeItem);
      if (pszId)
      {
         id = DeviceForm::nameToID (const_cast<char*>(pszId));
         if (-1 == id)
         {
            cerr << "Error: Unknown DeviceForm \"" << pszId << "\"" << endl;
            setErrorCondition ();
            goto done;
         }
      }

      pdi_d->listForms.push_back (string (pszId));

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      pszCapabilities = bundleStringData (nodeItem);

      if (!DeviceForm::isReservedKeyword (pszCapabilities))
      {
         listType.push_back (string (pszCapabilities));
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      pszBinaryData = bundleStringData (nodeItem);

      if (parseBinaryData (pszBinaryData, &pbData, &cbData))
      {
         pdbData = new BinaryData (pbData, cbData);
      }
      else
      {
         cerr << "Error: binary data not understood \"" << pszBinaryData << "\" for " << pszId << "." << endl;
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      // Move down into the HardCopyCap tree
      nodeHCC = nodeItem->childs;
      if (nodeHCC == 0)
         goto done;

      iLeft = bundleIntegerData (nodeHCC);

      // Move to the next one
      nodeHCC = nextNode (nodeHCC);
      if (nodeHCC == 0)
         goto done;

      iTop = bundleIntegerData (nodeHCC);

      // Move to the next one
      nodeHCC = nextNode (nodeHCC);
      if (nodeHCC == 0)
         goto done;

      iRight = bundleIntegerData (nodeHCC);

      // Move to the next one
      nodeHCC = nextNode (nodeHCC);
      if (nodeHCC == 0)
         goto done;

      iBottom = bundleIntegerData (nodeHCC);

      ostrstream oss;

      oss << "   case "
          << pszId
          << ": return new "
          << *pstringNameOut
          << " ("
          << pszId
          << ","
          << pszCapabilities
          << ",new BinaryData (abData"
          << pszId
          << ",dimof (abData"
          << pszId
          << ")),new HardCopyCap ("
          << iLeft
          << ","
          << iTop
          << ","
          << iRight
          << ","
          << iBottom
          << "));"
          << ends;

      mapData2[string (pszId)]   = pbData;
      mapData1[string (pszId)]   = new BinaryData (pbData, cbData);
      mapElement[string (pszId)] = new string (oss.str ());

      if (fDebugOutput)
      {
         cerr << dec << "id              = " << id              << endl;
         cerr        << "pszCapabilities = " << pszCapabilities << endl;
         cerr        << "pszBinaryData   = " << pszBinaryData   << endl;
         cerr        << "pdbData         = " << *pdbData        << endl;
         cerr << dec << "   iLeft        = " << iLeft           << endl;
         cerr        << "   iTop         = " << iTop            << endl;
         cerr        << "   iRight       = " << iRight          << endl;
         cerr        << "   iBottom      = " << iBottom         << endl;
      }

      // Get the next deviceForm node
      deviceFormNode = nextNode (deviceFormNode);

      if (pszId)
      {
         free ((char *)pszId);
      }
      if (pszCapabilities)
      {
         free ((char *)pszCapabilities);
      }
      if (pszBinaryData)
      {
         free ((char *)pszBinaryData);
      }
   }

   pdi_d->listForms.sort ();
   pdi_d->listForms.unique ();

   rc = true;

   if (shouldCreateFile (*pstrXMLFile_d))
   {
      cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

      openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

      outputHeader (pofstreamHPP);

      *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
      *pofstreamHPP << "#define _" << *pstringNameOut << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#include <BinaryData.hpp>" << endl;
      *pofstreamHPP << "#include <DeviceForm.hpp>" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceForm" << endl;
      *pofstreamHPP << "{" << endl;
      *pofstreamHPP << "public:" << endl;

      *pofstreamHPP << "   enum {" << endl;
      listType.sort ();
      listType.unique ();
      iCount = listType.size ();
      for ( FormList::iterator next = listType.begin () ;
            next != listType.end () ;
            next++, iCount-- )
      {
         *pofstreamHPP << "      " << *next;
         if (1 < iCount)
            *pofstreamHPP << ",";
         *pofstreamHPP << endl;
      }
      *pofstreamHPP << "   };" << endl;
      *pofstreamHPP << endl;

      *pofstreamHPP << *pstringNameOut << " (int          id," << endl;
      *pofstreamHPP << "                          int          iCapabilities," << endl;
      *pofstreamHPP << "                          BinaryData  *data," << endl;
      *pofstreamHPP << "                          HardCopyCap *hcInfo);" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "   static  DeviceForm  *create          (int   id);" << endl;
      *pofstreamHPP << "   static  DeviceForm  *create          (char *pszId);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual bool         isFormSupported (int   id);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual Enumeration *getEnumeration  ();" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "protected:" << endl;
      *pofstreamHPP << "   virtual DeviceForm  *createV         (int   id);" << endl;

      *pofstreamHPP << "};" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#endif" << endl;

      outputHeader (pofstreamCPP);

      *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << *pstringNameOut << "::" << endl;
      *pofstreamCPP << *pstringNameOut << " (int          id," << endl;
      *pofstreamCPP << "                          int          iCapabilities," << endl;
      *pofstreamCPP << "                          BinaryData  *data," << endl;
      *pofstreamCPP << "                          HardCopyCap *hcInfo)" << endl;
      *pofstreamCPP << "      : DeviceForm (id," << endl;
      *pofstreamCPP << "                    iCapabilities," << endl;
      *pofstreamCPP << "                    data," << endl;
      *pofstreamCPP << "                    hcInfo)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceForm * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "createV (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return " << *pstringNameOut << "::create (id);" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceForm * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (char *pszId)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return create (DeviceForm::nameToID (pszId));" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceForm * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (int id)" << endl;
      *pofstreamCPP << "{" << endl;

      for ( FormMapData1::iterator next = mapData1.begin () ;
            next != mapData1.end () ;
            next++ )
      {
         BinaryData *bdValue = (BinaryData *)next->second;
         PBYTE       pbValue = bdValue->getData ();
         int         cbValue = bdValue->getLength ();

         *pofstreamCPP << "   static byte abData" << next->first << "[] = {" << dec;

         for (int i = 0; i < cbValue; i++)
         {
            *pofstreamCPP << (int)pbValue[i];
            if (i < cbValue - 1)
               *pofstreamCPP << ",";
         }

         *pofstreamCPP << "};" << endl;
      }

      *pofstreamCPP << endl;
      *pofstreamCPP << "   switch (id)" << endl;
      *pofstreamCPP << "   {" << endl;

      for ( FormMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++ )
      {
         string *pElement = (string *)next->second;

         *pofstreamCPP << *pElement << endl;
      }

      *pofstreamCPP << "   default: return 0;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "static int aSupportedForms[] = {" << endl;

      iCount = mapElement.size ();
      for ( FormMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << "   DeviceForm::" << next->first;
         if (1 < iCount)
            *pofstreamCPP << ",";
         *pofstreamCPP << endl;
      }

      *pofstreamCPP << "};" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "bool " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "isFormSupported (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   for (int i = 0; i < (int)dimof (aSupportedForms); i++)" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "      if (aSupportedForms[i] == id)" << endl;
      *pofstreamCPP << "         return true;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return false;" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "Enumeration * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "getEnumeration ()" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   class FormEnumerator : public Enumeration" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "   public:" << endl;
      *pofstreamCPP << "      FormEnumerator (int cForms, int aForms[])" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         iForm_d  = 0;" << endl;
      *pofstreamCPP << "         cForms_d = cForms;" << endl;
      *pofstreamCPP << "         aForms_d = aForms;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual bool hasMoreElements ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iForm_d < cForms_d)" << endl;
      *pofstreamCPP << "            return true;" << endl;
      *pofstreamCPP << "         else" << endl;
      *pofstreamCPP << "            return false;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual void *nextElement ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iForm_d > cForms_d - 1)" << endl;
      *pofstreamCPP << "            return 0;" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "         return create (aForms_d[iForm_d++]);" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   private:" << endl;
      *pofstreamCPP << "      int  iForm_d;" << endl;
      *pofstreamCPP << "      int  cForms_d;" << endl;
      *pofstreamCPP << "      int *aForms_d;" << endl;
      *pofstreamCPP << "   };" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return new FormEnumerator (dimof (aSupportedForms), aSupportedForms);" << endl;
      *pofstreamCPP << "}" << endl;
   }

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printXMLDocument (deviceFormsNode->doc, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringFormClassName = 0;
   }

   for ( FormMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( FormMapData1::iterator next = mapData1.begin () ;
         next != mapData1.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( FormMapData2::iterator next = mapData2.begin () ;
         next != mapData2.end () ;
         next++ )
   {
      delete[] (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDeviceMedias (xmlNodePtr deviceMediasNode)
{
   if (deviceMediasNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringMediaClassName = pstringNameOut;

   typedef map <string, string *>     MediaMapElement;
   typedef map <string, BinaryData *> MediaMapData1;
   typedef map <string, byte *>       MediaMapData2;

   xmlNodePtr      deviceMediaNode = deviceMediasNode->childs;
   ofstream       *pofstreamHPP    = 0;
   ofstream       *pofstreamCPP    = 0;
   MediaMapElement mapElement;
   MediaMapData1   mapData1;
   MediaMapData2   mapData2;
   int             iCount          = 0;
   int             rc              = false;

   if (deviceMediaNode == 0)
      goto done;

   while (deviceMediaNode != 0)
   {
      xmlNodePtr  nodeItem             = deviceMediaNode->childs;
      const char *pszId                = 0;
      int         id;
      const char *pszBinaryData        = 0;
      byte       *pbData               = 0;
      int         cbData               = 0;
      BinaryData *pdbData;
      int         iColorAdjustRequired;
      const char *pszAbsorption        = 0;

      if (nodeItem == 0)
         goto done;

      pszId = bundleStringData (nodeItem);
      if (pszId)
      {
         id = DeviceMedia::nameToID (const_cast<char*>(pszId));
         if (-1 == id)
         {
            cerr << "Error: Unknown DeviceMedia \"" << pszId << "\"" << endl;
            setErrorCondition ();
            goto done;
         }
      }

      pdi_d->listMedias.push_back (string (pszId));

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      pszBinaryData = bundleStringData (nodeItem);

      if (parseBinaryData (pszBinaryData, &pbData, &cbData))
      {
         pdbData = new BinaryData (pbData, cbData);
      }
      else
      {
         cerr << "Error: binary data not understood \"" << pszBinaryData << "\" for " << pszId << "." << endl;
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iColorAdjustRequired = bundleIntegerData (nodeItem);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      pszAbsorption = bundleStringData (nodeItem);

      if (  0 != strcmp (pszAbsorption, "MEDIA_NO_ABSORPTION")
         && 0 != strcmp (pszAbsorption, "MEDIA_LIGHT_ABSORPTION")
         && 0 != strcmp (pszAbsorption, "MEDIA_HEAVY_ABSORPTION")
         )
         goto done;

      ostrstream oss;

      oss << "   case "
          << pszId
          << ": return new "
          << *pstringNameOut
          << " ("
          << pszId
          << ",new BinaryData (abData"
          << pszId
          << ",dimof (abData"
          << pszId
          << ")),"
          << iColorAdjustRequired
          << ","
          << pszAbsorption
          << ");"
          << ends;

      mapData2[string (pszId)]   = pbData;
      mapData1[string (pszId)]   = new BinaryData (pbData, cbData);
      mapElement[string (pszId)] = new string (oss.str ());

      if (fDebugOutput)
      {
         cerr << dec << "id                   = " << id                   << endl;
         cerr        << "pszBinaryData        = " << pszBinaryData        << endl;
         cerr        << "pdbData              = " << *pdbData             << endl;
         cerr        << "iColorAdjustRequired = " << iColorAdjustRequired << endl;
         cerr        << "pszAbsorption        = " << pszAbsorption        << endl;
      }

      // Get the next deviceMedia node
      deviceMediaNode = nextNode (deviceMediaNode);

      if (pszId)
      {
         free ((char *)pszId);
      }
      if (pszBinaryData)
      {
         free ((char *)pszBinaryData);
      }
      if (pszAbsorption)
      {
         free ((char *)pszAbsorption);
      }
   }

   pdi_d->listMedias.sort ();
   pdi_d->listMedias.unique ();

   rc = true;

   if (shouldCreateFile (*pstrXMLFile_d))
   {
      cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

      openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

      outputHeader (pofstreamHPP);

      *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
      *pofstreamHPP << "#define _" << *pstringNameOut << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#include <BinaryData.hpp>" << endl;
      *pofstreamHPP << "#include <DeviceMedia.hpp>" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceMedia" << endl;
      *pofstreamHPP << "{" << endl;
      *pofstreamHPP << "public:" << endl;

      *pofstreamHPP << *pstringNameOut << " (int          id," << endl;
      *pofstreamHPP << "                           BinaryData *data," << endl;
      *pofstreamHPP << "                           int         iColorAdjustRequired," << endl;
      *pofstreamHPP << "                           int         iAbsorption);" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "   static  DeviceMedia *create           (int   id);" << endl;
      *pofstreamHPP << "   static  DeviceMedia *create           (char *pszId);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual bool         isMediaSupported (int   id);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual Enumeration *getEnumeration   ();" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "protected:" << endl;
      *pofstreamHPP << "   virtual DeviceMedia *createV          (int   id);" << endl;

      *pofstreamHPP << "};" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#endif" << endl;

      outputHeader (pofstreamCPP);

      *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << *pstringNameOut << "::" << endl;
      *pofstreamCPP << *pstringNameOut << " (int          id," << endl;
      *pofstreamCPP << "                           BinaryData *data," << endl;
      *pofstreamCPP << "                           int         iColorAdjustRequired," << endl;
      *pofstreamCPP << "                           int         iAbsorption)" << endl;
      *pofstreamCPP << "      : DeviceMedia (id," << endl;
      *pofstreamCPP << "                     data," << endl;
      *pofstreamCPP << "                     iColorAdjustRequired," << endl;
      *pofstreamCPP << "                     iAbsorption)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceMedia * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "createV (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return " << *pstringNameOut << "::create (id);" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceMedia * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (char *pszId)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return create (DeviceMedia::nameToID (pszId));" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceMedia * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (int id)" << endl;
      *pofstreamCPP << "{" << endl;

      for ( MediaMapData1::iterator next = mapData1.begin () ;
            next != mapData1.end () ;
            next++ )
      {
         BinaryData *bdValue = (BinaryData *)next->second;
         PBYTE       pbValue = bdValue->getData ();
         int         cbValue = bdValue->getLength ();

         *pofstreamCPP << "   static byte abData" << next->first << "[] = {" << dec;

         for (int i = 0; i < cbValue; i++)
         {
            *pofstreamCPP << (int)pbValue[i];
            if (i < cbValue - 1)
               *pofstreamCPP << ",";
         }

         *pofstreamCPP << "};" << endl;
      }

      *pofstreamCPP << endl;
      *pofstreamCPP << "   switch (id)" << endl;
      *pofstreamCPP << "   {" << endl;

      for ( MediaMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++ )
      {
         string *pElement = (string *)next->second;

         *pofstreamCPP << *pElement << endl;
      }

      *pofstreamCPP << "   default: return 0;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "static int aSupportedMedias[] = {" << endl;

      iCount = mapElement.size ();
      for ( MediaMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << "   DeviceMedia::" << next->first;
         if (1 < iCount)
            *pofstreamCPP << ",";
         *pofstreamCPP << endl;
      }

      *pofstreamCPP << "};" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "bool " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "isMediaSupported (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   for (int i = 0; i < (int)dimof (aSupportedMedias); i++)" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "      if (aSupportedMedias[i] == id)" << endl;
      *pofstreamCPP << "         return true;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return false;" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "Enumeration * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "getEnumeration ()" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   class MediaEnumerator : public Enumeration" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "   public:" << endl;
      *pofstreamCPP << "      MediaEnumerator (int cMedias, int aMedias[])" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         iMedia_d  = 0;" << endl;
      *pofstreamCPP << "         cMedias_d = cMedias;" << endl;
      *pofstreamCPP << "         aMedias_d = aMedias;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual bool hasMoreElements ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iMedia_d < cMedias_d)" << endl;
      *pofstreamCPP << "            return true;" << endl;
      *pofstreamCPP << "         else" << endl;
      *pofstreamCPP << "            return false;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual void *nextElement ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iMedia_d > cMedias_d - 1)" << endl;
      *pofstreamCPP << "            return 0;" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "         return create (aMedias_d[iMedia_d++]);" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   private:" << endl;
      *pofstreamCPP << "      int  iMedia_d;" << endl;
      *pofstreamCPP << "      int  cMedias_d;" << endl;
      *pofstreamCPP << "      int *aMedias_d;" << endl;
      *pofstreamCPP << "   };" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return new MediaEnumerator (dimof (aSupportedMedias), aSupportedMedias);" << endl;
      *pofstreamCPP << "}" << endl;
   }

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printXMLDocument (deviceMediasNode->doc, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringMediaClassName = 0;
   }

   for ( MediaMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( MediaMapData1::iterator next = mapData1.begin () ;
         next != mapData1.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( MediaMapData2::iterator next = mapData2.begin () ;
         next != mapData2.end () ;
         next++ )
   {
      delete[] (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDeviceConnections (xmlNodePtr deviceConnectionsNode)
{
   if (deviceConnectionsNode == 0)
      return false;

   xmlNodePtr deviceConnectionNode = deviceConnectionsNode->childs;
   int        rc                   = false;

   if (deviceConnectionNode == 0)
      goto done;

   while (deviceConnectionNode != 0)
   {
      xmlNodePtr nodeItem = deviceConnectionNode->childs;

      if (nodeItem == 0)
         goto done;

      // name

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      // form

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      // tray

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      // media

      // Get the next deviceConnection node
      deviceConnectionNode = nextNode (deviceConnectionNode);
   }

   rc = true;

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printXMLDocument (deviceConnectionsNode->doc, 0);
   }

   // Clean up!

   return rc;
}

bool OmniDomParser::
processDeviceGammas (xmlNodePtr deviceGammasNode)
{
   if (deviceGammasNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringGammaClassName = pstringNameOut;

   if (!shouldCreateFile (*pstrXMLFile_d))
      return true;

   if (!pdi_d->pstringResolutionClassName)
   {
      cerr << "Error: DeviceResolution not filled in." << endl;
      return false;
   }

   if (!pdi_d->pstringMediaClassName)
   {
      cerr << "Error: DeviceMedia not filled in." << endl;
      return false;
   }

   if (!pdi_d->pstringPrintModeClassName)
   {
      cerr << "Error: DevicePrintMode not filled in." << endl;
      return false;
   }

   cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

   typedef map <string, string *> GammaMapElement;

   xmlNodePtr      deviceGammaNode = deviceGammasNode->childs;
   ofstream       *pofstreamHPP    = 0;
   ofstream       *pofstreamCPP    = 0;
   GammaMapElement mapElement;
   int             iCount          = 0;
   int             rc              = false;

   if (deviceGammaNode == 0)
      goto done;

   while (deviceGammaNode != 0)
   {
      xmlNodePtr  nodeItem            = deviceGammaNode->childs;
      const char *pszResolutionID     = 0;
      const char *pszMediaID          = 0;
      const char *pszPrintModeID      = 0;
      const char *pszDitherCatagoryID = 0;
      int         iCGamma;
      int         iMGamma;
      int         iYGamma;
      int         iKGamma;
      int         iCBias;
      int         iMBias;
      int         iYBias;
      int         iKBias;

      if (nodeItem == 0)
         goto done;

      if (  !(pszResolutionID = bundleStringData (nodeItem))
         || -1 == DeviceResolution::nameToID (const_cast<char*>(pszResolutionID))
         )
      {
         cerr << "Error: Unknown DeviceResolution \"" << pszResolutionID << "\"" << endl;
         setErrorCondition ();
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      if (  !(pszMediaID = bundleStringData (nodeItem))
         || -1 == DeviceMedia::nameToID (const_cast<char*>(pszMediaID))
         )
      {
         cerr << "Error: Unknown DeviceMedia \"" << pszMediaID << "\"" << endl;
         setErrorCondition ();
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      if (  !(pszPrintModeID = bundleStringData (nodeItem))
         || -1 == DevicePrintMode::nameToID (const_cast<char*>(pszPrintModeID))
         )
      {
         cerr << "Error: Unknown DevicePrintMode \"" << pszPrintModeID << "\"" << endl;
         setErrorCondition ();
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      if (  !(pszDitherCatagoryID = bundleStringData (nodeItem))
         || !DeviceDither::ditherCatagoryValid (const_cast<char*>(pszDitherCatagoryID))
         )
      {
         cerr << "Error: Unknown Dither type \"" << pszDitherCatagoryID << "\"" << endl;
         setErrorCondition ();
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iCGamma = bundleIntegerData (nodeItem);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iMGamma = bundleIntegerData (nodeItem);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iYGamma = bundleIntegerData (nodeItem);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iKGamma = bundleIntegerData (nodeItem);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iCBias = bundleIntegerData (nodeItem);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iMBias = bundleIntegerData (nodeItem);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iYBias = bundleIntegerData (nodeItem);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      iKBias = bundleIntegerData (nodeItem);

      ostrstream oss;

      oss << "   "
          << "{"
          << *pdi_d->pstringResolutionClassName
          << "::"
          << pszResolutionID
          << ","
          << *pdi_d->pstringMediaClassName
          << "::"
          << pszMediaID
          << ","
          << *pdi_d->pstringPrintModeClassName
          << "::"
          << pszPrintModeID
          << ","
          << "\""
          << pszDitherCatagoryID
          << "\","
          << iCGamma
          << ","
          << iMGamma
          << ","
          << iYGamma
          << ","
          << iKGamma
          << ","
          << iCBias
          << ","
          << iMBias
          << ","
          << iYBias
          << ","
          << iKBias
          << "}"
          << ends;

      ostrstream ossName;

      // Create a unique name
      ossName << pszResolutionID
              << "/"
              << pszMediaID
              << "/"
              << pszPrintModeID
              << "/"
              << pszDitherCatagoryID
              << ends;

      mapElement[string (ossName.str ())] = new string (oss.str ());

      if (fDebugOutput)
      {
         cerr << dec << "pszResolutionID     = " << pszResolutionID     << endl;
         cerr <<        "pszMediaID          = " << pszMediaID          << endl;
         cerr <<        "pszPrintModeID      = " << pszPrintModeID      << endl;
         cerr <<        "pszDitherCatagoryID = " << pszDitherCatagoryID << endl;
         cerr <<        "iCGamma =             " << iCGamma             << endl;
         cerr <<        "iMGamma =             " << iMGamma             << endl;
         cerr <<        "iYGamma =             " << iYGamma             << endl;
         cerr <<        "iKGamma =             " << iKGamma             << endl;
         cerr <<        "iCBias =              " << iCBias              << endl;
         cerr <<        "iMBias =              " << iMBias              << endl;
         cerr <<        "iYBias =              " << iYBias              << endl;
         cerr <<        "iKBias =              " << iKBias              << endl;
      }

      // Get the next deviceGamma node
      deviceGammaNode = nextNode (deviceGammaNode);

      if (pszResolutionID)
      {
         free ((char *)pszResolutionID);
      }
      if (pszMediaID)
      {
         free ((char *)pszMediaID);
      }
      if (pszPrintModeID)
      {
         free ((char *)pszPrintModeID);
      }
      if (pszDitherCatagoryID)
      {
         free ((char *)pszDitherCatagoryID);
      }
   }

   rc = true;

   openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

   outputHeader (pofstreamHPP);

   *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
   *pofstreamHPP << "#define _" << *pstringNameOut << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "#include <DeviceGammaTable.hpp>" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceGammaTable" << endl;
   *pofstreamHPP << "{" << endl;
   *pofstreamHPP << "public:" << endl;

   *pofstreamHPP << "static DeviceGamma *getDeviceGamma (int   iResolutionID," << endl;
   *pofstreamHPP << "                                    int   iMediaID," << endl;
   *pofstreamHPP << "                                    int   iPrintModeID," << endl;
   *pofstreamHPP << "                                    char *pszDitherID);" << endl;
   *pofstreamHPP << "};" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "#endif" << endl;

   outputHeader (pofstreamCPP);

   *pofstreamCPP << "#include <defines.hpp>" << endl;
   *pofstreamCPP << "#include <DeviceDither.hpp>" << endl;
   *pofstreamCPP << "#include <" << *pdi_d->pstringResolutionClassName << ".hpp>" << endl;
   *pofstreamCPP << "#include <" << *pdi_d->pstringMediaClassName << ".hpp>" << endl;
   *pofstreamCPP << "#include <" << *pdi_d->pstringPrintModeClassName << ".hpp>" << endl;
   *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << "typedef struct _GAMMATABLE {" << endl;
   *pofstreamCPP << "   int                iResolutionID;" << endl;
   *pofstreamCPP << "   int                iMediaID;" << endl;
   *pofstreamCPP << "   int                iPrintModeID;" << endl;
   *pofstreamCPP << "   char              *pszDitherCatagoryID;" << endl;
   *pofstreamCPP << "   unsigned short int usCGamma;" << endl;
   *pofstreamCPP << "   unsigned short int usMGamma;" << endl;
   *pofstreamCPP << "   unsigned short int usYGamma;" << endl;
   *pofstreamCPP << "   unsigned short int usKGamma;" << endl;
   *pofstreamCPP << "   unsigned short int usCBias;" << endl;
   *pofstreamCPP << "   unsigned short int usMBias;" << endl;
   *pofstreamCPP << "   unsigned short int usYBias;" << endl;
   *pofstreamCPP << "   unsigned short int usKBias;" << endl;
   *pofstreamCPP << "} GAMMATABLE, *PGAMMATABLE;" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << "GAMMATABLE gt" << *pstringNameOut << "[] = {" << endl;

   iCount = mapElement.size ();
   for ( GammaMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      string *pElement = (string *)next->second;

      *pofstreamCPP << *pElement;
      if (1 < iCount)
         *pofstreamCPP << ",";
      *pofstreamCPP << endl;
   }

   *pofstreamCPP << "};" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceGamma * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDeviceGamma (int   iResolutionID," << endl;
   *pofstreamCPP << "                int   iMediaID," << endl;
   *pofstreamCPP << "                int   iPrintModeID," << endl;
   *pofstreamCPP << "                char *pszDitherID)" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   char *pszDitherCatagoryID = DeviceDither::getDitherCatagory (pszDitherID);" << endl;
   *pofstreamCPP << "" << endl;
   *pofstreamCPP << "   for (int i = 0; i < (int)dimof (gt" << *pstringNameOut << "); i++)" << endl;
   *pofstreamCPP << "   {" << endl;
   *pofstreamCPP << "      if (  iResolutionID     == gt" << *pstringNameOut << "[i].iResolutionID" << endl;
   *pofstreamCPP << "         && iMediaID          == gt" << *pstringNameOut << "[i].iMediaID" << endl;
   *pofstreamCPP << "         && iPrintModeID      == gt" << *pstringNameOut << "[i].iPrintModeID" << endl;
   *pofstreamCPP << "         && 0 == strcmp (pszDitherCatagoryID, gt" << *pstringNameOut << "[i].pszDitherCatagoryID)" << endl;
   *pofstreamCPP << "         )" << endl;
   *pofstreamCPP << "      {" << endl;
   *pofstreamCPP << "         return new DeviceGamma ((int)gt" << *pstringNameOut << "[i].usCGamma," << endl;
   *pofstreamCPP << "                                 (int)gt" << *pstringNameOut << "[i].usMGamma," << endl;
   *pofstreamCPP << "                                 (int)gt" << *pstringNameOut << "[i].usYGamma," << endl;
   *pofstreamCPP << "                                 (int)gt" << *pstringNameOut << "[i].usKGamma," << endl;
   *pofstreamCPP << "                                 (int)gt" << *pstringNameOut << "[i].usCBias,"  << endl;
   *pofstreamCPP << "                                 (int)gt" << *pstringNameOut << "[i].usMBias,"  << endl;
   *pofstreamCPP << "                                 (int)gt" << *pstringNameOut << "[i].usYBias,"  << endl;
   *pofstreamCPP << "                                 (int)gt" << *pstringNameOut << "[i].usKBias);" << endl;
   *pofstreamCPP << "      }" << endl;
   *pofstreamCPP << "   }" << endl;
   *pofstreamCPP << "" << endl;
   *pofstreamCPP << "   return 0;" << endl;
   *pofstreamCPP << "}" << endl;

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printXMLDocument (deviceGammasNode->doc, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringGammaClassName = 0;
   }

   for ( GammaMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      delete (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDeviceDatas (xmlNodePtr deviceDatasNode)
{
   if (deviceDatasNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringDataClassName = pstringNameOut;

   if (!shouldCreateFile (*pstrXMLFile_d))
      return true;

   cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

   typedef map <string, BinaryData *> DataMapElement;
   typedef map <string, byte *>       DataMapData;

   xmlNodePtr      deviceDataNode = deviceDatasNode->childs;
   ofstream       *pofstreamHPP   = 0;
   ofstream       *pofstreamCPP   = 0;
   DataMapElement  mapElement;
   DataMapData     mapData;
   int             rc             = false;

   if (deviceDataNode == 0)
      goto done;

   while (deviceDataNode != 0)
   {
      const char *pszDataName  = (char *)xmlGetProp (deviceDataNode, (xmlChar *)"name");
      const char *pszDataValue = 0;
      const char *pszTypeName  = (char *)xmlGetProp (deviceDataNode, (xmlChar *)"type");
      byte       *pbData       = 0;
      int         cbData;

      if (pszDataName == 0)
         goto done;

      if (pszTypeName == 0)
         goto done;

      pszDataValue = bundleStringData (deviceDataNode);

      if (fDebugOutput) cerr << "name  = " << pszDataName  << ", type = " << pszTypeName << endl;
      if (fDebugOutput) cerr << "value = " << pszDataValue << endl;

      if (0 == strcmp (pszTypeName, "string"))
      {
         cbData = strlen (pszDataValue) + 1;
         pbData = new byte [cbData];

         if (pbData)
         {
            strcpy ((char *)pbData, pszDataValue);
         }
         else
         {
            goto done;
         }
      }
      else if (0 == strcmp (pszTypeName, "boolean"))
      {
         cbData = 4;
         pbData = new byte [cbData];

         if (pbData)
         {
            memset (pbData, 0, cbData);

            if (0 == strcasecmp (pszDataValue, "true"))
            {
               pbData[0] = 1;
            }
            else if (0 == strcasecmp (pszDataValue, "false"))
            {
               pbData[0] = 0;
            }
            else
            {
               delete[] pbData;
               goto done;
            }
         }
         else
         {
            goto done;
         }
      }
      else if (0 == strcmp (pszTypeName, "integer"))
      {
         cbData = 4;
         pbData = new byte [cbData];

         if (pbData)
         {
            if (0 == sscanf (pszDataValue, "%d", (int *)pbData))
            {
               delete[] pbData;
               goto done;
            }
         }
         else
         {
            goto done;
         }
      }
      else if (0 == strcmp (pszTypeName, "byte"))
      {
         cbData = 1;
         pbData = new byte [cbData];

         if (pbData)
         {
            if (0 == sscanf (pszDataValue, "%c", (char *)pbData))
            {
               delete[] pbData;
               goto done;
            }
         }
         else
         {
            goto done;
         }
      }
      else if (0 == strcmp (pszTypeName, "binary"))
      {
         if (!parseBinaryData (pszDataValue, &pbData, &cbData))
         {
            cerr << "Error: binary data not understood \"" << pszDataValue << "\" for " << pszDataName << "." << endl;
            delete[] pbData;
            goto done;
         }
      }
      else if (0 == strcmp (pszTypeName, "bytearray"))
      {
         int iCnt,
             iTempCnt = 0;

         for (iCnt = 0; *(pszDataValue+iCnt); iCnt++)
         {
            if (!isxdigit (*(pszDataValue+iCnt)))
            {
               if (isspace (*(pszDataValue+iCnt)) == 0)
               {
                  cerr << "Error: Data in bytearray is not understood \"" << *(pszDataValue+iCnt) << "\" for " << pszDataName << "." << endl;
                  goto done;
               }
            }
            else
            {
               iTempCnt++;
            }
         }

         cbData = iTempCnt/2;
         pbData = new byte [cbData];
         if (!pbData)
         {
            goto done;
         }

         int          iEnd = strlen (pszDataValue);
         int          iLoc = 0;
         bool         bOK;
         char         cTemp[2];
         unsigned int iValue;

         iTempCnt = 0;

         for (iCnt = 0; iCnt < iEnd; iCnt++)
         {
            if (isxdigit (*(pszDataValue+iCnt)))
            {
               cTemp[iLoc++] = *(pszDataValue+iCnt);
               if (iLoc == 2)
               {
                  bOK = parseHexGroup ((const char *)cTemp, &iValue);
                  if (!bOK)
                  {
                     cerr << "Error: Data in bytearray is not understood \"" << *(pszDataValue+iCnt) << "\" for " << pszDataName << "." << endl;
                     delete[] pbData;
                     goto done;
                  }

                  *(pbData+iTempCnt++) = (unsigned char)iValue;
                  iLoc = 0;
               }
            }
         }

         if (iLoc)
         {
            cerr << "Error: Data in bytearray is missing bytes\"" << *(pszDataValue+iCnt) << "\" for " << pszDataName << "." << endl;
            delete[] pbData;
            goto done;
         }
      }
      else
      {
         cerr << "Error: Unknown type " << pszTypeName << endl;
         goto done;
      }

      if (!pbData)
      {
         cerr << "Error: pbData is not allocated." << endl;
         goto done;
      }

      mapData[string (pszDataName)]    = pbData;
      mapElement[string (pszDataName)] = new BinaryData (pbData, cbData);

      // Move to the next one
      deviceDataNode = nextNode (deviceDataNode);

      if (pszDataValue)
      {
         free ((char *)pszDataValue);
      }
      if (pszDataName)
      {
         free ((char *)pszDataName);
      }
      if (pszTypeName)
      {
         free ((char *)pszTypeName);
      }
   }

   rc = true;

   openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

   outputHeader (pofstreamHPP);

   *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
   *pofstreamHPP << "#define _" << *pstringNameOut << endl;
   *pofstreamHPP << "#include <DeviceData.hpp>" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceData" << endl;
   *pofstreamHPP << "{" << endl;
   *pofstreamHPP << "public:" << endl;
   *pofstreamHPP << "   " << *pstringNameOut << " ();" << endl;
   *pofstreamHPP << "};" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "#endif" << endl;

   outputHeader (pofstreamCPP);

   *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << *pstringNameOut << "::" << endl;
   *pofstreamCPP << *pstringNameOut << " ()" << endl;
   *pofstreamCPP << "{" << endl;

   for ( DataMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      BinaryData *bdValue = (BinaryData *)next->second;
      PBYTE       pbValue = bdValue->getData ();
      int         cbValue = bdValue->getLength ();

      *pofstreamCPP << "   static byte " << next->first << "[] = {" << dec;

      for (int i = 0; i < cbValue; i++)
      {
         *pofstreamCPP << (int)pbValue[i];
         if (i < cbValue - 1)
            *pofstreamCPP << ",";
      }

      *pofstreamCPP << "};" << endl;
   }

   *pofstreamCPP << endl;

   for ( DataMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      *pofstreamCPP << "   add (\""
                    << next->first
                    << "\", new BinaryData ("
                    << next->first
                    << ", dimof ("
                    << next->first << ")));"
                    << endl;
   }

   *pofstreamCPP << "}" << endl;

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printXMLDocument (deviceDatasNode->doc, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringDataClassName = 0;
   }

   for ( DataMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( DataMapData::iterator next = mapData.begin () ;
         next != mapData.end () ;
         next++ )
   {
      delete[] (*next).second;
   }

   return rc;
}


bool OmniDomParser::
preprocessDriver (xmlNodePtr driverNode)
{
   // This routine does two things.
   //    It fill is pstringInstanceClassName and pstringBlitterClassName
   //       in the DeviceInfo structure.
   //    It sees if it needs to process (copy) the instance and blitter files.

   xmlNodePtr  nodeItem    = driverNode->childs;
   const char *pszNodeName = 0;

   while (  NULL != nodeItem
         && 0 != strcmp ((char *)nodeItem->name, "Instance")
         )
   {
      // Move to the next one
      nodeItem = nextNode (nodeItem);
   }

   if (nodeItem == 0)
      return false;

   pszNodeName = bundleStringData (nodeItem);

   pdi_d->pstringInstanceClassName = converToClassName (pszNodeName);

   copyFile (pszNodeName);

   // Move to the next node
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      return false;

   if (0 != strcmp ((char *)nodeItem->name, "Instance"))
      return false;

   if (pszNodeName)
   {
      free ((char *)pszNodeName);
   }

   pszNodeName = bundleStringData (nodeItem);

   pdi_d->pstringInstanceClassName = converToClassName (pszNodeName);

   copyFile (pszNodeName);

   while (  NULL != nodeItem
         && 0 != strcmp ((char *)nodeItem->name, "Blitter")
         )
   {
      // Move to the next one
      nodeItem = nextNode (nodeItem);
   }

   if (nodeItem == 0)
      return false;

   if (pszNodeName)
   {
      free ((char *)pszNodeName);
   }

   pszNodeName = bundleStringData (nodeItem);

   pdi_d->pstringBlitterClassName = converToClassName (pszNodeName);

   copyFile (pszNodeName);

   // Move to the next node
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      return false;

   if (0 != strcmp ((char *)nodeItem->name, "Blitter"))
      return false;

   if (pszNodeName)
   {
      free ((char *)pszNodeName);
   }

   pszNodeName = bundleStringData (nodeItem);

   pdi_d->pstringBlitterClassName = converToClassName (pszNodeName);

   copyFile (pszNodeName);

   if (pszNodeName)
   {
      free ((char *)pszNodeName);
   }

   return true;
}

bool OmniDomParser::
processDriver (xmlNodePtr driverNode)
{
   bool  fFoundItem = false;
   char *pszComma   = 0;

   if (driverNode == 0)
      return false;

   preprocessDriver (driverNode);

   if (!shouldCreateFile (*pstrXMLFile_d))
   {
      cout << "Parsing:    \"" << *pstrXMLFile_d << "\"" << endl;
   }
   else
   {
      cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;
   }

   // Make sure required fields in the Device Info struture are filled in
   if (!pdi_d->pstringCommandClassName)
   {
      cerr << "Error: DeviceCommands not filled in." << endl;
      setErrorCondition ();
      return false;
   }
   if (!pdi_d->pstringResolutionClassName)
   {
      cerr << "Error: DeviceResolution not filled in." << endl;
      setErrorCondition ();
      return false;
   }
   if (!pdi_d->pstringPrintModeClassName)
   {
      cerr << "Error: DevicePrintMode not filled in." << endl;
      setErrorCondition ();
      return false;
   }
   if (!pdi_d->pstringTrayClassName)
   {
      cerr << "Error: DeviceTray not filled in." << endl;
      setErrorCondition ();
      return false;
   }
   if (!pdi_d->pstringFormClassName)
   {
      cerr << "Error: DeviceForm not filled in." << endl;
      setErrorCondition ();
      return false;
   }
   if (!pdi_d->pstringMediaClassName)
   {
      cerr << "Error: DeviceMedia not filled in." << endl;
      setErrorCondition ();
      return false;
   }
   if (!pdi_d->pstringOrientationClassName)
   {
      cerr << "Error: DeviceOrientation not filled in." << endl;
      setErrorCondition ();
      return false;
   }

   if (!shouldCreateFile (*pstrXMLFile_d))
      return true;

   typedef list <string> DriverList;

   const char      *pszDeviceName          = 0;
   const char      *pszDriverName          = 0;
   const char      *pszTypeName            = 0;
   const char      *pszMajorPDLLevel       = 0;
   const char      *pszMinorPDLLevel       = 0;
   const char      *pszMajorRevisionLevel  = 0;
   const char      *pszMinorRevisionLevel  = 0;
   const char      *pszOrientation         = 0;
   const char      *pszForm                = 0;
   const char      *pszTray                = 0;
   const char      *pszMedia               = 0;
   const char      *pszResolution          = 0;
   const char      *pszDither              = 0;
   const char      *pszPrintMode           = 0;
   xmlNodePtr       nodeItem               = driverNode->childs;
   string          *pstringNameOut         = getNameOut ();
   ofstream        *pofstreamHPP           = 0;
   ofstream        *pofstreamCPP           = 0;
   ofstream        *pofstreamMAK           = 0;
   DriverList       listCapabilities;
   DriverList       listRasterCapabilities;
   DriverList       listDeviceOptions;
   int              iCount                 = 0;
   xmlNodePtr       nodeOther              = 0;
   int              rc                     = false;

   pszDeviceName = (char *)xmlGetProp (driverNode, (xmlChar *)"name");

   if (fDebugOutput) cerr << "pszDeviceName = " << pszDeviceName << endl;

   pszDriverName = bundleStringData (nodeItem);

   if (fDebugOutput) cerr << "pszDriverName = " << pszDriverName << endl;

   // Move to the next node
   nodeItem = nextNode (nodeItem);
   if (nodeItem == NULL)
      goto done;

   if (fDebugOutput) printNodeName (nodeItem);

   while (  NULL != nodeItem
         && 0 == strcmp ((char *)nodeItem->name, "Capability")
         )
   {
      pszTypeName = (char *)xmlGetProp (nodeItem, (xmlChar *)"type");

      if (fDebugOutput) cerr << "pszTypeName = " << pszTypeName << endl;

      listCapabilities.push_back (string (pszTypeName));

      // Move to the next node
      nodeItem = nextNode (nodeItem);

      if (fDebugOutput && nodeItem) printNodeName (nodeItem);

      if (pszTypeName)
      {
         free ((char *)pszTypeName);
      }
   }

   if (nodeItem == NULL)
      goto done;

   while (  NULL != nodeItem
         && 0 == strcmp ((char *)nodeItem->name, "RasterCapabilities")
         )
   {
      pszTypeName = (char *)xmlGetProp (nodeItem, (xmlChar *)"type");

      if (fDebugOutput) cerr << "pszTypeName = " << pszTypeName << endl;

      listRasterCapabilities.push_back (string (pszTypeName));

      // Move to the next node
      nodeItem = nextNode (nodeItem);

      if (fDebugOutput && nodeItem) printNodeName (nodeItem);

      if (pszTypeName)
      {
         free ((char *)pszTypeName);
      }
   }

   if (nodeItem == NULL)
      goto done;

   while (  NULL != nodeItem
         && 0 == strcmp ((char *)nodeItem->name, "DeviceOptions")
         )
   {
      pszTypeName = (char *)xmlGetProp (nodeItem, (xmlChar *)"type");

      if (fDebugOutput) cerr << "pszTypeName = " << pszTypeName << endl;

      listDeviceOptions.push_back (string (pszTypeName));

      // Move to the next node
      nodeItem = nextNode (nodeItem);

      if (fDebugOutput && nodeItem) printNodeName (nodeItem);

      if (pszTypeName)
      {
         free ((char *)pszTypeName);
      }
   }

   if (nodeItem == NULL)
      goto done;

   if (0 != strcmp ((char *)nodeItem->name, "PDL"))
   {
      cerr << "Error: Could not find PDL tag" << endl;
      setErrorCondition ();
      goto done;
   }

   pszMajorPDLLevel = (char *)xmlGetProp (nodeItem, (xmlChar *)"level");
   if (pszMajorPDLLevel == NULL)
      goto done;
   if (fDebugOutput) cerr << "pszMajorPDLLevel = " << pszMajorPDLLevel << endl;

   pszMinorPDLLevel = (char *)xmlGetProp (nodeItem, (xmlChar *)"sublevel");
   if (pszMinorPDLLevel == NULL)
      goto done;
   if (fDebugOutput) cerr << "pszMinorPDLLevel = " << pszMinorPDLLevel << endl;

   pszMajorRevisionLevel = (char *)xmlGetProp (nodeItem, (xmlChar *)"major");
   if (pszMajorRevisionLevel == NULL)
      goto done;
   if (fDebugOutput) cerr << "pszMajorRevisionLevel = " << pszMajorRevisionLevel << endl;

   pszMinorRevisionLevel = (char *)xmlGetProp (nodeItem, (xmlChar *)"minor");
   if (pszMinorRevisionLevel == NULL)
      goto done;
   if (fDebugOutput) cerr << "pszMinorRevisionLevel = " << pszMinorRevisionLevel << endl;

   // Move to the next node
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   if (fDebugOutput) printNodeName (nodeItem);

   while (  NULL != nodeItem
         && (  0 == strcmp ((char *)nodeItem->name, "Uses")
            || 0 == strcmp ((char *)nodeItem->name, "Has")
            )
         )
   {
      // Move to the next node
      nodeItem = nextNode (nodeItem);

      if (fDebugOutput && nodeItem) printNodeName (nodeItem);
   }

   if (nodeItem == 0)
      goto done;

   if (0 != strcmp ((char *)nodeItem->name, "Instance"))
   {
      cerr << "Error: Could not find Instance tag" << endl;
      setErrorCondition ();
      goto done;
   }

   // Move to the next node
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   if (0 != strcmp ((char *)nodeItem->name, "Instance"))
   {
      cerr << "Error: Could not find Instance tag" << endl;
      setErrorCondition ();
      goto done;
   }

   // Move to the next node
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   if (0 != strcmp ((char *)nodeItem->name, "Blitter"))
   {
      cerr << "Error: Could not find Blitter tag" << endl;
      setErrorCondition ();
      goto done;
   }

   // Move to the next node
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   if (0 != strcmp ((char *)nodeItem->name, "Blitter"))
   {
      cerr << "Error: Could not find Blitter tag" << endl;
      setErrorCondition ();
      goto done;
   }

   // Move to the next node
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   if (0 == strcmp ((char *)nodeItem->name, "DeviceData"))
   {
      // Move to the next node
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;
   }

   // Move down into the DefaultJobProperties
   nodeItem = nodeItem->childs;
   if (nodeItem == 0)
      goto done;

   pszOrientation = bundleStringData (nodeItem);

   if (  NULL == pszOrientation
      || -1 == DeviceOrientation::nameToID (const_cast<char*>(pszOrientation))
      )
   {
      cerr << "Error: Unknown DeviceOrientation \"" << pszOrientation << "\"" << endl;
      setErrorCondition ();
      goto done;
   }

   fFoundItem = false;
   for ( OrientationList::iterator next = pdi_d->listOrientations.begin () ;
         next != pdi_d->listOrientations.end () ;
         next++ )
   {
      if (0 == strcmp (next->c_str (), pszOrientation))
      {
         fFoundItem = true;
      }
   }

   if (!fFoundItem)
   {
      cerr << "Error: \"" << pszOrientation << "\" is not defined for this device." << endl;
      setErrorCondition ();
      goto done;
   }

   if (fDebugOutput) cerr << "pszOrientation = " << pszOrientation << endl;

   // Move to the next node
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   pszForm = bundleStringData (nodeItem);

   if (pszForm)
   {
      pszComma = strchr (pszForm, ',');
      if (pszComma)
      {
         *pszComma = '\0';
      }
      if (-1 == DeviceForm::nameToID (const_cast<char*>(pszForm)))
      {
         cerr << "Error: Unknown DeviceForm \"" << pszForm << "\"" << endl;
         setErrorCondition ();
         goto done;
      }
   }
   else
   {
      cerr << "Error: Null DeviceForm \"" << pszForm << "\"" << endl;
      setErrorCondition ();
      goto done;
   }

   fFoundItem = false;
   for ( FormList::iterator next = pdi_d->listForms.begin () ;
         next != pdi_d->listForms.end () ;
         next++ )
   {
      if (0 == strcmp (next->c_str (), pszForm))
      {
         fFoundItem = true;
      }
   }

   if (!fFoundItem)
   {
      cerr << "Error: \"" << pszForm << "\" is not defined for this device." << endl;
      setErrorCondition ();
      goto done;
   }

   if (pszComma)
   {
      *pszComma = ',';
   }

   if (fDebugOutput) cerr << "pszForm = " << pszForm << endl;

   // Move to the next node
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   pszTray = bundleStringData (nodeItem);

   if (  NULL == pszTray
      || -1 == DeviceTray::nameToID (const_cast<char*>(pszTray))
      )
   {
      cerr << "Error: Unknown DeviceTray \"" << pszTray << "\"" << endl;
      setErrorCondition ();
      goto done;
   }

   fFoundItem = false;
   for ( TrayList::iterator next = pdi_d->listTrays.begin () ;
         next != pdi_d->listTrays.end () ;
         next++ )
   {
      if (0 == strcmp (next->c_str (), pszTray))
      {
         fFoundItem = true;
      }
   }

   if (!fFoundItem)
   {
      cerr << "Error: \"" << pszTray << "\" is not defined for this device." << endl;
      setErrorCondition ();
      goto done;
   }

   if (fDebugOutput) cerr << "pszTray = " << pszTray << endl;

   // Move to the next node
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   pszMedia = bundleStringData (nodeItem);

   if (  NULL == pszMedia
      || -1 == DeviceMedia::nameToID (const_cast<char*>(pszMedia))
      )
   {
      cerr << "Error: Unknown DeviceMedia \"" << pszMedia << "\"" << endl;
      setErrorCondition ();
      goto done;
   }

   fFoundItem = false;
   for ( MediaList::iterator next = pdi_d->listMedias.begin () ;
         next != pdi_d->listMedias.end () ;
         next++ )
   {
      if (0 == strcmp (next->c_str (), pszMedia))
      {
         fFoundItem = true;
      }
   }

   if (!fFoundItem)
   {
      cerr << "Error: \"" << pszMedia << "\" is not defined for this device." << endl;
      setErrorCondition ();
      goto done;
   }

   if (fDebugOutput) cerr << "pszMedia = " << pszMedia << endl;

   // Move to the next node
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   pszResolution = bundleStringData (nodeItem);

   if (  NULL == pszResolution
      || -1 == DeviceResolution::nameToID (const_cast<char*>(pszResolution))
      )
   {
      cerr << "Error: Unknown DeviceResolution \"" << pszResolution << "\"" << endl;
      setErrorCondition ();
      goto done;
   }

   fFoundItem = false;
   for ( ResolutionList::iterator next = pdi_d->listResolutions.begin () ;
         next != pdi_d->listResolutions.end () ;
         next++ )
   {
      if (0 == strcmp (next->c_str (), pszResolution))
      {
         fFoundItem = true;
      }
   }

   if (!fFoundItem)
   {
      cerr << "Error: \"" << pszResolution << "\" is not defined for this device." << endl;
      setErrorCondition ();
      goto done;
   }

   if (fDebugOutput) cerr << "pszResolution = " << pszResolution << endl;

   // Move to the next node
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   pszDither = bundleStringData (nodeItem);

   if (  NULL == pszDither
      || !DeviceDither::ditherNameValid (const_cast<char*>(pszDither))
      )
   {
      cerr << "Error: Unknown DeviceDither \"" << pszDither << "\"" << endl;
      setErrorCondition ();
      goto done;
   }

   if (fDebugOutput) cerr << "pszDither = " << pszDither << endl;

   // Move to the next node
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   pszPrintMode = bundleStringData (nodeItem);

   if (  NULL == pszPrintMode
      || -1 == DevicePrintMode::nameToID (const_cast<char*>(pszPrintMode))
      )
   {
      cerr << "Error: Unknown DevicePrintMode \"" << pszPrintMode << "\"" << endl;
      setErrorCondition ();
      goto done;
   }

   fFoundItem = false;
   for ( PrintModeList::iterator next = pdi_d->listPrintModes.begin () ;
         next != pdi_d->listPrintModes.end () ;
         next++ )
   {
      if (0 == strcmp (next->c_str (), pszPrintMode))
      {
         fFoundItem = true;
      }
   }

   if (!fFoundItem)
   {
      cerr << "Error: \"" << pszPrintMode << "\" is not defined for this device." << endl;
      setErrorCondition ();
      goto done;
   }

   if (fDebugOutput) cerr << "pszPrintMode = " << pszPrintMode << endl;

   // Move to the next node
   nodeItem = nextNode (nodeItem);

   nodeOther = nodeItem;

   // process optional other(s)
   while (nodeItem)
   {
      if (0 != (strcmp ((const char *)nodeItem->name, "other")))
          goto done;

      // Move to the next node
      nodeItem = nextNode (nodeItem);
   }

   rc = true;

   listCapabilities.sort ();
   listRasterCapabilities.sort ();
   listDeviceOptions.sort ();
   listCapabilities.unique ();
   listRasterCapabilities.unique ();
   listDeviceOptions.unique ();

   openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

   outputHeader (pofstreamHPP);

   *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
   *pofstreamHPP << "#define _" << *pstringNameOut << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "#include <" << *pdi_d->pstringFormClassName << ".hpp>" << endl;
   *pofstreamHPP << "#include <" << *pdi_d->pstringTrayClassName << ".hpp>" << endl;
   *pofstreamHPP << "#include <" << *pdi_d->pstringMediaClassName << ".hpp>" << endl;
   *pofstreamHPP << "#include <" << *pdi_d->pstringResolutionClassName << ".hpp>" << endl;
   *pofstreamHPP << "#include <" << *pdi_d->pstringOrientationClassName << ".hpp>" << endl;
   *pofstreamHPP << "#include <" << *pdi_d->pstringCommandClassName << ".hpp>" << endl;
   if (pdi_d->pstringDataClassName)
      *pofstreamHPP << "#include <" << *pdi_d->pstringDataClassName << ".hpp>" << endl;
   *pofstreamHPP << "#include <" << *pdi_d->pstringPrintModeClassName << ".hpp>" << endl;
   if (pdi_d->pstringInstanceClassName)
      *pofstreamHPP << "#include <" << *pdi_d->pstringInstanceClassName << ".hpp>" << endl;
   if (pdi_d->pstringBlitterClassName)
      *pofstreamHPP << "#include <" << *pdi_d->pstringBlitterClassName << ".hpp>" << endl;
   if (pdi_d->pstringGammaClassName)
      *pofstreamHPP << "#include <" << *pdi_d->pstringGammaClassName << ".hpp>" << endl;
   *pofstreamHPP << "#include <DeviceInstance.hpp>" << endl;
   *pofstreamHPP << "#include <PrintDevice.hpp>" << endl;
   *pofstreamHPP << "#include <JobProperties.hpp>" << endl;
   *pofstreamHPP << "#include <OmniProxy.hpp>" << endl;
   *pofstreamHPP << endl;

   *pofstreamHPP << "class " << *pstringNameOut << " : public PrintDevice" << endl;
   *pofstreamHPP << "{" << endl;
   *pofstreamHPP << "public:" << endl;

   iCount = listRasterCapabilities.size ();
   if (0 < iCount)
   {
      int iValue = 1;

      *pofstreamHPP << "   enum {" << endl;

      for ( DriverList::iterator next = listRasterCapabilities.begin () ;
            next != listRasterCapabilities.end () ;
            next++, iCount-- )
      {
         *pofstreamHPP << "      " << *next << " = " << iValue;
         if (1 < iCount)
            *pofstreamHPP << ",";
         *pofstreamHPP << endl;

         iValue *= 2;
      }

      *pofstreamHPP << "   };" << endl;
      *pofstreamHPP << endl;
   }

   *pofstreamHPP << "   " << *pstringNameOut << " ();" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "   " << *pstringNameOut << " (char *pszJobProperties);" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "   virtual DeviceGamma       *getCurrentGamma          ();" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "   virtual DeviceOrientation *getDefaultOrientation    ();" << endl;
   *pofstreamHPP << "   virtual char              *getDefaultDitherID       ();" << endl;
   *pofstreamHPP << "   virtual DeviceForm        *getDefaultForm           ();" << endl;
   *pofstreamHPP << "   virtual DeviceTray        *getDefaultTray           ();" << endl;
   *pofstreamHPP << "   virtual DeviceMedia       *getDefaultMedia          ();" << endl;
   *pofstreamHPP << "   virtual DeviceResolution  *getDefaultResolution     ();" << endl;
   *pofstreamHPP << "   virtual DeviceCommand     *getDefaultCommands       ();" << endl;
   *pofstreamHPP << "   virtual DeviceData        *getDefaultData           ();" << endl;
   *pofstreamHPP << "   virtual DevicePrintMode   *getDefaultPrintMode      ();" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "private:" << endl;
   *pofstreamHPP << "   void commonInit ();" << endl;
   *pofstreamHPP << "};" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "#endif" << endl;

   outputHeader (pofstreamCPP);

   *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
   *pofstreamCPP << endl;

   *pofstreamCPP << "Device *" << endl;
   *pofstreamCPP << "newDevice (bool fAdvanced)" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   PrintDevice *pPrintDevice = new " << *pstringNameOut << " ();" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << "   pPrintDevice->initialize ();" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << "   if (fAdvanced)" << endl;
   *pofstreamCPP << "      return pPrintDevice;" << endl;
   *pofstreamCPP << "   else" << endl;
   *pofstreamCPP << "      return new OmniProxy (pPrintDevice);" << endl;
   *pofstreamCPP << "}" << endl;
   *pofstreamCPP << endl;

   *pofstreamCPP << "Device *" << endl;
   *pofstreamCPP << "newDevice (char *pszJobProperties, bool fAdvanced)" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   PrintDevice *pPrintDevice = new " << *pstringNameOut << " (pszJobProperties);" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << "   pPrintDevice->initialize ();" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << "   if (fAdvanced)" << endl;
   *pofstreamCPP << "      return pPrintDevice;" << endl;
   *pofstreamCPP << "   else" << endl;
   *pofstreamCPP << "      return new OmniProxy (pPrintDevice);" << endl;
   *pofstreamCPP << "}" << endl;
   *pofstreamCPP << endl;

   *pofstreamCPP << "void" << endl;
   *pofstreamCPP << "deleteDevice (Device *pDevice)" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   delete pDevice;" << endl;
   *pofstreamCPP << "}" << endl;
   *pofstreamCPP << endl;

   *pofstreamCPP << "void " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "commonInit ()" << endl;
   *pofstreamCPP << "{" << endl;

   iCount = listCapabilities.size ();
   if (0 < iCount)
   {
      *pofstreamCPP << "   setCapabilities ( ";

      for ( DriverList::iterator next = listCapabilities.begin () ;
            next != listCapabilities.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << "Capability::" << *next << endl;
         if (1 < iCount)
            *pofstreamCPP << "                   | ";
      }

      *pofstreamCPP << "                   );" << endl;
   }

   iCount = listRasterCapabilities.size ();
   if (0 < iCount)
   {
      *pofstreamCPP << endl;
      *pofstreamCPP << "   setRasterCapabilities ( ";

      for ( DriverList::iterator next = listRasterCapabilities.begin () ;
            next != listRasterCapabilities.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << *next << endl;
         if (1 < iCount)
            *pofstreamCPP << "                   | ";
      }

      *pofstreamCPP << "                         );" << endl;
   }

   iCount = listDeviceOptions.size ();
   if (0 < iCount)
   {
      *pofstreamCPP << endl;
      *pofstreamCPP << "   setDeviceOptions ( ";

      for ( DriverList::iterator next = listDeviceOptions.begin () ;
            next != listDeviceOptions.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << *pdi_d->pstringInstanceClassName << "::" << *next << endl;
         if (1 < iCount)
            *pofstreamCPP << "                    | ";
      }

      *pofstreamCPP << "                    );" << endl;
   }

   *pofstreamCPP << endl;
   if (pdi_d->pstringInstanceClassName)
      *pofstreamCPP << "   setDeviceInstance (new " << *pdi_d->pstringInstanceClassName << " (this));" << endl;
   if (pdi_d->pstringBlitterClassName)
      *pofstreamCPP << "   setDeviceBlitter (new " << *pdi_d->pstringBlitterClassName << " (this));" << endl;

   if (  pszMajorPDLLevel
      && pszMinorPDLLevel
      && pszMajorRevisionLevel
      && pszMinorRevisionLevel
      )
   {
      *pofstreamCPP << "   setPDL (new PDL (PDL::"
                    << pszMajorPDLLevel
                    << ",PDL::"
                    << pszMinorPDLLevel
                    << ","
                    << pszMajorRevisionLevel
                    << ","
                    << pszMinorRevisionLevel
                    << "));"
                    << endl;
   }

   *pofstreamCPP << "}" << endl;
   *pofstreamCPP << endl;

   *pofstreamCPP << *pstringNameOut << "::" << endl;
   *pofstreamCPP << *pstringNameOut << " ()" << endl;
   *pofstreamCPP << "   : PrintDevice (\"" << pszDriverName << "\"," << endl;
   *pofstreamCPP << "                  \"" << pszDeviceName << "\"," << endl;
   *pofstreamCPP << "                  \"" << *pstringNameOut << "\"," << endl;
   *pofstreamCPP << "                  \"orientation=" << pszOrientation;
   *pofstreamCPP << " dither=" << pszDither;
   *pofstreamCPP << " form=" << pszForm;
   *pofstreamCPP << " tray=" << pszTray;
   *pofstreamCPP << " media=" << pszMedia;
   *pofstreamCPP << " resolution=" << pszResolution;
   *pofstreamCPP << " printmode=" << pszPrintMode;

   // process optional other(s)
   while (nodeOther)
   {
      const char *pszId = bundleStringData (nodeOther);

      *pofstreamCPP << " " << pszId;

      // Move to the next node
      nodeOther = nextNode (nodeOther);

      if (pszId)
      {
         free ((char *)pszId);
      }
   }

   *pofstreamCPP << "\")" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   commonInit ();" << endl;
   *pofstreamCPP << "}" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << *pstringNameOut << "::" << endl;
   *pofstreamCPP << *pstringNameOut << " (char *pszJobProperties)" << endl;
   *pofstreamCPP << "   : PrintDevice (\"" << pszDriverName << "\"," << endl;
   *pofstreamCPP << "                  \"" << pszDeviceName << "\"," << endl;
   *pofstreamCPP << "                  \"" << *pstringNameOut << "\"," << endl;
   *pofstreamCPP << "                  pszJobProperties)" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   commonInit ();" << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceGamma * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getCurrentGamma ()" << endl;
   *pofstreamCPP << "{" << endl;
   if (pdi_d->pstringGammaClassName)
   {
      *pofstreamCPP << "   return " << *pdi_d->pstringGammaClassName << "::getDeviceGamma (getCurrentResolution ()->getID ()," << endl;
      *pofstreamCPP << "                            getCurrentMedia ()->getID ()," << endl;
      *pofstreamCPP << "                            getCurrentPrintMode ()->getID ()," << endl;
      *pofstreamCPP << "                            getCurrentDitherID ());" << endl;
   }
   else
   {
      *pofstreamCPP << "   return 0;" << endl;
   }
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceOrientation * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultOrientation ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return " << *pdi_d->pstringOrientationClassName << "::create (" << *pdi_d->pstringOrientationClassName << "::" << pszOrientation << ");" << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "char * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultDitherID ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return \"" << pszDither << "\";" << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceForm * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultForm ()" << endl;
   *pofstreamCPP << "{" << endl;

   if (pszComma)
   {
      int iX = 0,
          iY = 0;

      *pszComma = '\0';

      *pofstreamCPP << "   DeviceForm *pDF = " << *pdi_d->pstringFormClassName << "::create (" << *pdi_d->pstringFormClassName << "::" << pszForm << ");"  << endl;

      pszComma++;
      iX = parseSize (&pszComma);

      if (',' == *pszComma)
      {
         pszComma++;
         iY = parseSize (&pszComma);
      }

      if (  iX != 0
         && iY != 0
         )
      {
         *pofstreamCPP << "   pDF->setCx (" << iX << ");" << endl;
         *pofstreamCPP << "   pDF->setCy (" << iY << ");" << endl;
      }
      else if (iX != 0)
      {
         *pofstreamCPP << "   pDF->setCy (" << iX << ");" << endl;
      }

      *pofstreamCPP << "   return pDF;" << endl;
   }
   else
   {
      *pofstreamCPP << "   return " << *pdi_d->pstringFormClassName << "::create (" << *pdi_d->pstringFormClassName << "::" << pszForm << ");"  << endl;
   }

   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceTray * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultTray ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return " << *pdi_d->pstringTrayClassName << "::create (" << *pdi_d->pstringTrayClassName << "::" << pszTray << ");"  << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceMedia * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultMedia ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return " << *pdi_d->pstringMediaClassName << "::create (" << *pdi_d->pstringMediaClassName << "::" << pszMedia << ");"  << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceResolution * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultResolution ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return " << *pdi_d->pstringResolutionClassName << "::create (" << *pdi_d->pstringResolutionClassName << "::" << pszResolution << ");"  << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceCommand * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultCommands ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return new " << *pdi_d->pstringCommandClassName << " ();" << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceData * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultData ()" << endl;
   *pofstreamCPP << "{" << endl;
   if (pdi_d->pstringDataClassName)
      *pofstreamCPP << "   return new " << *pdi_d->pstringDataClassName << " ();" << endl;
   else
      *pofstreamCPP << "   return 0;" << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DevicePrintMode * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultPrintMode ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return " << *pdi_d->pstringPrintModeClassName << "::create (" << *pdi_d->pstringPrintModeClassName << "::" << pszPrintMode << ");"  << endl;
   *pofstreamCPP << "}" << endl;

   openMakeFile (pstringNameOut, &pofstreamMAK);

   *pofstreamMAK << "DEVICENAME            = " << *pstringNameOut << endl;
   *pofstreamMAK << "DEVCFILES             = " << *pdi_d->pstringPrintModeClassName << "~ \\" << endl;
   *pofstreamMAK << "                        " << *pdi_d->pstringResolutionClassName << "~ \\" << endl;
   *pofstreamMAK << "                        " << *pdi_d->pstringOrientationClassName << "~ \\" << endl;
   *pofstreamMAK << "                        " << *pdi_d->pstringCommandClassName << "~ \\" << endl;
   if (pdi_d->pstringGammaClassName)
      *pofstreamMAK << "                        " << *pdi_d->pstringGammaClassName << "~ \\" << endl;
   *pofstreamMAK << "                        " << *pdi_d->pstringTrayClassName << "~ \\" << endl;
   *pofstreamMAK << "                        " << *pdi_d->pstringFormClassName << "~ \\" << endl;
   *pofstreamMAK << "                        " << *pdi_d->pstringBlitterClassName << "~ \\" << endl;
   *pofstreamMAK << "                        " << *pdi_d->pstringInstanceClassName << "~ \\" << endl;
   *pofstreamMAK << "                        " << *pdi_d->pstringMediaClassName << "~ \\" << endl;
   if (pdi_d->pstringDataClassName)
      *pofstreamMAK << "                        " << *pdi_d->pstringDataClassName << "~ \\" << endl;
   *pofstreamMAK << "                        " << *pstringNameOut << "~" << endl;
   *pofstreamMAK << "OMNI_PATH             = .." << endl;
   *pofstreamMAK << endl;
   *pofstreamMAK << "all:    lib$(DEVICENAME).so" << endl;
   *pofstreamMAK << endl;
   *pofstreamMAK << "include ../common.mak" << endl;
   *pofstreamMAK << endl;
   *pofstreamMAK << *pstringNameOut
                 << ".o:\t"
                 << *pstringNameOut
                 << ".cpp "
                 << *pdi_d->pstringFormClassName
                 << ".o "
                 << *pdi_d->pstringResolutionClassName
                 << ".o "
                 << *pdi_d->pstringMediaClassName
                 << ".o "
                 << *pdi_d->pstringTrayClassName
                 << ".o "
                 << *pdi_d->pstringOrientationClassName
                 << ".o "
                 << *pdi_d->pstringCommandClassName
                 << ".o "
                 << "../Device.o"
                 << endl;
   *pofstreamMAK << *pdi_d->pstringFormClassName
                 << ".o:\t"
                 << *pdi_d->pstringFormClassName
                 << ".cpp "
                 << *pdi_d->pstringFormClassName
                 << ".hpp "
                 << "../DeviceForm.o"
                 << endl;
   *pofstreamMAK << *pdi_d->pstringResolutionClassName
                 << ".o:\t"
                 << *pdi_d->pstringResolutionClassName
                 << ".cpp "
                 << *pdi_d->pstringResolutionClassName
                 << ".hpp "
                 << "../DeviceResolution.o"
                 << endl;
   *pofstreamMAK << *pdi_d->pstringMediaClassName
                 << ".o:\t"
                 << *pdi_d->pstringMediaClassName
                 << ".cpp "
                 << *pdi_d->pstringMediaClassName
                 << ".hpp "
                 << "../DeviceMedia.o"
                 << endl;
   *pofstreamMAK << *pdi_d->pstringTrayClassName
                 << ".o:\t"
                 << *pdi_d->pstringTrayClassName
                 << ".cpp "
                 << *pdi_d->pstringTrayClassName
                 << ".hpp "
                 << "../DeviceTray.o"
                 << endl;
   *pofstreamMAK << *pdi_d->pstringOrientationClassName
                 << ".o:\t"
                 << *pdi_d->pstringOrientationClassName
                 << ".cpp "
                 << *pdi_d->pstringOrientationClassName
                 << ".hpp "
                 << "../DeviceOrientation.o"
                 << endl;
   *pofstreamMAK << *pdi_d->pstringCommandClassName
                 << ".o:\t"
                 << *pdi_d->pstringCommandClassName
                 << ".cpp "
                 << *pdi_d->pstringCommandClassName
                 << ".hpp "
                 << "../DeviceCommand.o"
                 << endl;
   *pofstreamMAK << *pdi_d->pstringPrintModeClassName
                 << ".o:\t"
                 << *pdi_d->pstringPrintModeClassName
                 << ".cpp "
                 << *pdi_d->pstringPrintModeClassName
                 << ".hpp "
                 << "../DevicePrintMode.o"
                 << endl;
///*pofstreamMAK << *pdi_d->pstringConnectionClassName
///              << ".o:\t"
///              << *pdi_d->pstringConnectionClassName
///              << ".cpp "
///              << *pdi_d->pstringConnectionClassName
///              << ".hpp "
///              << "../DeviceConnection.o"
///              << endl;
   if (pdi_d->pstringGammaClassName != 0)
   {
      *pofstreamMAK << *pdi_d->pstringGammaClassName
                    << ".o:\t"
                    << *pdi_d->pstringGammaClassName
                    << ".cpp "
                    << *pdi_d->pstringGammaClassName
                    << ".hpp "
                    << *pdi_d->pstringResolutionClassName
                    << ".hpp "
                    << *pdi_d->pstringMediaClassName
                    << ".hpp "
                    << *pdi_d->pstringPrintModeClassName
                    << ".hpp "
                    << "../DeviceGammaTable.hpp "
                    << "../DeviceGamma.hpp "
                    << "../DeviceResolution.hpp "
                    << "../DeviceMedia.hpp "
                    << "../DevicePrintMode.hpp "
                    << "../Enumeration.hpp "
                    << "../BinaryData.hpp "
                    << "../DeviceDither.hpp "
                    << "../defines.hpp"
                    << endl;
   }
   *pofstreamMAK << *pdi_d->pstringBlitterClassName
                 << ".o:\t"
                 << *pdi_d->pstringBlitterClassName
                 << ".cpp "
                 << *pdi_d->pstringBlitterClassName
                 << ".hpp "
                 << "../DeviceBlitter.o "
                 << *pdi_d->pstringInstanceClassName
                 << ".hpp "
                 << *pdi_d->pstringPrintModeClassName
                 << ".hpp "
                 << *pdi_d->pstringResolutionClassName
                 << ".hpp "
                 << "../JobProperties.hpp"
                 << endl;
   *pofstreamMAK << *pdi_d->pstringInstanceClassName
                 << ".o:\t"
                 << *pdi_d->pstringInstanceClassName
                 << ".cpp "
                 << *pdi_d->pstringInstanceClassName
                 << ".hpp "
                 << "../DeviceInstance.o"
                 << endl;
   if (pdi_d->pstringDataClassName)
   {
      *pofstreamMAK << *pdi_d->pstringDataClassName
                    << ".o:\t"
                    << *pdi_d->pstringDataClassName
                    << ".cpp "
                    << *pdi_d->pstringDataClassName
                    << ".hpp "
                    << "../DeviceData.o"
                    << endl;
   }

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printXMLDocument (driverNode->doc, 0);
   }

   // Clean up!
   if (pszDeviceName)
   {
      free ((char *)pszDeviceName);
   }
   if (pszDriverName)
   {
      free ((char *)pszDriverName);
   }
   if (pszOrientation)
   {
      free ((char *)pszOrientation);
   }
   if (pszForm)
   {
      free ((char *)pszForm);
   }
   if (pszTray)
   {
      free ((char *)pszTray);
   }
   if (pszMedia)
   {
      free ((char *)pszMedia);
   }
   if (pszResolution)
   {
      free ((char *)pszResolution);
   }
   if (pszDither)
   {
      free ((char *)pszDither);
   }
   if (pszPrintMode)
   {
      free ((char *)pszPrintMode);
   }
   if (pszMajorPDLLevel)
   {
      free ((char *)pszMajorPDLLevel);
   }
   if (pszMinorPDLLevel)
   {
      free ((char *)pszMinorPDLLevel);
   }
   if (pszMajorRevisionLevel)
   {
      free ((char *)pszMajorRevisionLevel);
   }
   if (pszMinorRevisionLevel)
   {
      free ((char *)pszMinorRevisionLevel);
   }

   delete pofstreamHPP;
   delete pofstreamCPP;
   delete pofstreamMAK;
   delete pstringNameOut;

   return rc;
}

void OmniDomParser::
processDOMNode (xmlNodePtr node)
{
   char *pszElementName = (char *)node->name;

   if (  0 == strcmp ("Uses", pszElementName)
      || 0 == strcmp ("Has", pszElementName)
      )
   {
      char *pszFilename = (char *)xmlNodeListGetString (node->doc, node->childs, 1);

      OmniDomParser parser (pszFilename, pdi_d);

      parser.parse ();

      if (pszFilename)
      {
         free ((char *)pszFilename);
      }
   }
   else if (0 == strcmp ("deviceOrientations", pszElementName))
      processDeviceOrientations (node);
   else if (0 == strcmp ("deviceCommands", pszElementName))
      processDeviceCommands (node);
   else if (0 == strcmp ("deviceResolutions", pszElementName))
      processDeviceResolutions (node);
   else if (0 == strcmp ("devicePrintModes", pszElementName))
      processDevicePrintModes (node);
   else if (0 == strcmp ("deviceTrays", pszElementName))
      processDeviceTrays (node);
   else if (0 == strcmp ("deviceForms", pszElementName))
      processDeviceForms (node);
   else if (0 == strcmp ("deviceMedias", pszElementName))
      processDeviceMedias (node);
   else if (0 == strcmp ("deviceConnections", pszElementName))
      processDeviceConnections (node);
   else if (0 == strcmp ("deviceGammaTables", pszElementName))
      processDeviceGammas (node);
   else if (0 == strcmp ("deviceDatas", pszElementName))
      processDeviceDatas (node);
   else if (0 == strcmp ("Device", pszElementName))
   {
      xmlNodePtr cur = node->childs;

      while (cur)
      {
         if (fDebugOutput) printXMLNode (cur);

         processDOMNode (cur);

         // Move to the next one
         cur = nextNode (cur);
      }

      processDriver (node);
   }
}

bool OmniDomParser::
parse ()
{
   const char *pszXMLFile = pstrXMLFile_d->c_str ();
   xmlValidCtxt validator = {
      this,
      MyValidityErrorFunc,
      MyValidityWarningFunc
   };

   if (fDebugOutput)
      cerr << "DOM Parsing: " << pszXMLFile << endl;

   doc_d = xmlParseFile (pszXMLFile);

   if (fDebugOutput)
      cerr << "doc_d = 0x" << hex << (int)doc_d << dec << endl;

   if (doc_d)
   {
      xmlValidateDocument (&validator, doc_d);

      if (fDebugOutput)
         cerr << "Document validated." << endl;

      processDOMNode (xmlDocGetRootElement (doc_d));
   }
   else
   {
      cerr << "Error: File not found \"" << pszXMLFile << "\"" << endl;

      fErrorCondition_d = true;
   }

   return !fErrorCondition_d;
}
